﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace gfx_convert_library.Store
{
    class Save
    {
        static public bool write(string fileName, writeMode_t type, DataStream datas, long forceLength, int dataWidth = 32,string extra = "")
        {
            if (type == writeMode_t.RAW)
            {
                return writeRaw(fileName, datas, forceLength);
            }
            else if (type == writeMode_t.MIF)
            {
                return writeMif(fileName, datas, forceLength, dataWidth);
            }
            else if (type == writeMode_t.ASM)
            {
                return writeASM(fileName, datas, forceLength, dataWidth, extra);
            }
            else if (type == writeMode_t.C)
            {
                return writeC(fileName, datas, forceLength, dataWidth);
            }
            return false;
        }

        static bool writeRaw(string fileName, DataStream datas, long forceLength)
        {
            FileStream fs = File.Create(fileName + ".raw");
            BinaryWriter bw = new BinaryWriter(fs);

            long dataLength = datas.Length;
            long totalLength = forceLength > dataLength ? forceLength : dataLength;

            if (totalLength >= 0)
            {

                byte[] buffer = new byte[totalLength];

                for (int i = 0; i < totalLength; i++)
                {
                    buffer[i] = 0;
                }

                datas.Read(buffer, 0, (int)dataLength);

                bw.Write(buffer);
            }

            bw.Close();
            fs.Close();

            return true;
        }

        static bool writeMif(string fileName, DataStream datas, long forceLength, int dataWidth)
        {
            using (FileStream fs = File.Open(fileName + ".mif", FileMode.Create, FileAccess.Write, FileShare.None))
            {
                long dataLength = datas.Length;
                long totalLength = forceLength > dataLength ? forceLength : dataLength;

                //

                StringBuilder stringBuilder = new StringBuilder();

                // generate head

                stringBuilder.Append(string.Format("WIDTH={0};\n"
                                    + "DEPTH={1};\n"
                                    + "ADDRESS_RADIX=UNS;\n"
                                    + "DATA_RADIX=HEX;\n"
                                    + "CONTENT BEGIN\n", dataWidth, totalLength * 8 / dataWidth));


                // data

                if (dataWidth == 8)
                {
                    for (int adr = 0; adr < totalLength; adr++)
                    {
                        byte v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("    {0:D8} : {1:X2};\n", adr, v));
                    }
                }
                else if (dataWidth == 16)
                {
                    for (int adr = 0; adr < totalLength / 2; adr++)
                    {
                        UInt16 v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("    {0:D8} : {1:X4};\n", adr, v));
                    }
                }
                else if (dataWidth == 32)
                {
                    for (int adr = 0; adr < totalLength / 4; adr++)
                    {
                        UInt32 v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("    {0:D8} : {1:X8};\n", adr, v));
                    }
                }
                else
                {
                    throw new Exception("unsupported bit witdh");
                }

                stringBuilder.Append("END;\n");

                String outString = stringBuilder.ToString();

                // write data

                foreach (char c in outString)
                {
                    fs.WriteByte(Convert.ToByte(c));
                }

                fs.Close();
            }

            return true;
        }

        static bool writeASM(string fileName, DataStream datas, long forceLength, int dataWidth, string opcodeName)
        {
            using (FileStream fs = File.Open(fileName + ".asm", FileMode.Create, FileAccess.Write, FileShare.None))
            {
                long dataLength = datas.Length;
                long totalLength = forceLength > dataLength ? forceLength : dataLength;

                if (opcodeName.Length == 0)
                {
                    opcodeName = "word";
                }

                //

                StringBuilder stringBuilder = new StringBuilder();

                // generate head

                stringBuilder.Append(string.Format("\t\t; bit width={0} length = {1}\n", dataWidth, totalLength * 8 / dataWidth));


                // data

                if (dataWidth == 8)
                {
                    for (int adr = 0; adr < totalLength; adr++)
                    {
                        byte v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("    "+opcodeName+"    ${0:X2}  ;at {1:D8}\n", v, adr));
                    }
                }
                else if (dataWidth == 16)
                {
                    for (int adr = 0; adr < totalLength / 2; adr++)
                    {
                        UInt16 v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("    " + opcodeName + "    ${0:X4}  ;at {1:D8}\n", v, adr));
                    }
                }
                else if (dataWidth == 32)
                {
                    for (int adr = 0; adr < totalLength / 4; adr++)
                    {
                        UInt32 v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("    " + opcodeName + "    ${0:X8}  ;at {1:D8}\n", v, adr));
                    }
                }
                else
                {
                    throw new Exception("unsupported bit witdh");
                }

                String outString = stringBuilder.ToString();

                // write data

                foreach (char c in outString)
                {
                    fs.WriteByte(Convert.ToByte(c));
                }

                fs.Close();
            }

            return true;
        }

        static bool writeC(string fileName, DataStream datas, long forceLength, int dataWidth)
        {
            using (FileStream fs = File.Open(fileName + ".c", FileMode.Create, FileAccess.Write, FileShare.None))
            {
                long dataLength = datas.Length;
                long totalLength = forceLength > dataLength ? forceLength : dataLength;

                //

                StringBuilder stringBuilder = new StringBuilder();

                // generate head

                stringBuilder.Append("data = {");


                // data

                if (dataWidth == 8)
                {
                    for (int adr = 0; adr < totalLength; adr++)
                    {
                        byte v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("0x{0:X2}{1}", v, adr != (totalLength - 1) ? "," : ""));

                        if (adr % 1024 == 0)
                        {
                            stringBuilder.Append("\n");
                        }
                    }
                }
                else if (dataWidth == 16)
                {
                    for (int adr = 0; adr < totalLength / 2; adr++)
                    {
                        UInt16 v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("0x{0:X4}{1}", v, adr != (totalLength / 2 - 1) ? "," : ""));

                        if (adr % 1024 == 1023)
                        {
                            stringBuilder.Append("\n");
                        }
                    }
                }
                else if (dataWidth == 32)
                {
                    for (int adr = 0; adr < totalLength / 4; adr++)
                    {
                        UInt32 v = 0;
                        if (adr < dataLength)
                        {
                            datas.Read(out v, adr);
                        }
                        stringBuilder.Append(string.Format("0x{0:X8}{1}", v, adr != (totalLength / 4 - 1) ? "," : ""));

                        if (adr % 1024 == 0)
                        {
                            stringBuilder.Append("\n");
                        }
                    }
                }
                else
                {
                    throw new Exception("unsupported bit witdh");
                }

                stringBuilder.Append("}\n");
                String outString = stringBuilder.ToString();

                // write data

                foreach (char c in outString)
                {
                    fs.WriteByte(Convert.ToByte(c));
                }

                fs.Close();
            }

            return true;
        }

    }
}
