﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;
using System.ComponentModel;

namespace raw_device_io
{
    public class raw_device_access
    {
        #region EXT_ACCESS
        const int FILE_FLAG_NO_BUFFERING = unchecked((int)0x20000000);
        const int FILE_FLAG_SEQUENTIAL_SCAN = unchecked((int)0x08000000);

        [DllImport("kernel32", SetLastError = true, CharSet = CharSet.Unicode)]
        static extern IntPtr CreateFile(String fileName,
        int
        desiredAccess,
        System.IO.FileShare
        shareMode,
        IntPtr
        securityAttrs,
        System.IO.FileMode
        creationDisposition,
        int
        flagsAndAttributes,
        IntPtr
        templateFile);
        #endregion

        public List<DriveInfo> getDevices(params DriveType[] deviceFilter)
        {
            List<DriveInfo> rList = new List<DriveInfo>();
            DriveInfo[] driveInfos = DriveInfo.GetDrives();

            foreach (DriveInfo di in driveInfos)
            {
                bool filter = true;
                foreach (DriveType type in deviceFilter)
                {
                    if (type == di.DriveType)
                    {
                        filter = false;
                        break;
                    }
                }

                if (!filter)
                {
                    rList.Add(di);
                }
            }

            return rList;
        }

        public byte[] readFromDevice(string deviceName, int readStart, int readLength)
        {
            Byte[] rBytes = new Byte[readLength];

            try
            {

                string volumeString = string.Format("\\\\.\\{0}", deviceName.TrimEnd('\\'));

                FileStream fs;
                SafeFileHandle handle;

                int flags = FILE_FLAG_NO_BUFFERING | FILE_FLAG_SEQUENTIAL_SCAN;
                /* Call the Windows CreateFile() API to open the file */
                IntPtr ptr = CreateFile(volumeString,
                (int)FileAccess.Read,
                FileShare.ReadWrite,
                IntPtr.Zero,
                FileMode.Open,
                flags,
                IntPtr.Zero);

                handle = new SafeFileHandle(ptr, true);
                if (!handle.IsInvalid)
                {
                    fs = new FileStream(handle, FileAccess.Read, readLength, false);
                    UTF8Encoding temp = new UTF8Encoding(true);
                    fs.Seek(readStart, SeekOrigin.Begin);
                    fs.Read(rBytes, 0, rBytes.Length);
                    fs.Close();
                }
                else
                {
                    try
                    {
                        throw new Win32Exception(Marshal.GetLastWin32Error()); // translate number to string
                    }
                    catch (Exception ex)
                    {
                        throw new Exception("Failed to open device, reason = " + ex.Message + "\n");
                    }
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Read error, reason = " + ex.Message + "\n");
            }

            return rBytes;
        }

        public void writeToDevice(string deviceName, int writeStart, int writeLength, byte[] wBytes, bool swapBytes)
        {
            try
            {
                // swap bytes

                if (swapBytes)
                {
                    for (int i = 0; i < writeLength; i += 2)
                    {
                        byte a = wBytes[i];
                        wBytes[i] = wBytes[i + 1];
                        wBytes[i + 1] = a;
                    }
                }

                //


                string volumeString = string.Format("\\\\.\\{0}", deviceName.TrimEnd('\\'));

                FileStream fs;
                SafeFileHandle handle;

                int flags = FILE_FLAG_NO_BUFFERING | FILE_FLAG_SEQUENTIAL_SCAN;
                /* Call the Windows CreateFile() API to open the file */
                IntPtr ptr = CreateFile(volumeString,
                (int)FileAccess.Write,
                FileShare.ReadWrite,
                IntPtr.Zero,
                FileMode.Open,
                flags,
                IntPtr.Zero);


                handle = new SafeFileHandle(ptr, true);
                if (!handle.IsInvalid)
                {
                    fs = new FileStream(handle, FileAccess.Write, writeLength, false);
                    UTF8Encoding temp = new UTF8Encoding(true);
                    fs.Seek(writeStart, SeekOrigin.Begin);
                    fs.Write(wBytes, 0, wBytes.Length);
                    fs.Flush();
                    fs.Close();
                }
                else
                {
                    try
                    {
                        throw new Win32Exception(Marshal.GetLastWin32Error()); // translate number to string
                    }
                    catch (Exception ex)
                    {
                        throw new Exception("Failed to open device, reason = " + ex.Message + "\n");
                    }
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Write error (Code " + Marshal.GetLastWin32Error()  + "), reason = " + ex.Message + "\nTry to clear device first!\n");
            }
        }
    }
}
