﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ucore_assembler_gui.Properties;
using ucore_assembler_library;
using xcore_tools;
using xcore_tools.Common;
using ucore_assembler_library.Data;

namespace ucore_assembler_gui
{
    public partial class ucoreAssemblerForm : Form, IPrint
    {
        Assembler m_assembler;
        bool m_enableCloseDiaglog = true;

        public ucoreAssemblerForm()
        {
            InitializeComponent();

            // fix tab bug

            this.Load += (o, e) =>
            {
                for (int i = Tools.TabCount - 1; i >= 0; i--)
                {
                    Tools.SelectedIndex = i;
                }

            };        

            m_assembler = new Assembler(this);
            setStatus(assmblerStatus_t.ready);
        }

        public bool enableCloseDialog
        {
            get { return m_enableCloseDiaglog; }
            set { m_enableCloseDiaglog = value; }

        }

        enum assmblerStatus_t
        {
            ready,
            pass,
            fail
        }

        void setStatus(assmblerStatus_t status, List<Assembler.storeTypes_t> sTypes = null)
        {
            switch (status)
            {
                case assmblerStatus_t.ready:    assemblerStatusLabel.Text = "READY";
                                                assemblerStatusLabel.ForeColor = Color.Blue; break;
                case assmblerStatus_t.pass:
                    {
                        string aStatus = "PASS";
                        if (sTypes != null)
                        {
                            aStatus += " / store(";
                            int sTypeCount = sTypes.Count();
                            int i = 0;
                            foreach (Assembler.storeTypes_t sType in sTypes)
                            {
                                aStatus += sType.ToString().ToUpper().Remove(0,3);
                                i++;
                                if (i != sTypeCount)
                                {
                                    aStatus += ",";
                                }
                                
                            }
                            aStatus += ")";
                        }
                        assemblerStatusLabel.Text = aStatus;
                        assemblerStatusLabel.ForeColor = Color.Green;
                    }
                    break;
                case assmblerStatus_t.fail:     assemblerStatusLabel.Text = "FAIL"; 
                                                assemblerStatusLabel.ForeColor = Color.Red; break;
            }
        }

        #region FileDialogLoadSave

        private void sourceSelectButton_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();

                if (sourceFileBox.Text.Length != 0)
                {
                    openFileDialog.FileName = sourceFileBox.Text;
                }

                openFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                openFileDialog.Filter = "Assembler files (*.asm;*.s)|*.asm;*.s|All files (*.*)|*.*";
                openFileDialog.FilterIndex = 2;
                openFileDialog.RestoreDirectory = true;

                if (openFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog.FileName;
                    sourceFileBox.Text = fileName;

                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void destinationSelectButton_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();

                if (destFileBox.Text.Length != 0)
                {
                    saveFileDialog.FileName = destFileBox.Text;
                }

                saveFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                saveFileDialog.Filter = "All files (*.*)|*.*";
                saveFileDialog.FilterIndex = 1;
                saveFileDialog.RestoreDirectory = true;

                if (saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog.FileName;
                    destFileBox.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        #endregion

        private void InvokeIfRequired(Control target, Delegate methodToInvoke)
        {
            if (target.InvokeRequired)
            {
                target.Invoke(methodToInvoke);
            }
            else
            {
                methodToInvoke.DynamicInvoke();
            }
        }

        public void print(string txt)
        {
            InvokeIfRequired(this, (MethodInvoker)delegate()
            {
                debugBox.AppendText(txt);
            });
            
        }

        public void dprint(string txt)
        {
            if (checkBoxDebugPrint.Checked == true)
            {
                print(txt);
            }
        }

        private void ucoreAssemblerForm_Load(object sender, EventArgs e)
        {
            Settings.Default.Reload();
        }

        private void ucoreAssemblerForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (showCloseDialog() == DialogResult.No)
            {
                e.Cancel = true;
            }
            else
            {
                Settings.Default.Save();
            }
        }

        private DialogResult showCloseDialog()
        {
            DialogResult dr = DialogResult.Yes;
            if (enableCloseDialog)
            {
                dr = MessageBox.Show("Quit ucore assembler?", "Quit?", MessageBoxButtons.YesNo);
            }
            return dr;
        }

        private void buttonAssemble_Click(object sender, EventArgs e)
        {
            setStatus(assmblerStatus_t.fail);

            hiresCounter hc = new hiresCounter();
            hc.start();

            if (checkBoxCLS.Checked == true)
            {
                debugBox.Clear();
            }
         
            m_assembler.sourceFileName = sourceFileBox.Text;
            m_assembler.destFileName = destFileBox.Text;

            // load
            print("Load source\n");
            if (m_assembler.loadFile() == false)
            {
                print("error: load file to assemble fail!\n");
                return;
            }

            // enable hazard check if required

            m_assembler.warningLevel = (uint)numericUpDownWarningLevel.Value;
        
            // assemble

            print("Assemble\n");

            if (m_assembler.assemble() == false)
            {
                return;
            }
           

            // trace
            if (checkBoxTrace.Checked == true)
            {
                print("Trace opcodes\n");
                m_assembler.trace();
            }

            // store
            print("Store to destination\n");
            List<Assembler.storeTypes_t> m_storeTypes = new List<Assembler.storeTypes_t>();
            if (checkBoxStoreMif.Checked == true)
            {
                m_storeTypes.Add(Assembler.storeTypes_t.st_mif);
            }

            if (checkBoxStoreRaw.Checked == true)
            {
                m_storeTypes.Add(Assembler.storeTypes_t.st_raw);
            }

            if (checkBoxStoreDebug.Checked == true)
            {
                m_storeTypes.Add(Assembler.storeTypes_t.st_debug);
            }

            try
            {
                if (m_assembler.storeFile(m_storeTypes, (uint)numericUpDownStoreLength.Value) == false)
                {
                    print("error: store assembled file(s) fail!\n");
                    return;
                }
            }
            catch (Exception ex)
            {
                print("error: store assembled file(s) fail! Reason '" + ex.Message + "'\n");
                return;
            }
            

            double dataCount = (double) m_assembler.dataCount;
            double lineCount = (double)m_assembler.lineCount;
            double maxCount = (double) numericUpDownStoreLength.Value;

            print(string.Format("Use {0} of {1} words ({2:##.##}%)\n",dataCount,maxCount,100.0/maxCount*dataCount));
            hc.stop();

            double elapsedTime = hc.elapsed();

            print(string.Format("Finished to assemble {0} lines in {1:.00}s ({2:.0} ls)\n", lineCount,elapsedTime, lineCount / elapsedTime));

            setStatus(assmblerStatus_t.pass, m_storeTypes);
        }

        private void clearToolStripMenuItem_Click(object sender, EventArgs e)
        {
            debugBox.Clear();
        }
    

        private void exitToolStripMenuItem_Click_1(object sender, EventArgs e)
        {
            Close();
        }

        private void buttonPrintOpcodes_Click(object sender, EventArgs e)
        {
            Opcode o = new Opcode(this);
            List<Opcode.opcodeDefinition> opcodeDefinitions = o.getOpcodes();

            foreach (Opcode.opcodeDefinition oD in opcodeDefinitions)
            {
                print(String.Format("{0}\n",oD.name));
            }
        }

        
    }
}
