﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_assembler_library
{
    public enum label_TypeEnum
    {
        OPCODE,
        UNKNOWN
    }

    [Serializable]
    class Label
    {
        public Label(string iname,uint iaddress, label_TypeEnum itype)
        {
            address = iaddress;
            type = itype;
            name = iname;
        }

        public Label()
        {
            address = 0;
            type = label_TypeEnum.UNKNOWN;
            name = "";
        }

        public string name
        {
            get { return m_name; }
            set { m_name = value; }
        }

        public uint address
        {
            get { return m_address; }
            set { m_address = value; }
        }

        public label_TypeEnum type
        {
            get { return m_type; }
            set { m_type = value; }
        }

        private string m_name;
        private uint m_address;
        private label_TypeEnum m_type;
    }

    [Serializable]
    public class Labels
    {
        public Labels()
        {

        }

        // insert

        public bool insert(string name, uint address)
        {
            return insert(name, address, label_TypeEnum.OPCODE);
        }

        public bool insert(string name, uint address, label_TypeEnum type)
        {
            if (existsName(name))
            {
                return false;
            }

            m_labels.Add(new Label(name, address, type));
            m_labelNames.Add(name);

            return true;
        }

        // search

        public bool search(string name, out uint address, out label_TypeEnum type)
        {
            address = 0;
            type = label_TypeEnum.UNKNOWN;

            if (!existsName(name))
            {
                return false;
            }

            foreach (Label d in m_labels)
            {
                if (d.name == name)
                {
                    address = d.address;
                    type = d.type;

                    return true;
                }
            }

            return false;
        }

        public bool search(uint address, label_TypeEnum type, ref string name)
        {
            foreach (Label d in m_labels)
            {
                if (d.address == address && d.type == type)
                {
                    name = d.name;
                    return true;
                }
            }

            return false;
        }

        public bool search(string name)
        {
            return existsName(name);
        }

        //

        public void clear()
        {
            m_labelNames.Clear();
            m_labels.Clear();
        }

        public List<string> getLabels()
        {
            return m_labelNames;
        }

        //

        public int count()
        {
            return m_labels.Count;
        }

        private bool existsName(string name)
        {
            foreach (string s in m_labelNames)
            {
                if (s == name)
                {
                    return true;
                }
            }

            return false;
        }


        private List<Label> m_labels = new List<Label>();
        private List<string> m_labelNames = new List<string>();



    }
}
