﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO.Compression;

namespace ucore_assembler_library
{

    [Serializable] public class DebugData 
    {
        public DebugData(string iinfo, List<IOpcode<UInt16>> iopcodes, Labels ilabels)
        {
            m_info = iinfo;
            m_opcodes = iopcodes;
            m_labels = ilabels;
        }

        public DebugData()
        {
            m_info = null;
            m_opcodes = null;
            m_labels = null;
        }

        public Labels labels
        {
            get { return m_labels;  }
        }

        public string info
        {
            get { return m_info;  }
        }

        public List<IOpcode<UInt16>> opcodes
        {
            get { return m_opcodes; }
        }

        List<IOpcode<UInt16>> m_opcodes;
        Labels m_labels;
        string m_info;
    };

    public class SaveDebug : ISave
    {
        public string name()
        {
            return ("debug");
        }

        public bool supportLength()
        {
            return false;
        }


        public bool save(string fileName, List<IOpcode<UInt16>> opcodes, uint length)
        {
          
            string processFileName = fileName;

            if (processFileName.EndsWith(".debug", StringComparison.CurrentCultureIgnoreCase) == false)
            {
                processFileName += ".debug";
            }
                
            BinaryFormatter formater = new BinaryFormatter();

            Labels labels = (opcodes.Count > 0)?opcodes[0].labelDictonary:null;
            DebugData debugData = new DebugData("ucore Assembler debug data version 1.0", opcodes, labels);

            using (FileStream fs = File.Open(processFileName, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                formater.Serialize(fs, debugData);
                fs.Close();
            }
           
            return true;
        }

        private bool m_enable = false;

        public bool enable
        {
            get
            {
                return m_enable;
            }
            set
            {
                m_enable = value;
            }
        }
    }
}
