﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Threading.Tasks;

namespace ucore_assembler_library
{
    public class SaveMif : ISave
    {
        public string name()
        {
            return ("mif");
        }

        public bool supportLength()
        {
            return true;
        }

        public bool save(string fileName, List<IOpcode<UInt16>> opcodes, uint length)
        {
          
            string processFileName = fileName;

            if (processFileName.EndsWith(".mif", StringComparison.CurrentCultureIgnoreCase) == false)
            {
                processFileName += ".mif";
            }

            using (FileStream fs = File.Open(processFileName, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                StringBuilder outString = new StringBuilder();

                // generate head

                string mifHead = string.Format("WIDTH=16;\n"
                                    + "DEPTH={0};\n"
                                    + "ADDRESS_RADIX=UNS;\n"
                                    + "DATA_RADIX=HEX;\n"
                                    + "CONTENT BEGIN\n", length);

                outString.Append(mifHead);

                // generate data

                if (opcodes.Count > length)
                {
                    throw new Exception("Error: Listing is greater as mif address count");
                }

                SortedDictionary<uint, string> wData = new SortedDictionary<uint, string>();                
                foreach (IOpcode<UInt16> o in opcodes)
                {
                    UInt16 value = (UInt16)o.generate();
                    uint pc = o.pc;

                    if (o.isBinary)
                    {
                        for (int i = 0; i < o.lengthBinary; i ++)
                        {
                            wData[(uint)(pc + i)] = string.Format("    {0:D8} : {1:X4};-- binary\n", (uint)(pc + i ), o.dataBinary[i]);
                        }
                    }
                    else
                    {
                        wData[pc] = string.Format("    {0:D8} : {1:X4};-- {2}\n", pc, value, o.trace());
                    }
                }
                

                for (uint adr = 0; adr < length; adr++)
                {
                    if (wData.ContainsKey(adr) == true)
                    {
                        outString.Append(wData[adr]);
                    }
                    else
                    {
                        outString.Append(string.Format("    {0:D8} : {1:X4};-- {2}\n", adr, 0xffff, "empty"));
                    }
                }
       
               outString.Append("END;\n");

                // write data (shit convert ....)

                char []dataRaw = outString.ToString().ToCharArray();
                byte[] bytesRaw = new byte[dataRaw.Length];

                for (int i = 0; i < dataRaw.Length; i++)
                {
                    bytesRaw[i] = (byte) dataRaw[i];
                }

                fs.Write(bytesRaw, 0, dataRaw.Length);
                fs.Close();
            }
          
            return true;
        }

        private bool m_enable = false;

        public bool enable
        {
            get
            {
                return m_enable;
            }
            set
            {
                m_enable = value;
            }
        }
    }
};