﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using xcore_tools;

namespace ucore_assembler_library
{
    public class SaveRaw : ISave, IPrint
    {
        IPrint m_iprint = null;

        public SaveRaw(IPrint iprint)
        {
            m_iprint = iprint;
        }

        public string name()
        {
            return ("raw");
        }

        public bool supportLength()
        {
            return false;
        }

        public bool save(string fileName, List<IOpcode<UInt16>> opcodes, uint length)
        {
            string processFileName = fileName;

            if (processFileName.EndsWith(".raw", StringComparison.CurrentCultureIgnoreCase) == false)
            {
                processFileName += ".raw";
            }

            using (FileStream fs = File.Open(processFileName, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                SortedDictionary<uint, IOpcode<UInt16>> wData = new SortedDictionary<uint, IOpcode<UInt16>>();

                foreach (IOpcode<UInt16> o in opcodes)
                {
                    UInt16 value = (UInt16)o.generate();
                    uint pc = o.pc;

                    wData[pc] = o;
                }

                for (uint adr = 0; adr < length; adr++)
                {
                    UInt16 value = 0xffff;  // empty

                    if (wData.ContainsKey(adr) == true)
                    {
                        IOpcode<UInt16> o = wData[adr];

                        if (o.isBinary)
                        {
                            uint dataLength = o.lengthBinary>length?length:o.lengthBinary;
 
                            dprint(String.Format("SaveRaw: store binary from {0} length {1} words\n",adr,dataLength));

                            for (int i = 0; i < dataLength; i++)
                            {
                                fs.WriteByte((byte)(o.dataBinary[i] >> 8));
                                fs.WriteByte((byte)(o.dataBinary[i]));
                            }
                           
                            adr += dataLength-1;
                        }
                        else 
                        {
                            value = o.generate();

                            fs.WriteByte(Convert.ToByte(value >> 8));
                            fs.WriteByte(Convert.ToByte(value & 0xff));
                        }
                       
                    }
                    else
                    {
                        fs.WriteByte(Convert.ToByte(value >> 8));
                        fs.WriteByte(Convert.ToByte(value & 0xff));
                    }
                }

                fs.Close();
            }
                    
            return true;
        }

        private bool m_enable = false;

        public bool enable
        {
            get
            {
                return m_enable;
            }
            set
            {
                m_enable = value;
            }
        }

        public void print(string txt)
        {
            if (m_iprint != null)
            {
                m_iprint.print(txt);
            }
        }

        public void dprint(string txt)
        {
            if (m_iprint != null)
            {
                m_iprint.dprint(txt);
            }
        }
    }
}
