﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using ucore_emulator.core;
using ucore_emulator.core.components.memory.ctrl;
using ucore_emulator.common;
using System.Threading;
using ucore_emulator.Properties;
using ucore_emulator.core.components;
using ucore_emulator.Load;
using System.Drawing.Drawing2D;
using ucore_emulator.core.components.video;
using System.Drawing.Imaging;
using System.Threading.Tasks;
using ucore_emulator.core.components.extMemory;
using ucore_emulator.core.components.memory;
using ucore_assembler_library;
using ucore_assembler_gui;
using Ionic.Zip;
using System.IO;
using System.Xml;
using GfxConvert;
using RAWEdit;
using xcore_tools;
using System.Reflection;

namespace ucore_emulator
{
    public partial class EmulatorForm : Form, IPrint, IVideoEvent
    {
        //

        emulator m_emulator = null;
        hiresCounter m_hiresCounter = new hiresCounter();
        ScanCodeTranslator m_scanCodeTranslator = new ScanCodeTranslator();


        const int m_perfInf_ucore = 0;
        const int m_perfInf_hcores = 1;
        performaceInformations []m_performaceInformations = new performaceInformations[2];

        long m_lastTime = 0;
        UInt64 m_lastUCoreClkCount = 0;
        UInt64 m_lastHCoresClkCount = 0;
        UInt64 m_lastFrameCount = 0;
        UInt64 m_frameCount = 0;
        UInt32 m_frameCheckDelay = 0;

        // ext

        ucoreAssemblerForm m_ucoreAssemblerForm = null;
        GfxConvertForm m_gfxConnvertForm = null;
        RAWEditMainForm m_RawEditForm = null;

        // gfx


        Object _lockResize = new Object();
        uint m_resizeCount = 0;
        Object _lockRepaint = new Object();
        Object _vBlankLock = new Object();
        Bitmap m_screenBitmap = null;
        Rectangle m_screenBufferRectangle = new Rectangle(0, 0, 0, 0);

        //

        public EmulatorForm()
        {
            for (int i = 0; i < m_performaceInformations.Count(); i++)
            {
                m_performaceInformations[i] = new performaceInformations();
            }

            InitializeComponent();

            // version to title

            this.Text += getBuildVersion();

            // fix tab bug

            this.Load += (o, e) =>
            {
                for (int i = tabControl1.TabCount-1; i >= 0; i--)
                {
                    tabControl1.SelectedIndex = i;
                }

            };           

            //

            m_emulator = new emulator(this, this, m_scanCodeTranslator);
            m_emulator.ucore.bus.ctrlMemory.identifyAsEmulator = checkBoxCheckIdentifyAsEmulator.Checked;
            m_emulator.ucore.ext_bus.sdCard.sdCard_emulation.setFile(textBoxSDRawFile.Text);
            m_emulator.ucore.bus.ctrlMemory.audio.enableRealAudio = checkBoxEnableReadAudio.Checked;

            numericUpDownHCoreCount.Value = m_emulator.hcoreCount;

            // hcore

            int processorCount = Environment.ProcessorCount;
            processorCount-= 3; // - ucore -video -gui
            if(processorCount < 1)
            {
                processorCount = 1;
            }
            checkBoxHCoreMultithread.Text = "use multithread for HCore (recommended max. " + processorCount + " Threads on your System)";

            // 
        }

        private static string getBuildVersion()
        {
            string versionString = "";
            try
            {
                object[] asAttrs = Assembly.GetExecutingAssembly().GetCustomAttributes(true);
                foreach (object asAttr in asAttrs)
                {
                    if (asAttr is AssemblyFileVersionAttribute)
                    {
                        AssemblyFileVersionAttribute vAttr = asAttr as AssemblyFileVersionAttribute;
                        Version fileVersion = new Version(vAttr.Version);
                        versionString = string.Format(" V1.{0}.{1:00}.{2:00}", fileVersion.Major - 13, fileVersion.Minor, fileVersion.Build);
                    }
                }
            }
            catch (Exception)
            { /* not found so ignore version */
                versionString = "";
            }
            return versionString;
        }

        private void ctrlDisplayTimer_Tick(object sender, EventArgs e)
        {
            if (m_resizeCount > 0)
            {
                m_resizeCount--;
                videoBox.Refresh();
            }
            maskedTextBoxSDBlock.Text = String.Format("{0:00000000}",m_emulator.ucore.ext_bus.sdCard.sdCard_emulation.currentReadBlock);
            pictureBoxCtrlDisplay.Refresh();

            InvokeIfRequired(this, (MethodInvoker)delegate()
            {                
                debugBox.AppendText(m_debugText.ToString());
                m_debugText.Clear();
            });

             
        }

        #region board_ctrl_paint
        private void pictureBoxCtrlDisplay_Paint(object sender, PaintEventArgs e)
        {
            BoardCtrl boardCtrl = m_emulator.boardCtrl;

            Graphics g = e.Graphics;

            // hex 

            Pen p = new Pen(Color.DarkRed);
            p.Width = 2;
            Pen d = new Pen(Color.LightGray);
            d.Width = 1;

            int xo;
            int yo;

            xo = 0;
            yo = 4;

            /*   --     0
                 --    ###
                 --  5 # # 1
                 --    #6#
                 --  4 # # 2
                 --    ###
                 --     3
             */

            int hex_w = 8;
            int hex_h = 8;
            uint hexCount = boardCtrl.hexCount;
            for (uint h = 0; h < hexCount; h++)
            {
                xo = (int)((hexCount - 1 - h) * (hex_w + 4)) + 4;

                g.DrawLine(boardCtrl.get_hex((uint)h, 0) ? p : d, xo, yo, xo + hex_w, yo); // -0
                g.DrawLine(boardCtrl.get_hex((uint)h, 5) ? p : d, xo, yo, xo, yo + hex_h); // |5
                g.DrawLine(boardCtrl.get_hex((uint)h, 1) ? p : d, xo + hex_w, yo, xo + hex_w, yo + hex_h); // |1
                g.DrawLine(boardCtrl.get_hex((uint)h, 6) ? p : d, xo, yo + hex_h, xo + hex_w, yo + hex_h); // -6
                g.DrawLine(boardCtrl.get_hex((uint)h, 4) ? p : d, xo, yo + hex_h, xo, yo + hex_h + hex_h); // |4
                g.DrawLine(boardCtrl.get_hex((uint)h, 2) ? p : d, xo + hex_w, yo + hex_h, xo + hex_w, yo + hex_h + hex_h); // |2
                g.DrawLine(boardCtrl.get_hex((uint)h, 3) ? p : d, xo, yo + hex_h + hex_h, xo + hex_w, yo + hex_h + hex_h); // -3
            }

           Font ledFont = new Font("courier", 5);
           yo = 21;

           uint ledXOffset = 14;
           uint rledCount = boardCtrl.rledCount;
           int led_w = 10;
           for (uint h = 0; h < rledCount; h++)
            {
                xo = (int)((h * led_w) + ledXOffset);
                g.FillEllipse(boardCtrl.get_rled(rledCount - h - 1) ? Brushes.DarkRed : Brushes.LightCoral, xo, yo + 4, led_w, led_w);
               g.DrawString("" + (rledCount - h - 1), ledFont, Brushes.Black, xo + (h>=8?2:1), yo+6);   
            }

           yo = 4;
           uint gledCount = boardCtrl.gledCount;
           for (uint h = 0; h < gledCount; h++)
            {
                xo = (int)((h * led_w) + (led_w * (rledCount - 9)) + ledXOffset);
                g.FillEllipse(boardCtrl.get_gled(gledCount - h - 1) ? Brushes.DarkGreen : Brushes.LightGreen, xo, yo + 4, led_w, led_w);
                g.DrawString("" + (gledCount - h - 1), ledFont, Brushes.Black, xo + 3, yo+6);               
            }

            // draw KEYBOARD LED

            string[] keyBoardLeds = new string[3];

            keyBoardLeds[0] = "CAPS";
            keyBoardLeds[1] = "SCRL";
            keyBoardLeds[2] = "NUM";

            yo = 4;
            int keyBoardLedsCount = keyBoardLeds.Count();
            for (int h = 0; h < keyBoardLedsCount; h++)
            {
                xo = (int)((rledCount * led_w)) + 20;

                Brush keyBoardTextBrush;

                if (boardCtrl.getKeyBoardStatus(BoardCtrl.keyBoardStatusToEnum(keyBoardLeds[h])))
                {
                    keyBoardTextBrush = Brushes.DarkBlue;
                }
                else
                {
                    keyBoardTextBrush = Brushes.LightGray;
                }

                g.DrawString(keyBoardLeds[h], ledFont, keyBoardTextBrush, xo, yo+6+(h *8));
            }

            // SD CMD

          /*  yo = 13;
            xo = (int)((keyBoardLedsCount * 24) + (hexCount * (hex_w + 4)) + (gledCount * led_w)) + 80;

            UInt16 lastSDCmd = m_emulator.ucore.ext_bus.sdCard.sdCard_emulation.reciveCmd.cmd;

            g.DrawString(string.Format("SD CMD{0}",lastSDCmd),  new Font("courier", 8), Brushes.DarkSlateBlue, xo, yo);
           */
            
           // time calculate

           UInt64 ucoreClkCount = m_emulator.clkCountUCore;
           UInt64 hcoresClkCount = m_emulator.clkCountHCores;

           long current = m_hiresCounter.current();
           double elapsed = m_hiresCounter.calculate(current, m_lastTime);

           // calc mips

           m_performaceInformations[m_perfInf_ucore].currentMips = (1.0 / elapsed) * (double)(ucoreClkCount - m_lastUCoreClkCount) / 1E+6;
           m_performaceInformations[m_perfInf_hcores].currentMips = (1.0 / elapsed) * (double)(hcoresClkCount - m_lastHCoresClkCount) / 1E+6;
         
           for (int i = 0; i < m_performaceInformations.Count(); i++)
           {
               if (m_performaceInformations[i].maxMips < m_performaceInformations[i].currentMips)
               {
                   m_performaceInformations[i].maxMips = m_performaceInformations[i].currentMips;
               }
           }

           m_lastUCoreClkCount = ucoreClkCount;
           m_lastHCoresClkCount = hcoresClkCount;

           m_lastTime = current;

           
           if (m_frameCheckDelay > 10)
           {
               m_performaceInformations[m_perfInf_ucore].fps = 0.1 / elapsed * (((double)m_frameCount) - (((double)m_lastFrameCount)));
               m_lastFrameCount = m_frameCount;
               m_frameCheckDelay = 0;
           }
           m_frameCheckDelay++;

           uint hcoreNotIdleCounter = m_emulator.hcoreCount;
           for (uint i = 0; i < m_emulator.hcoreCount; i++)
           {
               if (m_emulator.getHCore(i).idle)
               {
                   hcoreNotIdleCounter--;
               }
           }

           uint hcoreCount = m_emulator.hcoreCount;
     
           toolStripStatusLabelPerformance.Text = string.Format("Video {0:00.00} FPS; UCore/HCores(C" + hcoreCount +";W"+ hcoreNotIdleCounter+") {1:00.00}/{2:00.00} MIPS (MAX: {3:00.00}/{4:00.00}) {5}/{6}% of native speed; Audio max freq. {7:#####0} Hz",
                 m_performaceInformations[m_perfInf_ucore].fps,
               m_performaceInformations[m_perfInf_ucore].currentMips, m_performaceInformations[m_perfInf_hcores].currentMips,
               m_performaceInformations[m_perfInf_ucore].maxMips, m_performaceInformations[m_perfInf_hcores].maxMips,
              (int)((m_performaceInformations[m_perfInf_ucore].currentMips / 106.48) * 100.0),
              (int)((m_performaceInformations[m_perfInf_hcores].currentMips / 100.0) * 100.0),
              m_emulator.ucore.bus.ctrlMemory.audio.currentMaxFrequency);

        }
    #endregion

        #region menu_items
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Close();
        }
        #endregion

        private void EmulatorForm_FormClosing(object sender, FormClosingEventArgs e)
        {

            if (MessageBox.Show(this, "Quit emulator?", "Quit?", MessageBoxButtons.YesNo) == DialogResult.No)
            {
                e.Cancel = true;
            }
            else
            {
                Settings.Default.Save();

                try{    m_emulator.stop();  }catch (Exception){}

                Thread.Sleep(1000);

                try{    m_emulator.kill();  }catch (Exception){}
               
            }

        }

        #region toolstrip_events
        private void toolStripButtonReset_Click(object sender, EventArgs e)
        {
            m_emulator.reset();
            m_scanCodeTranslator.reset();
        }

        private void toolStripButtonRun_Click(object sender, EventArgs e)
        {
            try
            {
                bool multiThreadingHCores = checkBoxHCoreMultithread.Checked;
                m_emulator.m_hcoreSingleThread = !multiThreadingHCores;
                bool disableHCores = checkBoxDisableHcores.Checked;
                m_emulator.start(disableHCores);
            }
            catch (Exception ex)
            {
                print(ex.Message);
            }
        }

        private void toolStripButtonStep_Click(object sender, EventArgs e)
        {
            try
            {
                m_emulator.step();
            }
            catch (Exception ex)
            {
                print(ex.Message);
            }
        }

        private void toolStripButtonStop_Click(object sender, EventArgs e)
        {
            try
            {
                m_emulator.stop();
            }
            catch (Exception ex)
            {
                print(ex.Message);
            }
        }
        #endregion

        StringBuilder m_debugText = new StringBuilder();

        public void print(string txt)
        {
            InvokeIfRequired(this, (MethodInvoker)delegate()
               {
                   m_debugText.Append(txt);                  
               });
        }

        public void dprint(string txt)
        {
            if (checkBoxDebugPrint.Checked == true)
            {
                print(txt);
            }
        }

        private void buttonSelectUCoreMem_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog1 = new OpenFileDialog();


                if (textBoxUCoreMemPath.Text != "")
                {
                    openFileDialog1.FileName = textBoxUCoreMemPath.Text;
                    openFileDialog1.InitialDirectory = Path.GetDirectoryName(textBoxUCoreMemPath.Text);
                }
                else
                {
                    openFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog1.Filter = "All files (*.*)|*.*";
                openFileDialog1.FilterIndex = 2;
                openFileDialog1.RestoreDirectory = true;

                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog1.FileName;
                    textBoxUCoreMemPath.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void EmulatorForm_Load(object sender, EventArgs e)
        {
            Settings.Default.Reload();
        }

        private void buttonLoadUCoreMem_Click(object sender, EventArgs e)
        {
            loadUCoreRaw(textBoxUCoreMemPath.Text);

        }

        private void loadUCoreRaw(string fileName)
        {
            ucore_bus bus = m_emulator.ucore.bus;

            int length = 0;
            try
            {
                LoadRAW loadRaw = new LoadRAW();
                loadRaw.load(this, fileName, bus, out length,true);
            }
            catch (Exception ex)
            {
                print("Error at load '" + ex.Message + "'\n");
            }
        }

        private void toolStripButtonUcoreView_Click(object sender, EventArgs e)
        {
            m_emulator.showUcoreForm();
        }

        private void videoBox_Paint(object sender, PaintEventArgs e)
        {
            lock (_lockRepaint)
            {
                Graphics g = e.Graphics;

                if (!checkBoxBlurGfx.Checked)
                {
                    g.PixelOffsetMode = PixelOffsetMode.Half;
                    g.SmoothingMode = SmoothingMode.None;
                    g.InterpolationMode = InterpolationMode.NearestNeighbor;
                }

                if(m_screenBitmap != null)
                {
                    g.DrawImage(m_screenBitmap, 0, 0, videoBox.Width, videoBox.Height);
                }

                if (m_resizeCount > 0)
                {
                    Pen aPen = new Pen(Brushes.Blue);
                    g.DrawLine(aPen, 0, 0, videoBox.Width, videoBox.Height);
                    g.DrawLine(aPen, videoBox.Width, 0, 0, videoBox.Height);

                    Font eFont = new Font("courier", 16,FontStyle.Bold);

                    int ratio =  videoBox.Height * 16 / videoBox.Width;

                    Brush bBrush = ratio==10?Brushes.LightGreen:Brushes.Yellow;
                    String infoMessage = String.Format("{0}x{1} [16x{2}]", videoBox.Width, videoBox.Height, ratio);

                    g.DrawString(infoMessage, eFont, bBrush, (videoBox.Width) / 2 - 100, (videoBox.Height) / 2 - 16);
                }
            }
        }

        private void InvokeIfRequired(Control target, Delegate methodToInvoke)
        {
            try
            {
                if (target.InvokeRequired)
                {
                    target.Invoke(methodToInvoke);
                }
                else
                {
                    methodToInvoke.DynamicInvoke();
                }
            }
            catch (Exception)
            { 
            }
        }

        int m_lastvisibleWidth = -1;
        int m_lastvisibleHeight = -1;

        unsafe public void verticalBlank()
        {
            lock (_vBlankLock)
            {              
               // m_screenBufferRectangle.Width = m_emulator.videoComponent.screenDefinitons.width;
              //  m_screenBufferRectangle.Height = m_emulator.videoComponent.screenDefinitons.height;



                UInt32[] screenBuffer = m_emulator.videoComponent.getScreenBuffer();

                int visibleWidth = m_emulator.videoComponent.visibleWidth();
                int visibleHeight = m_emulator.videoComponent.visibleHeight();
                int copyLength = visibleWidth * visibleHeight;

                if (visibleWidth != m_lastvisibleWidth || m_lastvisibleHeight != visibleHeight)
                {
                    m_screenBitmap = new Bitmap(visibleWidth, visibleHeight);
                }

                m_lastvisibleWidth = visibleWidth;
                m_lastvisibleHeight = visibleHeight;

                
                m_screenBufferRectangle.Width = visibleWidth;
                m_screenBufferRectangle.Height = visibleHeight;


                InvokeIfRequired(this, (MethodInvoker)delegate()
                {
                    BitmapData bitmapData = m_screenBitmap.LockBits(m_screenBufferRectangle, ImageLockMode.WriteOnly, PixelFormat.Format32bppRgb);
                    UInt32* bmData = (UInt32*)bitmapData.Scan0.ToPointer();

                    fixed (UInt32* bmSource = &screenBuffer[0])
                    {
                        for (int i = 0, length = copyLength; i < length; )
                        {
                            *bmData++ = bmSource[i++]; *bmData++ = bmSource[i++];
                            *bmData++ = bmSource[i++]; *bmData++ = bmSource[i++];
                            *bmData++ = bmSource[i++]; *bmData++ = bmSource[i++];
                            *bmData++ = bmSource[i++]; *bmData++ = bmSource[i++];
                        }
                    }

                    // time calculate

                    m_screenBitmap.UnlockBits(bitmapData);
                    m_frameCount++;

                    videoBox.Refresh();  // display

                });

            }
        }

       /* [obsolete] private void trackBar1_ValueChanged(object sender, EventArgs e)
        {
            m_emulator.videoComponent.vblank_v = (int)trackBar1.Value;
        }

        private void trackBar2_ValueChanged(object sender, EventArgs e)
        {
            m_emulator.videoComponent.vblank_h = (int)trackBar2.Value;
        }*/ 

        private void buttonSelectAllInOnePackage_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog1 = new OpenFileDialog();


                if (textBoxAllInOnePackage.Text != "")
                {
                    openFileDialog1.FileName = textBoxAllInOnePackage.Text;
                    openFileDialog1.InitialDirectory = Path.GetDirectoryName(textBoxAllInOnePackage.Text);
                }
                else
                {
                    openFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog1.Filter = "All files (*.*)|*.*";
                openFileDialog1.FilterIndex = 2;
                openFileDialog1.RestoreDirectory = true;

                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog1.FileName;
                    textBoxAllInOnePackage.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }    
        }  

        private void buttonSelectFlash_Click(object sender, EventArgs e)
        {            
            try
            {
                OpenFileDialog openFileDialog1 = new OpenFileDialog();


                if (textBoxFlashMemoryPath.Text != "")
                {
                    openFileDialog1.FileName = textBoxFlashMemoryPath.Text;
                    openFileDialog1.InitialDirectory = Path.GetDirectoryName(textBoxFlashMemoryPath.Text);
                }
                else
                {
                    openFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog1.Filter = "All files (*.*)|*.*";
                openFileDialog1.FilterIndex = 2;
                openFileDialog1.RestoreDirectory = true;

                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog1.FileName;
                    textBoxFlashMemoryPath.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }    
        }

        private void buttonLoadFlash_Click(object sender, EventArgs e)
        {
            loadFlash(textBoxFlashMemoryPath.Text);
        }

        private void loadFlash(string fileName)
        {
            ext_bus bus = m_emulator.ucore.ext_bus;

            int length = 0;
            try
            {
                LoadRAW loadRaw = new LoadRAW();

                bus.flash.writeProtection = false;  // disable write protection for load
                loadRaw.load(this, fileName , bus, out length, 0xe0000000);
                bus.flash.writeProtection = true;
            }
            catch (Exception ex)
            {
                print("Error at load '" + ex.Message + "'\n");
            }
        }

        #region IRAM_ASSEMBLER

        private void iramAssemblerButtonSourceFileSelect_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();

                if (iramTextBoxUcore1SourceFile.Text.Length != 0)
                {
                    openFileDialog.FileName = iramTextBoxUcore1SourceFile.Text;
                    openFileDialog.InitialDirectory = Path.GetDirectoryName(iramTextBoxUcore1SourceFile.Text);
                }
                else
                {
                    openFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog.Filter = "Assembler files (*.asm;*.s)|*.asm;*.s|All files (*.*)|*.*";
                openFileDialog.FilterIndex = 2;
                openFileDialog.RestoreDirectory = true;

                if (openFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog.FileName;
                    iramTextBoxUcore1SourceFile.Text = fileName;

                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void iramAssemblerDestinationSelectButton_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();

                if (iramAssemblerDestFileBox.Text.Length != 0)
                {
                    saveFileDialog.FileName = iramAssemblerDestFileBox.Text;
                    saveFileDialog.InitialDirectory = Path.GetDirectoryName(iramAssemblerDestFileBox.Text);
                }
                else
                {
                    saveFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                saveFileDialog.Filter = "All files (*.*)|*.*";
                saveFileDialog.FilterIndex = 1;
                saveFileDialog.RestoreDirectory = true;

                if (saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog.FileName;
                    iramAssemblerDestFileBox.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void iramAssemblerButtonAssembleAndLoad_Click(object sender, EventArgs e)
        {
            List<Assembler.storeTypes_t> storeTypes = new List<Assembler.storeTypes_t>();
            if (iramCheckBoxStoreMif.Checked == true)
            {
                storeTypes.Add(Assembler.storeTypes_t.st_mif);
            }

            if (iramCheckBoxStoreRaw.Checked == true)
            {
                storeTypes.Add(Assembler.storeTypes_t.st_raw);
            }

            if (iramCheckBoxStoreDebug.Checked == true)
            {
                storeTypes.Add(Assembler.storeTypes_t.st_debug);
            }
              
            try
            {
                assemble(   
                            "ucore",
                            iramTextBoxUcore1SourceFile.Text, iramAssemblerDestFileBox.Text,
                            (uint)numericUpDownIRAMAssemblerWarningLevel.Value,                          
                            iramCheckBoxTrace.Checked,iramCheckBoxTraceFileStats.Checked,false,
                            (uint)iramNumericUpDownStoreLength.Value, true, true,
                            storeTypes
                        );

             }
            catch (Exception ex)
            {
                print("error: '" + ex.Message + "'\n");
                return;
            }   
        }

        #endregion

        #region EXT_SDCARD_ASSEMBLER
        private void sdButtonUcoreAssembleFileSelect_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();

                if (sdTextBoxUcore1SourceFile.Text.Length != 0)
                {
                    openFileDialog.FileName = sdTextBoxUcore1SourceFile.Text;
                    openFileDialog.InitialDirectory = Path.GetDirectoryName(sdTextBoxUcore1SourceFile.Text);
                }
                else
                {
                    openFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog.Filter = "Assembler files (*.asm;*.s)|*.asm;*.s|All files (*.*)|*.*";
                openFileDialog.FilterIndex = 2;
                openFileDialog.RestoreDirectory = true;

                if (openFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog.FileName;
                    sdTextBoxUcore1SourceFile.Text = fileName;

                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void sdDestinationSelectButton_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog saveFileDialog = new SaveFileDialog();

                if (sdAssemblerDestFileBox.Text.Length != 0)
                {
                    saveFileDialog.FileName = sdAssemblerDestFileBox.Text;
                    saveFileDialog.InitialDirectory = Path.GetDirectoryName(sdAssemblerDestFileBox.Text);
                }
                else
                {
                    saveFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                saveFileDialog.Filter = "All files (*.*)|*.*";
                saveFileDialog.FilterIndex = 1;
                saveFileDialog.RestoreDirectory = true;

                if (saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog.FileName;
                    sdAssemblerDestFileBox.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void sdButtonAssembleAndLoad_Click(object sender, EventArgs e)
        {
            List<Assembler.storeTypes_t> storeTypes = new List<Assembler.storeTypes_t>();         
            if (sdCheckBoxStoreRaw.Checked == true)
            {
                storeTypes.Add(Assembler.storeTypes_t.st_raw);
            }

            if (sdCheckBoxStoreDebug.Checked == true)
            {
                storeTypes.Add(Assembler.storeTypes_t.st_debug);
            }

            if (sdCheckBoxStoreMif.Checked == true)
            {
                storeTypes.Add(Assembler.storeTypes_t.st_mif);
            }

            try
            {
                assemble(
                            "ucore",
                            sdTextBoxUcore1SourceFile.Text, sdAssemblerDestFileBox.Text,
                            (uint)numericUpDownWarningLevelSDAssembler.Value,
                            sdCheckBoxTrace.Checked,sdCheckBoxTraceFileStats.Checked,sdCheckBoxHCoreTrace.Checked,
                            0, false, false,
                            storeTypes
                        );

            }
            catch (Exception ex)
            {
                print("error: '" + ex.Message + "'\n");
                return;
            }   
        }
        #endregion

        private void assemble(  
                                string destinationType,
                                string sourceFileName,string destFileName, 
                                uint warningLevel,bool enableTrace,bool enableFileTrace,bool enableHCoreTrace,
                                uint maxLength,bool maxLengthEnabled,
                                bool copyDataToIram,
                                List<Assembler.storeTypes_t> fileStoreTypes
                             )
        {
            Assembler assembler = new Assembler(this);

            assembler.sourceFileName = sourceFileName;
            assembler.destFileName = destFileName;

            // set warning level / internal trace

            assembler.warningLevel = warningLevel;
            assembler.hcoreTrace = enableHCoreTrace;

            hiresCounter hc = new hiresCounter();
            hc.start();

            // load
            print("Load " + destinationType + " source\n");
            if (assembler.loadFile() == false)
            {
                print("error: load file to assemble fail!\n");
                return;
            }

            // assemble

            print("Start UCore assembler @" + DateTime.Now.ToString("HH:mm:ss") + "\n");
            if (assembler.assemble() == false)
            {
                return;
            }

            // trace
            if (enableTrace == true)
            {
                print("Trace opcodes\n");
                assembler.trace();
            }

            if (enableFileTrace == true)
            {
                assembler.traceIncludeStats();
            }
           
            double dataCount = (double)assembler.dataCount;
            double lineCount = (double)assembler.lineCount;
            double maxCount = (double)maxLength;

            if (maxLengthEnabled)
            {
                print(string.Format("Use {0} of {1} words ({2:##.##}%)\n", dataCount, maxCount, 100.0 / maxCount * dataCount));
            }
            else 
            {
                print(string.Format("Generate {0} word\n", dataCount));
            }
            
            hc.stop();

            double elapsedTime = hc.elapsed();

            print(string.Format("Finished to assemble {0} lines in {1:.00}s ({2:.0} ls)\n", lineCount, elapsedTime, lineCount / elapsedTime));

            uint storeLength = maxLengthEnabled ? maxLength : (uint)assembler.lastAddress;  

            if (copyDataToIram)
            {
                print("Copy data to iram\n");

                UInt16[] rawData = assembler.getRawData(storeLength); // get all
                ucore_bus bus = m_emulator.ucore.bus;

                for (uint i = 0; i < storeLength; i++)
                {
                    bus.write(i, rawData[i]);
                }
            }

            // store
       
            if (fileStoreTypes.Count != 0)
            {
                print("Store to destination\n");

                try
                {
                    if (assembler.storeFile(fileStoreTypes, storeLength) == false)
                    {
                        print("error: store assembled file(s) fail!\n");
                        return;
                    }
                }
                catch (Exception ex)
                {
                    print("error: store assembled file(s) fail! Reason '" + ex.Message + "'\n");
                    return;
                }
            }
        }

        private void clearToolStripMenuItem_Click(object sender, EventArgs e)
        {
            debugBox.Clear();
        }

        private void toolStripButtonAssemblerForm_Click(object sender, EventArgs e)
        {
            bool createNew = false;
            if (m_ucoreAssemblerForm == null)
            {
                createNew = true;
            }
            else
            {
                if (m_ucoreAssemblerForm.IsDisposed)
                {
                    createNew = true;
                }
            }

            if (createNew)
            {
                m_ucoreAssemblerForm = new ucoreAssemblerForm();
                m_ucoreAssemblerForm.enableCloseDialog = false;
            }

            m_ucoreAssemblerForm.Show();
        }


        private void toolStripButtonGfxConvertForm_Click(object sender, EventArgs e)
        {
            bool createNew = false;
            if (m_gfxConnvertForm == null)
            {
                createNew = true;
            }
            else
            {
                if (m_gfxConnvertForm.IsDisposed)
                {
                    createNew = true;
                }
            }

            if (createNew)
            {
                m_gfxConnvertForm = new GfxConvertForm();
                m_gfxConnvertForm.enableCloseDialog = false;
            }

            m_gfxConnvertForm.Show();
        }

        private void toolStripButtonRAWEdit_Click(object sender, EventArgs e)
        {
            bool createNew = false;
            if (m_RawEditForm == null)
            {
                createNew = true;
            }
            else
            {
                if (m_RawEditForm.IsDisposed)
                {
                    createNew = true;
                }
            }

            if (createNew)
            {
                m_RawEditForm = new RAWEditMainForm();
                m_RawEditForm.enableCloseDialog = false;
            }

            m_RawEditForm.clearFileBuffer();

            string sdRawFileName = sdAssemblerDestFileBox.Text;
            if (sdRawFileName != "")
            {
                string[] fileNames = new string[1];
                fileNames[0] = sdRawFileName;
                m_RawEditForm.addFilesToBuffer(fileNames);
                m_RawEditForm.refreshBufferView();
            }

            m_RawEditForm.Show();
        }   

        private void buttonLoadAllInOnePackage_Click(object sender, EventArgs e)
        {
            string fileName = textBoxAllInOnePackage.Text;

            // generate temp path

            string ucoreTempPath = System.IO.Path.GetTempPath() + "ucoreEmulatorFiles\\";
            
            // delete old stuff

            try
            {
                System.IO.Directory.Delete(ucoreTempPath, true);
            }
            catch (Exception ex)
            {
                // first run
            }

            // create new temp

            System.IO.Directory.CreateDirectory(ucoreTempPath);

            try
            {                         
                // extract all files into temp directory

                using (ZipFile zip1 = ZipFile.Read(fileName))
                {
                    foreach (ZipEntry entry in zip1)
                    {                        
                        entry.Extract(ucoreTempPath,ExtractExistingFileAction.OverwriteSilently);                        
                    }
                }

                // load xml config

                XmlTextReader xmlReader = new XmlTextReader( ucoreTempPath + "config.xml");
                XmlDocument xmlDocument = new XmlDocument();

                xmlDocument.Load(xmlReader);

                XmlElement root = xmlDocument["goldmomo_endlos"];
                XmlElement flashRawNode = root["flash_raw"];
                XmlElement ucoreRawNode = root["ucore_raw"];
                XmlElement ucoreSdCardNode = root["sd_card"];

                // load data if set

                if (flashRawNode != null)
                {
                    loadFlash(ucoreTempPath + flashRawNode.InnerText);
                }

                if (ucoreRawNode != null)
                {
                    loadUCoreRaw(ucoreTempPath + ucoreRawNode.InnerText);
                }

                if (ucoreSdCardNode != null)
                {
                    string sdCardFileName = ucoreTempPath + ucoreSdCardNode.InnerText;

                    textBoxSDRawFile.Text = sdCardFileName;
                    m_emulator.ucore.ext_bus.sdCard.sdCard_emulation.setFile(sdCardFileName);
                }

                xmlReader.Close();

            }
            catch (Exception ex)
            {
                print("error: load package file(s) fail! Reason '" + ex.Message + "'\n");
                return;
            }

           

        }

        private void checkBoxCheckIdentifyAsEmulator_CheckedChanged(object sender, EventArgs e)
        {
            m_emulator.ucore.bus.ctrlMemory.identifyAsEmulator = checkBoxCheckIdentifyAsEmulator.Checked;
        }

        private void videoBox_DragDrop(object sender, DragEventArgs e)
        {
            print(e.ToString()+"\n");
        }

        private void videoBox_DragEnter(object sender, DragEventArgs e)
        {
            print(e.ToString() + "\n");
        }

        private void videoBox_DragOver(object sender, DragEventArgs e)
        {
            print(e.ToString() + "\n");
        }

        long m_keyPressedLastTime = 0; 
        private void debugBox_KeyDown(object sender, KeyEventArgs e)
        {            
            long currentTime = DateTime.Now.Ticks / TimeSpan.TicksPerMillisecond;
            long timeDiff = currentTime - m_keyPressedLastTime;
            long minTimeDiff = 250; // ms

            if(timeDiff >= minTimeDiff) // only max every 250ms read pressed key to prevent slow emulator with flood of keypress
            {
                m_scanCodeTranslator.addKeys(m_scanCodeTranslator.translate(e.KeyCode, true));
                m_keyPressedLastTime = currentTime;
            }
            e.SuppressKeyPress = true;
        }

        private void debugBox_KeyUp(object sender, KeyEventArgs e)
        {
           m_scanCodeTranslator.addKeys(m_scanCodeTranslator.translate(e.KeyCode, false));
           e.SuppressKeyPress = true;            
        }

        private void copyAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
            Clipboard.SetText(debugBox.Text);
            }
            catch (Exception) { }
        }

        private void checkedListBoxSW_SelectedIndexChanged(object sender, EventArgs e)
        {
            for( int i=0;i<checkedListBoxSW.Items.Count;i++)
            {
                m_emulator.boardCtrl.set_sw(i, checkedListBoxSW.GetItemChecked(i));              
            }         
        }

        private void buttonSelectSDRawFile_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog1 = new OpenFileDialog();


                if (textBoxSDRawFile.Text != "")
                {
                    openFileDialog1.FileName = textBoxSDRawFile.Text;
                    openFileDialog1.InitialDirectory = Path.GetDirectoryName(textBoxSDRawFile.Text);
                }
                else
                {
                    openFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog1.Filter = "All files (*.*)|*.*";
                openFileDialog1.FilterIndex = 2;
                openFileDialog1.RestoreDirectory = true;

                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog1.FileName;
                    textBoxSDRawFile.Text = fileName;

                    m_emulator.ucore.ext_bus.sdCard.sdCard_emulation.setFile(fileName);
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }    
        }

        private void checkBoxSDHC_CheckedChanged(object sender, EventArgs e)
        {
            m_emulator.ucore.ext_bus.sdCard.sdCard_emulation.isSDHC = checkBoxSDHC.Checked;
        }

        private void checkBoxEnableReadAudio_CheckedChanged(object sender, EventArgs e)
        {
            m_emulator.ucore.bus.ctrlMemory.audio.enableRealAudio = checkBoxEnableReadAudio.Checked;
        }

        private void checkBoxEnableUCoreRuntimeCheck_CheckedChanged(object sender, EventArgs e)
        {
            m_emulator.ucore.enableRuntimeCheck = checkBoxEnableUCoreRuntimeCheck.Checked;
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show("goldmomo_endlos emulator "+getBuildVersion()+"\n"+
                            "Its for free use, do what you want, no warranty.\n\n"+
                            "Written by goldmomo (Marten Wobst)\n\nVisit www.goldmomo.org or www.youtube.com/user/goldmomo\n",
                            "About...");
        }

        private void videoBox_Resize(object sender, EventArgs e)
        {
            lock(_lockResize)
            {
                m_resizeCount = 10;
            }
        }

        private void checkBoxFillSDRamSRamffff_CheckedChanged(object sender, EventArgs e)
        {
            clearMode_t clearMode = clearMode_t.RANDOM;

            if (checkBoxFillSDRamSRamffff.Checked)
            {
                clearMode = clearMode_t.FILL_ffff;
            }

            m_emulator.ucore.ext_bus.dsram.setClearMode(clearMode);
            m_emulator.ucore.ext_bus.sram.setClearMode(clearMode);
        }

        private void numericUpDownHCoreCount_ValueChanged(object sender, EventArgs e)
        {
            m_emulator.hcoreCount = (uint)numericUpDownHCoreCount.Value;
        }


        uint screenShot_counter = 1;

        private void screenshotToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /* make screen shot*/
           
            string fileName = "";

            /*  lock (_vBlankLock)*/
            {
                try
                {
                    SaveFileDialog saveFileDialog = new SaveFileDialog();

                    saveFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                    saveFileDialog.Filter = "All files (*.*)|*.*";
                    saveFileDialog.FilterIndex = 1;
                    saveFileDialog.RestoreDirectory = true;
                    saveFileDialog.FileName = "screenShot"+screenShot_counter+".png";

                    if (saveFileDialog.ShowDialog() == DialogResult.OK)
                    {
                        fileName = saveFileDialog.FileName;
                        screenShot_counter++;
                    }


                    m_screenBitmap.Save(fileName, ImageFormat.Png);
                }
                catch (Exception)
                {
                    print("Error while selecting/open file or dir\n");
                }
            }
        }


    }
}
