﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_emulator.core.components
{
    public class Timer32 : IClocked
    {
        UInt32 m_loadValue;
        UInt32 m_count, m_last_count;
        bool m_start;
        bool m_stop;
        bool m_autoReload;
        bool m_timerHit;


        public enum timer_state_t
        {
            STIMER_WAIT,
            STIMER_RUN
        }

        timer_state_t m_timer_state;

        public bool timerHit
        {
            get { return m_timerHit; }
        }

        public Timer32()
        {
            reset();
        }

        // gui / debug get 

        public UInt32 loadValue
        {
            get { return m_loadValue; }
        }

        public UInt32 count
        {
            get { return m_count; }
        }

        public UInt32 last_count
        {
            get { return m_last_count; }
        }

        public bool start
        {
            get { return m_start; }
        }

        public bool stop
        {
            get { return m_stop; }
        }

        public bool autoReload
        {
            get { return m_autoReload; }
        }

        public timer_state_t timerState
        {
            get { return m_timer_state; }
        }

        //

        public void setLowCount(UInt16 din)
        {
            m_loadValue = (m_loadValue & 0xffff0000) | ((UInt32)din);
        }

        public void setHighCount(UInt16 din)
        {
            m_loadValue = (m_loadValue & 0x0000ffff) | (((UInt32)din) << 16);
        }

        public void setTimerStart(bool din)
        {
            m_start = din;
        }

        public void setTimerStop(bool din)
        {
            m_stop = din;
        }

        public void setTimerAutoReload(bool din)
        {
            m_autoReload = din;
        }

        public void reset()
        {
            m_loadValue = 0;
            m_start = false;
            m_stop = false;
            m_autoReload = false;
            m_timer_state = timer_state_t.STIMER_WAIT;
            m_count = 0;
            m_timerHit = false;
            m_last_count = 0;
        }

        public void step()
        {
            m_timerHit = false;
            m_last_count = m_count;
            switch (m_timer_state)
            {
                case timer_state_t.STIMER_WAIT:
                    if (m_start)
                    {
                        m_count = m_loadValue;
                        m_timer_state = timer_state_t.STIMER_RUN;
                    }
                    else
                    {
                        m_timer_state = timer_state_t.STIMER_WAIT;
                    }
                    break;
                case timer_state_t.STIMER_RUN:
                    m_count--;

                    if (m_stop)
                    {
                        m_timer_state = timer_state_t.STIMER_WAIT;
                    }
                    else
                    {
                        if (m_count == 0)
                        {
                            m_timerHit = true;

                            if (m_autoReload)
                            {
                                m_count = m_loadValue;
                            }
                            else
                            {
                                m_timer_state = timer_state_t.STIMER_WAIT;
                            }
                        }

                    }

                    break;
            }

            m_start = false;
            m_stop = false;
        }
    }

}
