﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_emulator.core.components.extMemory
{
    public class Flash : IExtMemory, IClocked
    {
        int m_size = 0;
        int m_sizeMask;
        bool m_writeProtection;

        UInt16 []m_flash = null;


        public Flash(int width = 22) // 1 megaword
        {
            m_size = 1 << width;
            m_sizeMask = m_size - 1;
            m_writeProtection = true;
            m_flash = new UInt16[m_size];

            for (int i = 0; i < m_size; i++)
            {
                m_flash[i] = 0xffff;
            }

            reset();
        }

        public void setClearMode(clearMode_t mode)
        {
            // nothing to do here
        }

        public bool writeProtection
        {
            get { return m_writeProtection;  }
            set { m_writeProtection = value;  }
        }


        public void write(uint adr, UInt16 value)
        {
            if (writeProtection)
            {
                throw new Exception(string.Format("Flash: illegal write access to ${0:X8} <= ${1:X4}", adr, value));
            }
            else
            {
                m_flash[adr & m_sizeMask] = value;
            }
        }

        public void read(uint adr, out UInt16 value)
        {
           value = m_flash[adr & m_sizeMask];
        }


        public void write(UInt32 adr, UInt32 value)
        {
            if (writeProtection)
            {
                throw new Exception(string.Format("Flash: illegal write access to ${0:X8} <= ${1:X8}", adr, value));
            }
            else 
            {
                m_flash[(adr * 2) & m_sizeMask] = (UInt16)(value >> 16);
                m_flash[((adr * 2) + 1) & m_sizeMask] = (UInt16)value;
            }
        }

        public void read(UInt32 adr, out UInt32 value)
        {
            value = (UInt32)m_flash[((adr * 2) + 1) & m_sizeMask];
            value |= ((UInt32)m_flash[((adr * 2)) & m_sizeMask]) << 16;
        }

        public void reset()
        {
        }

        public void step()
        {
           
        }


        public int size
        {
            get 
            { 
                return m_size; 
            }
        }
    }
}
