﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library;
using System.IO;
using ucore_emulator.core.components.extMemory.SDCard;
using xcore_tools;

namespace ucore_emulator.core.components.extMemory
{
    public class sdCard_ctrl : IExtMemory, IClocked
    {
        Random m_random = new Random();
        SDCardEmulation m_sdCard;
        SDClk m_sdClk;
        IPrint m_iprint;
        bool m_enableReadBurst = false;
        UInt32 m_readBurstAddress = 0;
        ext_bus m_ext_bus;

        //

        UInt16 m_test_register = 0;

        // latch

        UInt16 m_latch_prg_clk_msk;             
        UInt16 m_latch_prg_clk_edge_cmp_down;	
        UInt16 m_latch_prg_clk_edge_cmp_up;
        UInt16 m_latch_prg_clk_write_edge_cmp;
        UInt16 m_latch_prg_clk_read_edge_cmp;	

        //


        public sdCard_ctrl(IPrint iprint, ext_bus iext_bus) 
        {
            m_iprint  = iprint;
            m_ext_bus = iext_bus;

            m_sdCard = new SDCardEmulation(m_iprint);
            m_sdClk = new SDClk();

            reset();
        }

        public void setClearMode(clearMode_t mode)
        {
            // nothing to do here
        }

        public UInt32 burstAddress
        {
            get { return m_readBurstAddress;  }
        }

        public SDCardEmulation sdCard_emulation
        {
            get { return m_sdCard; }
        }

        public SDClk sdClk
        {
            get { return m_sdClk; }
        }

        UInt16 send_buffer_cmd;
        UInt32 send_buffer_param;

        public void write(uint adr, UInt16 value)
        {
            switch (adr & 0x7f)
            {
                case 0: send_buffer_cmd = (UInt16)(value & 0x3f);
                        
                        if ((value & 0x0100) != 0)   // start send
                        {
                            SDCommand cmd = new SDCommand(send_buffer_cmd, send_buffer_param, ((value & 0x0400) != 0));
                            m_sdCard.sendCmd(cmd);
                        }
                        if ((value & 0x0800) != 0)   // latch to clk
                        {
                            m_sdClk.prg_clk_msk                 = m_latch_prg_clk_msk;            
                            m_sdClk.prg_clk_edge_cmp_down       = m_latch_prg_clk_edge_cmp_down; 
                            m_sdClk.prg_clk_edge_cmp_up         = m_latch_prg_clk_edge_cmp_up;    
                            m_sdClk.prg_clk_write_edge_cmp      = m_latch_prg_clk_write_edge_cmp; 
                            m_sdClk.prg_clk_read_edge_cmp       = m_latch_prg_clk_read_edge_cmp; 
                        }
                        break;

                case 1: break;
                case 2: send_buffer_param = (send_buffer_param & 0x0000ffff) | (((UInt32)value) << 16);break;   // high send param
                case 3: send_buffer_param = (send_buffer_param & 0xffff0000) | (((UInt32)value)); break;        // low send param


                // clk

                case 8  : m_latch_prg_clk_msk               = (UInt16)(value & 0x1ff); break;                  					
				case 9 	: m_latch_prg_clk_edge_cmp_down	    = (UInt16)(value & 0x1ff); break;							
				case 10	: m_latch_prg_clk_edge_cmp_up       = (UInt16)(value & 0x1ff); break;
				case 11 : m_latch_prg_clk_write_edge_cmp    = (UInt16)(value & 0x1ff); break;
                case 12 : m_latch_prg_clk_read_edge_cmp	    = (UInt16)(value & 0x1ff); break;	
                
                //

                case 16: m_enableReadBurst = (value & 0x2) != 0;
                         break;

                case 17: break;
                case 18: m_readBurstAddress = (m_readBurstAddress & 0xffff0000) | (((UInt32)value)); break;
                case 19: m_readBurstAddress = (m_readBurstAddress & 0x0000ffff) | (((UInt32)value) << 16); break;
                case 25: m_test_register = value; break;
                default: break; // dont care
            }

        }

        public void read(uint adr, out UInt16 value)
        {
            switch (adr & 0x7f)
            {
                case 0:
                    {
                        SDCommandResponse cmdResponse = m_sdCard.receiveCmd;

                        value = (UInt16) (cmdResponse.commandPass?4:2);//(UInt16)((m_random.Next() & 0x1f) == 0 ? 4 : 0);   //recive ok random it
                        value |= (UInt16)(cmdResponse.cmd << 8);
                    }
                    break;
                case 1:
                    {
                        SDCommandResponse cmdResponse = m_sdCard.receiveCmd;

                        value = (UInt16)((cmdResponse.crc7expected << 8) | (cmdResponse.crc7));
                    }
                    break;

                case 2: value = m_sdCard.receiveCmd.arg[7]; break;
                case 3: value = m_sdCard.receiveCmd.arg[6]; break;
                case 4: value = m_sdCard.receiveCmd.arg[5]; break;
                case 5: value = m_sdCard.receiveCmd.arg[4]; break;
                case 6: value = m_sdCard.receiveCmd.arg[3]; break;
                case 7: value = m_sdCard.receiveCmd.arg[2]; break;
                case 8: value = m_sdCard.receiveCmd.arg[1]; break;
                case 9: value = m_sdCard.receiveCmd.arg[0]; break;

                case 16: value = 0xf; break; // read ever fine, FIXME

                case 21: value = (UInt16)(m_sdCard.dataInReadQueueDataCount() / 4); break;
                case 22: value = (UInt16)(m_sdCard.dataInReadQueueDataCount() / 4); break;
                case 23: value = (UInt16)(m_sdCard.dataInReadQueueDataCount() / 4); break;
                case 24: value = (UInt16)(m_sdCard.dataInReadQueueDataCount() / 4); break;
                case 25: value = m_test_register; break;
    
                default: value = 0xbada; break;
            }
        }


        public void write(UInt32 adr, UInt32 value)
        {
            throw new Exception("sd card doent accept 32 bit io");
        }

        public void read(UInt32 adr, out UInt32 value)
        {
            throw new Exception("sd card doent accept 32 bit io");
        }

        public void reset()
        {
            m_enableReadBurst = false;
            m_readBurstAddress = 0;

            m_sdCard.reset();
            m_sdClk.reset();
        }

        public void step()
        {
            if (m_enableReadBurst)
            {
                if (m_sdCard.dataInReadQueueDataCount() >= 2)
                {
                    byte h = m_sdCard.getByteFromReadQueue();
                    byte l = m_sdCard.getByteFromReadQueue();

                    UInt16 data = (UInt16)(((UInt16)h) << 8 | ((UInt16)l));
                    m_ext_bus.write(m_readBurstAddress, data);
                    m_readBurstAddress++;
                }
            }

            m_sdCard.step();
        }

        public int size
        {
            get
            {
                return 0;
            }
        }
    }
}
