﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library;
using System.IO;
using xcore_tools;

namespace ucore_emulator.core.components.extMemory.SDCard
{
    public class SDCardEmulation : IClocked, IPrint
    {
        IPrint m_iprint;
        bool m_appMode = false;
        SDCommandResponse m_sdCmdResponse;
        string m_fileName = "";
        Queue<byte> m_readQueue = new Queue<byte>();
        bool m_isSDHC = true;

        public UInt32 currentReadBlock  /* only for visualize in emulator */
        {
            get; 
            set;
        }

        SDCommand m_lastSendCmd = new SDCommand(0,0,false); // only for debug view

        public SDCardEmulation(IPrint iprint)
        {
            m_iprint = iprint;
            isSDHC = true;  // default is sdhc emulation (up to 2TB ...)
            m_sdCmdResponse = new SDCommandResponse(0, 0, 0);
        }

        public SDCommand lastSendCmd
        {
            get { return m_lastSendCmd; }
        }

        public bool isSDHC
        {
            get { return m_isSDHC; }
            set { m_isSDHC = value; }
        }

        public int dataInReadQueueDataCount()
        {
            return m_readQueue.Count;
        }

        public byte getByteFromReadQueue()
        {
            byte rByte = 0;
            if (m_readQueue.Count > 0)
            {
                rByte = m_readQueue.Dequeue();
            }
            return rByte;
        }

        public void setFile(string fileName)
        {
            m_fileName = fileName;
        }

        public void sendCmd(SDCommand c)
        {
            //dprint(String.Format("SDCard CMD recive CMD{0:00} Parameter {1:X8} next recive is {2}\n", c.cmd, c.param, c.resive136 ? "136Bit" : "48Bit"));
            m_lastSendCmd = c; // debug view

            switch (c.cmd)
            {
                case 0: // reset -> no response
                    break;

                case 2: // get cid
                    {
                        UInt16[] cid = new UInt16[8];
                        cid[0] = 0x0C1D;
                        cid[1] = 0x0123;
                        cid[2] = 0x4567;
                        cid[3] = 0x89ab;
                        cid[4] = 0xcdef;
                        cid[5] = 0xedcb;
                        cid[6] = 0xa987;
                        cid[7] = 0x6543;
                        m_sdCmdResponse = new SDCommandResponse(c.cmd, cid);
                    }
                    break;

                case 3: // get rca
                    m_sdCmdResponse = new SDCommandResponse(c.cmd, 0x05CA, 0); // last is byte is echo buffer
                    break;

                case 9: // get csd                       
                    {
                        UInt16[] csd = new UInt16[8];
                        csd[0] = 0x0C5D;
                        csd[1] = 0x0123;
                        csd[2] = 0x4567;
                        csd[3] = 0x89ab;
                        csd[4] = 0xcdef;
                        csd[5] = 0xedcb;
                        csd[6] = 0xa987;
                        csd[7] = 0x6543;
                        m_sdCmdResponse = new SDCommandResponse(c.cmd, csd);
                    }
                    break;

                case 17:    // read single block

                    m_sdCmdResponse = new SDCommandResponse(c.cmd, 0, 0); // 

                    // read from file

                    currentReadBlock = c.param;

                    int start = ((int)c.param) * (isSDHC ? 512 : 1);   // card is sdhc so param in blocks
                    byte[] blockBuffer = new byte[512];

                    try
                    {
                        using (FileStream fs = File.Open(m_fileName, FileMode.Open, FileAccess.Read, FileShare.None))
                        {
                            fs.Seek(start, SeekOrigin.Begin);
                            fs.Read(blockBuffer, 0, 512);

                            // push into readQueue

                            for (int i = 0; i < 512; i++)
                            {
                                m_readQueue.Enqueue(blockBuffer[i]);
                            }
                            fs.Close();
                        }
                    }
                    catch (Exception)
                    {
                        m_sdCmdResponse.commandPass = false;
                    }

                    //


                    break;

                case 55:    // switch to app
                    m_appMode = true;
                    m_sdCmdResponse = new SDCommandResponse(c.cmd, 0, 0); // 
                    break;
                case 41:    // ACMD
                    if (m_appMode)
                    {
                        UInt16 rValue = 0x8000; // bit15 -> sd card is not busy 
                        if (isSDHC)
                        {
                            rValue |= 0x4000;   // bit14 -> 0 = SD, 1 = SDHC
                        }
                        m_sdCmdResponse = new SDCommandResponse(c.cmd, rValue, 0); // not busy and SDHC (bit 15 & 14)
                    }
                    break;

                default: m_appMode = false; break;
            }
        }

        public SDCommandResponse receiveCmd
        {
            get
            {
                //dprint(String.Format("SDCard CMD result CMD{0:00} Parameter {1:X4}{2:X4}{3:X4}{4:X4}{5:X4}{6:X4}{7:X4}{8:X4} crc {9} excrc {10}\n", m_sdCmdResponse.cmd, m_sdCmdResponse.arg[0], m_sdCmdResponse.arg[1], m_sdCmdResponse.arg[2], m_sdCmdResponse.arg[3], m_sdCmdResponse.arg[4], m_sdCmdResponse.arg[5], m_sdCmdResponse.arg[6], m_sdCmdResponse.arg[7], m_sdCmdResponse.crc7, m_sdCmdResponse.crc7expected));
                return m_sdCmdResponse; // no queue here
            }
        }

        //



        //

        public void reset()
        {
            m_readQueue.Clear();
        }

        public void step()
        {

        }

        #region IPrint
        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
        #endregion
    }
}
