﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.stages;
using xcore_tools;

namespace ucore_emulator.core.components
{
    public struct hcore_thread_request
    {       
		 public uint max_cores;
		 public uint handover_address;
		 public uint jump_address;	
    };

    public struct hcore_thread_response
    {       
		 public uint max_cores;
         public uint current_core;
		 public uint handover_address;
		 public uint jump_address;	
    };

    public class hcore_thread_ctrl : IStage, IPrint
    {
        static object _lock = new object();

        IPrint m_iprint;
        SortedDictionary<uint, bool> m_idleCores;


        SortedDictionary<uint, hcore_thread_response> m_response;

        public hcore_thread_ctrl(IPrint iprint)
        {
            m_iprint = iprint;
            m_idleCores = new SortedDictionary<uint, bool>();
            m_response = new SortedDictionary<uint, hcore_thread_response>();
        }

        public void setCoreIdle(uint coreId, bool idle)
        {
            lock (_lock)
            {
                m_idleCores[coreId] = idle;
            }
        }

        public uint parallelRequest(hcore_thread_request req)
        {
            uint max_cores = 1; // ever one available (master self)

            lock (_lock)
            {
                

                foreach (KeyValuePair<uint, bool> ic in m_idleCores)
                {
                    if (ic.Value) // in idle
                    {
                        max_cores++;
                    }
               
                }

                uint coreNum=1;
                foreach (KeyValuePair<uint, bool> ic in m_idleCores)
                {
                    if(ic.Value) // in idle
                    {
                        uint coreId = ic.Key;

                        hcore_thread_response res = new hcore_thread_response();

                        res.current_core = coreNum++;
                        res.handover_address = req.handover_address;
                        res.jump_address = req.jump_address;
                        res.max_cores = max_cores;

                        m_response[coreId] = res;
                    }
                }
            }
            return max_cores;
        }

        public hcore_thread_response popResponse(uint coreId,out bool available)
        {
            hcore_thread_response res = new hcore_thread_response();

            available = false;

            lock (_lock)
            {
                if (m_response.ContainsKey(coreId) == true)
                {
                    res = m_response[coreId];
                    m_response.Remove(coreId);
                    available = true;
                }
            }

            return res;

        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }

        public void reset()
        {
            lock (_lock)
            {
                m_idleCores.Clear();
            }
        }
    }
}
