﻿using System;
using System.Collections.Generic;
using System.Text;
using ucore_emulator.common;

namespace ucore_emulator.core.components.memory.ctrl
{
    public class BoardCtrl
    {
        ScanCodeTranslator m_scanCodeTranslator = null;

        public BoardCtrl(ScanCodeTranslator iScanCodeTranslator)
        {
            m_scanCodeTranslator = iScanCodeTranslator;
            
            set_rled(0, 0xffffffff);
            set_gled(0, 0xffffffff);
            

            for (uint i = 0; i < hexCount; i++)
            {
                set_hex(i, 0x0);    //____ default
            }
            set_sw(0);
        }

        public ScanCodeTranslator scanCodeTranslator
        {
            get { return m_scanCodeTranslator; }
        }

        public void get_sw(out UInt32 sw)
        {
            sw = (UInt32)(m_sw & 0x7ffff); // hw support only 19 switches
        }

        public void set_sw(UInt32 sw)
        {
            m_sw = sw;
        }

        public void set_sw(int sw, bool on)
        {
            if (on)
            {
                m_sw |= (UInt32)(1 << sw);
            }
            else
            {
                m_sw &= (UInt32)~((1 << sw));
            }
        }

        public void set_rled(uint set, uint reset)
        {
            m_rled &= ~reset;
            m_rled |= set;
        }

        public void set_gled(uint set, uint reset)
        {
            m_gled &= ~reset;
            m_gled |= set;
        }

      

        public void set_hex(uint i,uint value)
        {
            if (i < m_hexCount)
            {
                m_hex[i] = value;
            }
        }

        public bool get_rled(uint i)
        {
            if (i < m_rledCount)
            {
                return ((m_rled >> (int)i) & 1) == 1;
            }
            return false;
        }       

        public bool get_gled(uint i)
        {
            if (i < m_gledCount)
            {
                return ((m_gled >> (int)i) & 1) == 1;
            }
            return false;
        }

        public bool get_hex(uint i, uint d)
        {
            if (i < m_hexCount )
            {
                return ((m_hex[i]>> (int)d) & 1) == 1;
            }
            return false;
        }

        public void get_joys(out UInt16 joys)
        {
            joys = m_joys;
        }

        public enum joystick_e
        {
            JS_UP,
            JS_DOWN,
            JS_LEFT,
            JS_RIGHT,
            JS_BC,
            JS_BB,
            JS_BA
        };

        public void set_joy(int port,joystick_e type,bool set)
        {
            UInt16 bits = 0;

            switch (type)
            {
                case joystick_e.JS_UP: bits = 1 << 7; break;
                case joystick_e.JS_DOWN: bits = 1 << 6; break;
                case joystick_e.JS_LEFT: bits = 1 << 5; break;
                case joystick_e.JS_RIGHT: bits = 1 << 4; break;
                case joystick_e.JS_BC: bits = 1 << 3; break;
                case joystick_e.JS_BB: bits = 1 << 2; break;
                case joystick_e.JS_BA: bits = 1 << 1; break;
            }

            if (set)
            {
                m_joys |= (UInt16)(bits << (port * 8));
            }
            else
            {
                m_joys &= (UInt16) ~((UInt16)(bits << (port * 8)));
            }
        }

        public uint hexCount
        {
            get { return m_hexCount; }
        }

        public uint rledCount
        {
            get { return m_rledCount; }
        }

        public uint gledCount
        {
            get { return m_gledCount; }
        }

        const int m_rledCount = 18;
        const int m_gledCount = 9;
        const int m_hexCount = 8;

        public enum keyBoardStatus
        {
            KS_CAPS,
            KS_SCROLL,
            KS_NUM
        }

        static public keyBoardStatus keyBoardStatusToEnum(string sel)
        {
            switch (sel)
            {
                case "CAPS": return keyBoardStatus.KS_CAPS;
                case "SCRL": return keyBoardStatus.KS_SCROLL;
                case "NUM": return keyBoardStatus.KS_NUM;

                default:
                    throw new Exception("keyBoardStatusToEnum: illegal input value");
            }
        }

        public bool getKeyBoardStatus(keyBoardStatus sel)
        {
            switch (sel)
            {
                case keyBoardStatus.KS_CAPS: return m_keyBoard_caps;
                case keyBoardStatus.KS_SCROLL: return m_keyBoard_scroll;
                case keyBoardStatus.KS_NUM: return m_keyBoard_num;

                default:
                    throw new Exception("getKeyBoardStatus: illegal input value");
            }
        }

        public void setKeyBoardStatus(keyBoardStatus sel,bool value)
        {
            switch (sel)
            {
                case keyBoardStatus.KS_CAPS: m_keyBoard_caps = value;break;
                case keyBoardStatus.KS_SCROLL: m_keyBoard_scroll = value;break;
                case keyBoardStatus.KS_NUM: m_keyBoard_num = value;break;

                default:
                    throw new Exception("setKeyBoardStatus: illegal input value");
            }
        }

       


        bool m_keyBoard_caps = false;
        bool m_keyBoard_scroll = false;
        bool m_keyBoard_num = false;

        UInt32 m_rled = 0;
        UInt32 m_gled = 0;
        UInt32[] m_hex = new UInt32[m_hexCount];

        UInt32 m_sw;
        UInt16 m_joys = 0;
    }
}
