﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library;
using xcore_tools;

namespace ucore_emulator.core.components.memory.ctrl
{
    public class I2C : IPrint
    {
        public enum I2CState_e
        {
            WAIT_START_A,
            WAIT_START_B,
            D0,
            D1,
            D2,
            D3,
            D4,
            D5,
            D6,
            D7,
            ACK
        }



        I2CState_e m_I2CState;
        char []m_bits = new char[8];

        IPrint m_iprint;

        public I2C(IPrint iprint) 
        {
            m_I2CState = I2CState_e.WAIT_START_A;
            m_iprint = iprint;

            for (int i = 0; i < 8; i++)
            {
                m_bits[i] = '?';
            }
        }

        public void write(bool sck, bool sda)
        {
            switch (m_I2CState)
            {
                case I2CState_e.WAIT_START_A:
                    if (sda && sck)
                    {
                        m_I2CState = I2CState_e.WAIT_START_B;
                    }
                    break;

                case I2CState_e.WAIT_START_B:
                    if (sck && !sda)
                    {
                        m_I2CState = I2CState_e.D0;
                    }
                    break;

                case I2CState_e.D0: 
                    if (sck)
                    {
                        m_bits[7] = sda?'1':'0'; m_I2CState = I2CState_e.D1;
                    }
                    break;
                case I2CState_e.D1:
                      if (sck)
                    {
                        m_bits[6] = sda?'1':'0'; m_I2CState = I2CState_e.D2;
                    }
                    break;
                case I2CState_e.D2: 
                      if (sck)
                    {
                        m_bits[5] = sda?'1':'0'; m_I2CState = I2CState_e.D3;
                    }
                    break;
                case I2CState_e.D3: 
                      if (sck)
                    {
                        m_bits[4] = sda?'1':'0'; m_I2CState = I2CState_e.D4;
                    }
                    break;
                case I2CState_e.D4: 
                      if (sck)
                    {
                        m_bits[3] = sda?'1':'0'; m_I2CState = I2CState_e.D5;
                    }
                    break;
                case I2CState_e.D5:
                      if (sck)
                    {
                        m_bits[2] = sda?'1':'0'; m_I2CState = I2CState_e.D6;
                    }
                    break;
                case I2CState_e.D6:
                      if (sck)
                    {
                        m_bits[1] = sda?'1':'0'; m_I2CState = I2CState_e.D7;
                    }
                    break;
                case I2CState_e.D7:
                    if (sck)
                    {
                        m_bits[0] = sda?'1':'0'; m_I2CState = I2CState_e.ACK;
                    }
                    break;
                case I2CState_e.ACK: 
                    if (sck)
                    {
                        m_I2CState = I2CState_e.WAIT_START_A;

                        string bitString = "";

                        int value = 0;
                        for (int i = 0; i < 8; i++)
                        {
                            value |= m_bits[7 - i] == '1' ? 1 : 0;
                            value <<= 1;
                            bitString += m_bits[7-i]; 
                        }

                        print(string.Format("I2C packet ${0:X2} %{1}\n",value, bitString ));
                    }
                    break;
            }
        }



        public void print(string txt)
        {
            if (m_iprint != null)
            { 
                m_iprint.print(txt); 
            }

        }

        public void dprint(string txt)
        {
            if (m_iprint != null)
            {
                m_iprint.dprint(txt);
            }
        }
    }
}
