﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_emulator.core.components.stages
{
    public class UCoreAnalaystic
    {
        public UCoreAnalaystic()
        {
            clear();
        }

        object _usageLock = new object();

        public void increasOpcodeUsage(opcode_t op)
        {
            lock (_usageLock)
            {
                if (m_opcodeUsage.ContainsKey(op))
                {
                    m_opcodeUsage[op]++;
                }
                else
                {
                    m_opcodeUsage[op] = 1;
                }
            }
        }

        public KeyValuePair<opcode_t, Int64>[] safeOpcodeUsage
        {
            get
            {
                lock(_usageLock)
                {
                    KeyValuePair<opcode_t, Int64>[] rout = new KeyValuePair<opcode_t, Int64>[m_opcodeUsage.Count];
                    m_opcodeUsage.CopyTo(rout, 0);

                    return rout;
                }
            }
        }

        public void incrementLineUsage(uint line)
        {
            lock (_usageLock)
            {
                if (m_lineUsage.ContainsKey(line))
                {
                    m_lineUsage[line]++;
                }
                else
                {
                    m_lineUsage[line]=1;
                }
            }
        }

        public uint getLineUsage(uint line)
        {
            uint rValue = 0;
            lock (_usageLock)
            {               
                if (m_lineUsage.ContainsKey(line))
                {
                    rValue = m_lineUsage[line];
                }
            }
            return rValue;
        }

        public void clear()
        {
            lock (_usageLock)
            {
                m_lastBranchCycle = 0;
                m_minBranchDelay = UInt64.MaxValue;
                m_maxBranchDelay = 0;
                m_lineUsage.Clear();
                m_opcodeUsage.Clear();
            }
        }

        // irq/bra latency

        UInt64 m_lastBranchCycle = 0;
        UInt64 m_minBranchDelay = UInt64.MaxValue;
        UInt64 m_maxBranchDelay = UInt64.MinValue;
        UInt64 m_aprBranchDelay = 0;
        UInt64 m_branchCount = 0;

        public UInt64 branchCount
        {
            get { return m_branchCount; }
            set { lock (_usageLock) { m_branchCount = value; } }
        }

        public UInt64 aprBranchDelay
        {
            get { return m_aprBranchDelay; }
            set { lock (_usageLock) { m_aprBranchDelay = value; } }
        }

        public UInt64 maxBranchDelay
        {
            get { lock (_usageLock) { return m_maxBranchDelay; } }
            set { lock (_usageLock) { m_maxBranchDelay = value; } }
        }

        public UInt64 minBranchDelay
        {
            get { lock (_usageLock) { return m_minBranchDelay; } }
            set { lock (_usageLock) { m_minBranchDelay = value; } }
        }

        public UInt64 lastBranchCycle
        {
            get { lock (_usageLock) { return m_lastBranchCycle; } }
            set { lock (_usageLock) { m_lastBranchCycle = value; } }
        }

        SortedDictionary<uint, uint> m_lineUsage = new SortedDictionary<uint, uint>();
        SortedDictionary<opcode_t, Int64> m_opcodeUsage = new SortedDictionary<opcode_t, Int64>();
        
    }
}
