﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.common;
using xcore_tools.Common;

namespace ucore_emulator.core.components.stages
{
    public class s2_decode : bitExtraction, IStage 
    {
        class decodeCache_line
        {
            public decodeCache_line()
            { 
                valid = false;
                dout = new s2_out_s3_in();
            }

            public bool valid;
            public s2_out_s3_in dout;
        }

        decodeCache_line[] decodeCache = new decodeCache_line[65536];

        public s2_decode()
        {
            for (int i = 0; i < 65536; i++)
            {
                decodeCache[i] = new decodeCache_line();
            }
        }

        public void reset()
        {
           
        }

        public s2_out_s3_in dataOut
        {
            get { return m_out; }
        }

        s2_out_s3_in m_out = new s2_out_s3_in();

        //

        UInt16 m_CurrentDiextended = 0;
        UInt16 m_NextDiextended = 0;

        //
       
        public void process(s1_out_s2_in din, out s2_out_s3_in dout)
        {
            // new decode cache (need more memory, but how cares ..)

          /*  decodeCache_line dl = decodeCache[(int)din.data_in];

            dout = dl.dout;

            if (dl.valid == true)
            {
                return;
            }
            else 
            {
                dl.valid = true;
            }*/

            // decoder immediate extension

            m_CurrentDiextended = m_NextDiextended;
            m_NextDiextended = 0;       // only valid for next cycle

            // old no cache
            dout = m_out;

            // default

            dout.decode_rc_we = false;
            dout.decode_imm_a = false;
            dout.debug_pc = din.debug_pc; // only for debug
			
			// opcode

            if ((din.data_in & 0x8000) == 0) // reg direct
            {
                switch ((din.data_in >> 11) & 0xf)//extr(din.data_in,14,11))
                {
                    case 0: dout.decode_opcode = opcode_t.ope_movei; dout.decode_rc_we = true; break;
                    case 1: dout.decode_opcode = opcode_t.ope_moveih; dout.decode_rc_we = true; break;
                    case 2: dout.decode_opcode = opcode_t.ope_addi; dout.decode_rc_we = true; break;
                    case 3: dout.decode_opcode = opcode_t.ope_subi; dout.decode_rc_we = true; break;
                    case 4: dout.decode_opcode = opcode_t.ope_lsri; dout.decode_rc_we = true; break;
                    case 5: dout.decode_opcode = opcode_t.ope_asri; dout.decode_rc_we = true; break;
                    case 6: dout.decode_opcode = opcode_t.ope_muli; dout.decode_rc_we = true; break;
                    case 7: dout.decode_opcode = opcode_t.ope_cmpeqi; break;
                    case 8: dout.decode_opcode = opcode_t.ope_cmploi; break;
                    case 9: dout.decode_opcode = opcode_t.ope_cmplosi; break;
                    case 10: dout.decode_opcode = opcode_t.ope_gpci; dout.decode_rc_we = true; break;
                    case 11: dout.decode_opcode = opcode_t.ope_jmpi; break;
                    case 12: dout.decode_opcode = opcode_t.ope_extri; break;
                    case 13: dout.decode_opcode = opcode_t.ope_rqldi; break;
                    case 14: dout.decode_opcode = opcode_t.ope_getsp; dout.decode_rc_we = true; break;

                    case 15: dout.decode_opcode = opcode_t.ope_dexti;                                     // dexti => special decoder instruction, send nop to alu (emulator will use dexti in alu for anlayse)
                                                    m_NextDiextended = (UInt16)(din.data_in & 0x07ff);    // 11 bit
                                                    break;
                }

                dout.decode_rc = (UInt16)((din.data_in >> 8) & 0x7); 
                dout.decode_rb = (UInt16)((din.data_in >> 8) & 0x7); 
                dout.decode_ra = 0;
                dout.decode_imm = (UInt16)((din.data_in & 0xff) | (m_CurrentDiextended << 8));
                dout.decode_imm_a = true;
            }
            else
            {
                if ((din.data_in & 0x4000) == 0) // (extr(din.data_in, 14, 14) == 0) // reg, reg, reg / reg, reg, imm3
                {
                    if ((din.data_in & 0x2000) == 0)//(extr(din.data_in, 13, 13) == 0) // reg, reg, reg
                    {
                        switch ((din.data_in >> 9) & 0xf)// (extr(din.data_in, 12, 9))
                        {
                            // reg, reg, reg

                            case 0: dout.decode_opcode = opcode_t.ope_add; dout.decode_rc_we = true; break;
                            case 1: dout.decode_opcode = opcode_t.ope_sub; dout.decode_rc_we = true; break;
                            case 2: dout.decode_opcode = opcode_t.ope_lsr; dout.decode_rc_we = true; break;
                            case 3: dout.decode_opcode = opcode_t.ope_asr; dout.decode_rc_we = true; break;
                            case 4: dout.decode_opcode = opcode_t.ope_mul; dout.decode_rc_we = true; break;
                            case 5: dout.decode_opcode = opcode_t.ope_addt; dout.decode_rc_we = true; break;
                            case 6: dout.decode_opcode = opcode_t.ope_subt; dout.decode_rc_we = true; break;
                            case 7: dout.decode_opcode = opcode_t.ope_muls; dout.decode_rc_we = true; break;
                            case 8: dout.decode_opcode = opcode_t.ope_movets; dout.decode_rc_we = true; break;
                            case 9: dout.decode_opcode = opcode_t.ope_and; dout.decode_rc_we = true; break;
                            case 10: dout.decode_opcode = opcode_t.ope_or; dout.decode_rc_we = true; break;
                            case 11: dout.decode_opcode = opcode_t.ope_xor; dout.decode_rc_we = true; break;
                            case 12: dout.decode_opcode = opcode_t.ope_bic; dout.decode_rc_we = true; break;
                            case 13: dout.decode_opcode = opcode_t.ope_illegal; break;
                            case 14: dout.decode_opcode = opcode_t.ope_illegal; break;
                            case 15: dout.decode_opcode = opcode_t.ope_illegal; break;
                        }

                        dout.decode_rc = (din.data_in >> 6) & 7;//extr(din.data_in, 8, 6);
                        dout.decode_rb = (din.data_in >> 3) & 7;//extr(din.data_in, 5, 3);
                        dout.decode_ra = (din.data_in) & 7;//extr(din.data_in, 2, 0);
                        dout.decode_imm = 0;
                        dout.decode_imm_a = false;
                    }
                    else // data_in(13) = '1'	-- reg, reg, imm3  (note: imm = imm3|(DIEXTENSION << 3))
                    {
                        if (((din.data_in >> 12)&1) == 0) // -- reg, reg, imm3 
                        {
                            switch ((din.data_in >> 9) & 7)//(extr(din.data_in, 11, 9))
                            {
                                // reg, reg, imm3

                                case 0: dout.decode_opcode = opcode_t.ope_addqi; dout.decode_rc_we = true; break;
                                case 1: dout.decode_opcode = opcode_t.ope_subqi; dout.decode_rc_we = true; break;
                                case 2: dout.decode_opcode = opcode_t.ope_lsrqi; dout.decode_rc_we = true; break;
                                case 3: dout.decode_opcode = opcode_t.ope_asrqi; dout.decode_rc_we = true; break;
                                case 4: dout.decode_opcode = opcode_t.ope_mulqi; dout.decode_rc_we = true; break;
                                case 5: dout.decode_opcode = opcode_t.ope_addtqi; dout.decode_rc_we = true; break;
                                case 6: dout.decode_opcode = opcode_t.ope_subtqi; dout.decode_rc_we = true; break;
                                case 7: dout.decode_opcode = opcode_t.ope_illegal; break;
                            }

                            dout.decode_rc = (din.data_in >> 6) & 7;
                            dout.decode_rb = (din.data_in >> 3) & 7;
                            dout.decode_ra = 0;
                            dout.decode_imm = (UInt16)(((din.data_in & 7) | (m_CurrentDiextended << 3)));
                            dout.decode_imm_a = true;
                        }
                        else
                        {
                            switch ((din.data_in >> 9) & 7)// reg, ref, simm3
                            {
                                // reg, reg, imm3

                                case 0: dout.decode_opcode = opcode_t.ope_edrqldi; break;
                                case 1: dout.decode_opcode = opcode_t.ope_illegal; break;
                                case 2: dout.decode_opcode = opcode_t.ope_illegal; break;
                                case 3: dout.decode_opcode = opcode_t.ope_illegal; break;
                                case 4: dout.decode_opcode = opcode_t.ope_illegal; break;
                                case 5: dout.decode_opcode = opcode_t.ope_illegal; break;
                                case 6: dout.decode_opcode = opcode_t.ope_illegal; break;
                                case 7: dout.decode_opcode = opcode_t.ope_illegal; break;
                            }

                            dout.decode_rc = 0;
                            dout.decode_rb = (din.data_in >> 6) & 7;
                            dout.decode_ra = (din.data_in >> 3) & 7;
                            dout.decode_imm = (UInt16)(((din.data_in & 7) | (m_CurrentDiextended << 3)));
                            dout.decode_imm_a = false;
                        }
                    }
                }
                else		// data(14) = '1'	
                {
                    if ((din.data_in & 0x3000) == 0x3000)  // (extr(din.data_in, 13, 12) == 3) // reg, reg / reg, imm3 / reg
                    {
                        if ((din.data_in & 0x0800) == 0)//(extr(din.data_in, 11, 11) == 0) // reg, reg / imm6
                        {
                            switch ((din.data_in >> 6) & 0x1f)//(extr(din.data_in, 10, 6))
                            {
                                case 0x00: dout.decode_opcode = opcode_t.ope_cmpeq; break;
                                case 0x04: dout.decode_opcode = opcode_t.ope_cmplo; break;
                                case 0x08: dout.decode_opcode = opcode_t.ope_cmplos; break;
                                case 0x0c: dout.decode_opcode = opcode_t.ope_espage; break;
                                case 0x10: dout.decode_opcode = opcode_t.ope_swp; dout.decode_rc_we = true; break;
                                case 0x14: dout.decode_opcode = opcode_t.ope_swptc; dout.decode_rc_we = true; break;
                                case 0x18: dout.decode_opcode = opcode_t.ope_st; break;
                                case 0x1c: dout.decode_opcode = opcode_t.ope_stinc; dout.decode_rc_we = true; break;

                                case 0x01: dout.decode_opcode = opcode_t.ope_bffo; dout.decode_rc_we = true; break;
                                case 0x05: dout.decode_opcode = opcode_t.ope_bset; dout.decode_rc_we = true; break;
                                case 0x09: dout.decode_opcode = opcode_t.ope_bclr; dout.decode_rc_we = true; break;
                                case 0x0d: dout.decode_opcode = opcode_t.ope_cmple; break;
                                case 0x11: dout.decode_opcode = opcode_t.ope_cmples; break;
                                case 0x15: dout.decode_opcode = opcode_t.ope_rqld; break;

                                case 0x19: dout.decode_opcode = opcode_t.ope_extb; dout.decode_rc_we = true; break;
                                case 0x1d: dout.decode_opcode = opcode_t.ope_stwo; break;

                                case 0x02: dout.decode_opcode = opcode_t.ope_andts; dout.decode_rc_we = true; break;
                                case 0x06: dout.decode_opcode = opcode_t.ope_orts; dout.decode_rc_we = true; break;
                                case 0x0a: dout.decode_opcode = opcode_t.ope_xorts; dout.decode_rc_we = true; break;
                                case 0x0e: dout.decode_opcode = opcode_t.ope_bicts; dout.decode_rc_we = true; break;
                                case 0x12: dout.decode_opcode = opcode_t.ope_lsrts; dout.decode_rc_we = true; break;
                                case 0x16: dout.decode_opcode = opcode_t.ope_asrts; dout.decode_rc_we = true; break;
                                case 0x1a: dout.decode_opcode = opcode_t.ope_mults; dout.decode_rc_we = true; break;
                                case 0x1e: dout.decode_opcode = opcode_t.ope_mulsts; dout.decode_rc_we = true; break;

                                // imm6

                                case 0x03: dout.decode_opcode = opcode_t.ope_erqldi; break;
                                case 0x07: dout.decode_opcode = opcode_t.ope_ssto; break;
                                case 0x0b: dout.decode_opcode = opcode_t.ope_udivinit; break;
                              
                                default: dout.decode_opcode = opcode_t.ope_illegal; break;
                            }

                            if (((din.data_in >> 6) & 0x3) == 3) // 	-- imm6
                            {
                                dout.decode_rc = 0;
                                dout.decode_rb = 0;
                                dout.decode_ra = 0;
                                dout.decode_imm = (UInt16)((din.data_in & 0x3f)  | (m_CurrentDiextended << 6));
								dout.decode_imm_a = true;
                            }
                            else // reg, reg
                            {
                                dout.decode_rc = (din.data_in >> 3) & 7; //extr(din.data_in, 5, 3);
                                dout.decode_rb = (din.data_in >> 3) & 7; // extr(din.data_in, 5, 3);
                                dout.decode_ra = (din.data_in) & 7; //extr(din.data_in, 2, 0);
                                dout.decode_imm = 0;
                                dout.decode_imm_a = false;
                            }
                        }
                        else	// data(11) = '1'
                        {
                            if((din.data_in & 0x07f0) == 0x7f0)//(extr(din.data_in, 10, 4) == 0x7f)	// no param
                            {
                                switch (din.data_in & 0xf)//(extr(din.data_in, 3, 0))
                                {
                                    case 0: dout.decode_opcode = opcode_t.ope_rqpop; break;
                                    case 1: dout.decode_opcode = opcode_t.ope_cli; break;
                                    case 2: dout.decode_opcode = opcode_t.ope_sei; break;
                                    case 3: dout.decode_opcode = opcode_t.ope_rti; break;
                                    case 4: dout.decode_opcode = opcode_t.ope_tnt; break;
                                    case 5: dout.decode_opcode = opcode_t.ope_vtt; break;
                                    case 6: dout.decode_opcode = opcode_t.ope_exl2derqld; break;
                                    case 7: dout.decode_opcode = opcode_t.ope_epushsadrl; break;
                                    case 8: dout.decode_opcode = opcode_t.ope_epushsadrh; break;
                                    case 9: dout.decode_opcode = opcode_t.ope_epopsadrl; break;
                                    case 10: dout.decode_opcode = opcode_t.ope_epopsadrh; break;
                                    case 11: dout.decode_opcode = opcode_t.ope_udivstep; break;
                                    case 12: dout.decode_opcode = opcode_t.ope_exl2den; break;                  // extension
                                    case 13: dout.decode_opcode = opcode_t.ope_exl2denerqld; break;             // extension
                                    case 14: dout.decode_opcode = opcode_t.ope_dump; break; // emulator only
                                    case 15: dout.decode_opcode = opcode_t.ope_nop; break;
                                }


                                dout.decode_rc = 0;
                                dout.decode_rb = 0;
                                dout.decode_ra = 0;
                                dout.decode_imm = 0;
                                dout.decode_imm_a = false;
                            }
                            else
                            {
                                if ((din.data_in & 0x0700) == 0x0700)// (extr(din.data_in, 10, 8) == 7) // reg		
                                {
                                    switch ((din.data_in >> 3) & 0x1f) //(extr(din.data_in, 7, 3))
                                    {
                                        case 0x00: dout.decode_opcode = opcode_t.ope_eld; dout.decode_rc_we = true; break;
                                        case 0x02: dout.decode_opcode = opcode_t.ope_push; break;
                                        case 0x04: dout.decode_opcode = opcode_t.ope_gmulhi; dout.decode_rc_we = true; break;
                                        case 0x06: dout.decode_opcode = opcode_t.ope_ld_pop; dout.decode_rc_we = true; break;
                                        case 0x08: dout.decode_opcode = opcode_t.ope_poparqp; dout.decode_rc_we = true; break;
                                        case 0x0a: dout.decode_opcode = opcode_t.ope_neg; dout.decode_rc_we = true; break;
                                        case 0x0c: dout.decode_opcode = opcode_t.ope_udivgq; dout.decode_rc_we = true; break;
                                        case 0x0e: dout.decode_opcode = opcode_t.ope_udivgr; dout.decode_rc_we = true; break;
                                        case 0x10: dout.decode_opcode = opcode_t.ope_udivgqh; dout.decode_rc_we = true; break;
                                        case 0x12: dout.decode_opcode = opcode_t.ope_udivgrh; dout.decode_rc_we = true; break;
                                        case 0x14: dout.decode_opcode = opcode_t.ope_udivsdd;  break;
                                        case 0x16: dout.decode_opcode = opcode_t.ope_udivsdv;  break;
                                        case 0x18: dout.decode_opcode = opcode_t.ope_udivsddh;  break;
                                        case 0x1a: dout.decode_opcode = opcode_t.ope_udivsdvh;  break;
                                        case 0x1c: dout.decode_opcode = opcode_t.ope_erqld; break;
                                        case 0x01: dout.decode_opcode = opcode_t.ope_setsp; break;
                                        case 0x03: dout.decode_opcode = opcode_t.ope_erqldb; break;
                                        case 0x05: dout.decode_opcode = opcode_t.ope_jmpts; break;
                                        case 0x07: dout.decode_opcode = opcode_t.ope_jmptc; break;
                                        case 0x09: dout.decode_opcode = opcode_t.ope_negts; dout.decode_rc_we = true; break;
                                        default: dout.decode_opcode = opcode_t.ope_illegal; break;

                                    }

                                    dout.decode_rc = (din.data_in ) & 7; //extr(din.data_in, 2, 0);
                                    dout.decode_rb = (din.data_in ) & 7; //extr(din.data_in, 2, 0);
                                    dout.decode_ra = (din.data_in ) & 7; //extr(din.data_in, 2, 0);
                                    dout.decode_imm = 0;
                                    dout.decode_imm_a = false;
                                }
                                else	// reg, imm3  (note: imm = imm3|(DIEXTENSION << 3))
                                {
                                    switch ((din.data_in>>8) & 7)// (extr(din.data_in, 10, 8))
                                    {
                                        case 0: dout.decode_opcode = opcode_t.ope_illegal; break;
                                        case 1: dout.decode_opcode = opcode_t.ope_illegal; break;
                                        case 2: dout.decode_opcode = opcode_t.ope_illegal; break;
                                        case 3: dout.decode_opcode = opcode_t.ope_est; break;
                                        case 4: dout.decode_opcode = opcode_t.ope_estts; break;
                                        case 5: dout.decode_opcode = opcode_t.ope_getssr; dout.decode_rc_we = true; break;
                                        case 6: dout.decode_opcode = opcode_t.ope_setssr; break;
                                        case 7: dout.decode_opcode = opcode_t.ope_illegal; break;		// can not occure
                                    }

                                    dout.decode_rc = (din.data_in>>3) & 7; //extr(din.data_in, 5, 3);
                                    dout.decode_rb = (din.data_in>>3) & 7; //extr(din.data_in, 5, 3);
                                    dout.decode_ra = 0;
                                    dout.decode_imm = (UInt16)((din.data_in & 7) | (m_CurrentDiextended << 3));
                                    dout.decode_imm_a = true;
                                }
                            }
                        }
                    }
                    else
                    {   // imm12
                        switch ((din.data_in >> 12) & 3)// (extr(din.data_in, 13, 12))
                        {
                            case 0: dout.decode_opcode = opcode_t.ope_br; break;
                            case 1: dout.decode_opcode = opcode_t.ope_brts; break;
                            case 2: dout.decode_opcode = opcode_t.ope_brtc; break;
                        }

                        dout.decode_rc = 0;
                        dout.decode_rb = 0;
                        dout.decode_ra = 0;
                        dout.decode_imm = (UInt16)(din.data_in & 0x0fff);//extr(din.data_in, 11, 0);
                        dout.decode_imm_a = true;

                    }
                }
            }
        }
    }
}
