﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components;
using ucore_emulator.core.components.stages;
using ucore_emulator.core.components.stages_hcore;
using xcore_tools;
using xcore_tools.Common;

namespace ucore_emulator.core.stages_hcore
{
    

    public class s4_execute : bitExtraction, IStage, IPrint
    {
        private IPrint m_iprint;
        private ext_bus m_bus;
        hcore_ctrl m_hcore_ctrl;
        hDCache m_dcache;
        public UInt64 m_dcacheHit = 0;
        public UInt64 m_dcacheMiss = 0;

        //

        public uint pc_reset = 0xd0100000;
        public uint pc = 0xd0100000;
        uint locked_pc = 0x0;
        uint m_hcore_id;
        public uint last;
        public bool t;
        public bool idle = false;

        //

        uint pa_handover_address = 0;
        uint pa_jump_address = 0; 
        uint pa_max_cores = 0;
        uint pa_current_core = 0;
        uint m_total_core_count = 0;

        hcore_thread_ctrl m_thread_ctrl;

        // debug

        public UInt64 skipCount = 0;
        public UInt64 cycleCount = 0;

        // reg offset

        uint m_roffset;

        // div steper

        uint udiv_remainder = 0;
        uint udiv_qoutient = 0;
        uint udiv_divisor = 0;
        uint udiv_divident = 0;
        uint udiv_counter = 0;

        //

        delay<Int64> m_mulDelay = new delay<Int64>(2 + 1);

     /*   struct bus_action
        {
            public bool we;
            public bool mode;
            public uint address;
            public uint data;

        };*/

     //   Queue<bus_action> m_busRequestQueue = new Queue<bus_action>();
     //   Queue<UInt32> m_loadQueue = new Queue<UInt32>();

        // out of order

            // memory

        enum ooMemoryCodes
        {
            oomc_ld,
            oomc_ldw,
            oomc_ldws,
            oomc_ldc,
            oomc_st,
            oomc_stw
        }

        struct ooToMemory
        {
           public ooMemoryCodes code;
           public uint address;
           public uint data;
           public uint rc;          
        }

        struct ooFromMemory
        {
            public uint data;
            public uint rc;          
        }

        Queue<ooToMemory> m_fifoToMemory = new Queue<ooToMemory>(256);
        Queue<ooFromMemory> m_fifoFromMemory = new Queue<ooFromMemory>(256);

            // uctrl/ucore

        public enum ooUctrlCodes
        {
            ootu_nop,
            ootu_setparam,
            ootu_getparam,
            ootu_call,
            ootu_semst,
            ootu_semld,
            ootu_seminc,
            ootu_semdec,
            ootu_allocsem,
            ootu_freesem
        }

        public struct ooToUctrl
        {
            public bool doReturn;
            public uint id;
            public uint code;
            public uint address;
            public uint data;
        }

        public struct ooUctrlDestHold
        {
            public uint rc;
        }

        public struct ooFromUctrl
        {
            public uint data;
        }

        /* fifo is shared over all hcore threads !!!*/

        static object m_fifoToUctrlLock = new object();
        static Queue<ooToUctrl> m_fifoToUctrl = new Queue<ooToUctrl>(256);

        //

        object m_fifoFromUctrlLock = new object();
        Queue<ooFromUctrl> m_fifoFromUctrl = new Queue<ooFromUctrl>(256);
        Queue<ooUctrlDestHold> m_fifoUctrlDestHold = new Queue<ooUctrlDestHold>(256);

        // external access

        public bool hasFifoToUctrlData
        {
            get
            {
                bool rValue = false;

                if (m_fifoToUctrlLock != null)
                {
                    lock (m_fifoToUctrlLock)
                    {
                        if (m_fifoToUctrl != null)
                        {
                            rValue = m_fifoToUctrl.Count > 0;
                        }
                    }
                }

                return rValue;
            }
        }

        public ooToUctrl popFifoToUctrlData()
        {
            ooToUctrl rValue;
            lock (m_fifoToUctrlLock)
            {
                rValue = m_fifoToUctrl.Dequeue();
            }
            return rValue;
        }

        public void pushFifoFromUctrl(ooFromUctrl from)
        {
            lock (m_fifoFromUctrlLock)
            {
                m_fifoFromUctrl.Enqueue(from);
            }
        }

        //

        public hc_s4_out_s5_in dataOut
        {
            get { return m_out; }
        }

        hc_s4_out_s5_in m_out = new hc_s4_out_s5_in();


        public s4_execute(ext_bus bus, IPrint iprint, uint ihcore_id, hcore_thread_ctrl thread_ctrl, uint total_core_count, hcore_ctrl hcore_ctrl,hDCache dcache)
        {         
            m_dcache = dcache;
            m_iprint = iprint;
            m_bus = bus;
            m_hcore_id = ihcore_id;
            idle = false;
            m_thread_ctrl = thread_ctrl;
            m_total_core_count = total_core_count;
            m_hcore_ctrl = hcore_ctrl;
        }

        public void reset()
        {
            m_fifoUctrlDestHold.Clear();
            m_fifoFromMemory.Clear();
            m_fifoFromUctrl.Clear();
            m_fifoToMemory.Clear();

            lock (m_fifoToUctrlLock)
            {
                m_fifoToUctrl.Clear();
            }

            m_dcacheHit = 0;
            m_dcacheMiss = 0;

            pc = pc_reset;
            idle = false;
            t = false; 

            skipCount = 0;
            cycleCount = 0;

            m_roffset = 0;
        }

        internal void process(bool lockPc, components.stages_hcore.hc_s3_out_s4_in din, out components.stages_hcore.hc_s4_out_s5_in dout, bool reset, out bool isIdle,out bool regLock,out bool dump)
        {
            dump = false;

            cycleCount++;

            // check config (from uctrl)

            while(m_hcore_ctrl.hasData(m_hcore_id))
            {
                ctrl_struct cs = m_hcore_ctrl.getData(m_hcore_id);
                switch(cs.address)
                {
                    case 0: pc = (pc & 0x0000ffff) | (((UInt32)cs.data) << 16); pc_reset = pc; break; // hi
                    case 1: pc = (pc & 0xffff0000) | (((UInt32)cs.data)); pc_reset = pc; break;     // lo
                    default: throw new Exception("HCore " + m_hcore_id + " got illegal config data\n");
                }
            }

            dout = m_out;

            //debug

            dout.debug_pc = din.debug_pc;

            //

            uint a = (din.a_as_reg_out) ? din.ra_data : din.imm_out;
            uint b = (din.b_as_reg_out) ? din.rb_data : last; 

            // default

            regLock = false;
            idle = false;
            dout.rc_we = false;          
            dout.rc = din.rc_through_out;
            dout.rc_unlock = false;
            dout.en_roffset_en = false;
            dout.roffset = m_roffset;

            uint tpc = pc;

            // calculate (note: optimzed nearly single instructions on emulator)

            uint res_add = a + b;
            UInt64 res_mul = (UInt64)m_mulDelay.process(((Int64)b) * ((Int64)a));

            if (reset == false)
            {
               // if(stallIn == false)
                {


                    // source register(s) not locked


                    // pc

                    tpc = pc + 2;

                    // reg we

                    if (((din.condition_out == hcondition_t.hcc_ever) ||
                                (din.condition_out == hcondition_t.hcc_ts && t == true) ||
                                (din.condition_out == hcondition_t.hcc_tc && t == false)) &&
                                din.c_as_reg_out == true)
                    {
                        dout.rc_we = din.rc_we_through_out; // if condition is true and c is a reg then push rc_we through											
                    }

                    // test condition 

                    if ((din.condition_out == hcondition_t.hcc_ever) ||
                                (din.condition_out == hcondition_t.hcc_ts && t == true) ||
                                (din.condition_out == hcondition_t.hcc_tc && t == false))
                    {
                        switch (din.opcode_out)
                        {
                            case hopcode_t.hope_skip: tpc = pc; skipCount++; break; //read register is locked
                            case hopcode_t.hope_nop:    break;
                            case hopcode_t.hope_dump: dump = true; break; /* only emulator */

                            // in order

                            case hopcode_t.hope_move: last = a; break;
                            case hopcode_t.hope_moveh: last = a << 16 | (b & 0xffff); break;
                            case hopcode_t.hope_add: last = res_add; break;
                            case hopcode_t.hope_sub: last = b - a; break;
                            case hopcode_t.hope_rsub: last = a - b; break;
                            case hopcode_t.hope_and: last = a & b; break;
                            case hopcode_t.hope_or: last = a | b; break;
                            case hopcode_t.hope_xor: last = a ^ b; break;
                            case hopcode_t.hope_bic: last = b & ~a; break;
                            case hopcode_t.hope_asr: last = (uint)(((int)b) >> (int)a); break;
                            case hopcode_t.hope_lsr: last = b >> (int)a; break;
                            case hopcode_t.hope_lsl: last = b << (int)a; break;
                            case hopcode_t.hope_cmplo: t = b < a; break;
                            case hopcode_t.hope_cmplos: t = ((int)b) < ((int)a); break;
                            case hopcode_t.hope_cmple: t = b <= a; break;
                            case hopcode_t.hope_cmples: t = ((int)b) <= ((int)a); break;
                            case hopcode_t.hope_cmpeq: t = b == a; break;
                            case hopcode_t.hope_br: tpc = pc + (a & 0xfffffffe); break;
                            case hopcode_t.hope_jmp: tpc = res_add & 0xfffffffe; break;
                            case hopcode_t.hope_gpc: last = pc + (a & 0xfffffffe); break;

                            case hopcode_t.hope_gssr:
                                {
                                    switch (a)
                                    {
                                        case 0: last = m_hcore_id; break;
                                        case 1: last = m_total_core_count; break;
                                        case 3: last = m_roffset & 0x7f; break;
                                        default: last = a & 0xf; break;
                                    }
                                }
                                break;

                            case hopcode_t.hope_sssr:
                                {
                                    switch ((uint)(a) & 0xf)
                                    {
                                        case 3: m_roffset = b & 0x7f;
                                                dout.roffset = m_roffset; // update to out singals
                                                break;
                                        // bla = b;
                                    }
                                }
                                break;

                            case hopcode_t.hope_roe:
                                dout.en_roffset_num = b & 0x7f;
								dout.en_roffset_data = (a & 1) == 1;
								dout.en_roffset_en	= true;
                                break;

                          /*  case hopcode_t.hope_udivinit:
                                {
                                    udiv_remainder = 0;
                                    udiv_qoutient = 0;
                                    udiv_divisor = a;
                                    udiv_divident = b;
                                    udiv_counter = din.imm_out & 0x1f;
                                }
                                break;


                            case hopcode_t.hope_udivstep:
                                {
                                    UInt32 udiv_temp = (udiv_remainder * 2) | extr(udiv_divident, (int)udiv_counter);
                                    if (udiv_temp >= udiv_divisor)
                                    {
                                        udiv_temp -= udiv_divisor;
                                        udiv_qoutient |= 1U << (int)udiv_counter;
                                    }
                                    udiv_remainder = udiv_temp;
                                    udiv_counter--;

                                    last = udiv_qoutient;
                                }
                                break;

                            case hopcode_t.hope_udivgrem:
                                {
                                    last = udiv_remainder;
                                }
                                break;
                                */
                            // delayed in order

                            case hopcode_t.hope_mul: last = (uint)(res_mul & 0xffffffff); break;
                            case hopcode_t.hope_mulh: last = (uint)(res_mul >> 32); break;

                            // out of order

                            // MEMORY

                            // with replay

                            case hopcode_t.hope_ld:
                            case hopcode_t.hope_ldc:
                            case hopcode_t.hope_ldw:
                            case hopcode_t.hope_ldws:
                                {
                                    ooToMemory ba = new ooToMemory();

                                    ba.address = res_add;
                                    ba.data = 0;

                                    if (din.opcode_out == hopcode_t.hope_ld)
                                    {
                                        ba.code = ooMemoryCodes.oomc_ld;
                                    }
                                    else if (din.opcode_out == hopcode_t.hope_ldc)
                                    {
                                        UInt32 cacheRValue = 0;
                                        if (m_dcache.read(ba.address, out cacheRValue) == true)
                                        {
                                            m_dcacheHit++;
                                        }
                                        else
                                        {
                                            m_dcacheMiss++;
                                        }

                                        ba.code = ooMemoryCodes.oomc_ldc;
                                    }
                                    else if (din.opcode_out == hopcode_t.hope_ldw)
                                    {
                                        ba.code = ooMemoryCodes.oomc_ldw;
                                    }
                                    else if (din.opcode_out == hopcode_t.hope_ldws)
                                    {
                                        ba.code = ooMemoryCodes.oomc_ldws;
                                    }

                                   
                                    ba.rc = din.rc_through_out;

                                    m_fifoToMemory.Enqueue(ba);
                                }
                                break;

                            // no RCWE

                            case hopcode_t.hope_st:
                            case hopcode_t.hope_stc:
                            case hopcode_t.hope_stw:
                                {

                                    ooToMemory ba = new ooToMemory();

                                    ba.address = (uint)(a + exts(din.imm_out, 6));
                                    ba.data = b;
                                    ba.rc = 0;

                                    if (din.opcode_out == hopcode_t.hope_st) 
                                    {
                                        ba.code = ooMemoryCodes.oomc_st;
                                    }
                                    else if (din.opcode_out == hopcode_t.hope_stc) 
                                    {
                                        ba.code = ooMemoryCodes.oomc_st;

                                        m_dcache.write(ba.address, ba.data); // write to dcache
                                    }
                                    else if (din.opcode_out == hopcode_t.hope_stw)
                                    {
                                        ba.code = ooMemoryCodes.oomc_stw;
                                    }

                                

                                    m_fifoToMemory.Enqueue(ba);

                                }
                                break;

                            // uctrl/ucore

                            case hopcode_t.hope_uexec:
                            case hopcode_t.hope_uexecs:

                                ooToUctrl ot = new ooToUctrl();


                                if(din.opcode_out == hopcode_t.hope_uexec)  // no return
                                {
                                    ot.doReturn = false;
                                }
                                else if (din.opcode_out == hopcode_t.hope_uexecs)  // with return
                                {
                                    ot.doReturn = true;
                                    ooUctrlDestHold oh = new ooUctrlDestHold();
                                    oh.rc = din.rc_through_out;
                                    m_fifoUctrlDestHold.Enqueue(oh);
                                }
                                else
                                {
                                    throw new Exception("hcore:execute Error: illegal opcode");
                                }

                                ot.id = m_hcore_id;             // 8 bit
                                ot.data = a;                    // 16 bit
                                ot.address = extr(b, 15, 8);    // 8 bit
                                ot.code = extr(b,7,0);          // 8 bit
                               
                                lock (m_fifoToUctrlLock)
                                {
                                    m_fifoToUctrl.Enqueue(ot);
                                }
                          
                                break;

                            // parallel

                            /*  case hopcode_t.hope_rake:
                                  {
                                      hcore_thread_request req = new hcore_thread_request();

                                      req.handover_address = pa_handover_address;
                                      req.jump_address = pa_jump_address;
                                      req.max_cores = pa_max_cores;

                                      pa_max_cores = m_thread_ctrl.parallelRequest(req);
                                      pa_current_core = 0;

                                      tpc = pa_jump_address;

                                      last = pa_max_cores;
                                  }
                                  break;*/

                            /* case hopcode_t.hope_spa:
                                 {
                                     switch (a)
                                     {
                                         case 0: pa_handover_address = b; break;
                                         case 1: pa_jump_address = b; break;
                                         case 2: pa_max_cores = b; break;
                                         default: break;
                                     }
                                 }
                                 break; */

                            /*  case hopcode_t.hope_gpa:    
                                   {
                                      switch (a)
                                      {
                                          case 0: last = pa_handover_address; break;
                                          case 1: last = pa_jump_address; break;
                                          case 2: last = pa_max_cores; break;
                                          case 3: last = pa_current_core; break;
                                          default: break;
                                      }
                                  }
                                  break;  */

                            case hopcode_t.hope_idle:
                                {
                                    bool available;
                                    hcore_thread_response res = m_thread_ctrl.popResponse(m_hcore_id, out available);
                                    if (available == true)
                                    {
                                        pa_handover_address = res.handover_address;
                                        pa_current_core = res.current_core;
                                        pa_jump_address = res.jump_address;
                                        pa_max_cores = res.max_cores;

                                        tpc = pa_jump_address;
                                    }
                                    else
                                    {
                                        tpc = pc; // dont cycle pc here (not important one hardware)
                                        idle = true;    // idle
                                    }
                                }
                                break;



                            default: break;
                        }
                    }

                  /*  if (m_hcore_id == 0)
                    {
                        dump = true;
                    }*/

                    dout.rc_data = last;
                }

                

                //--------------------------------------------------------------------------
                // out of order
                //--------------------------------------------------------------------------

                // comming in

                if (dout.rc_we == false) // only write if no 'we out is ongoing (in order instruction)
                {
                    // from memory fifo

                    memoryRecive(dout);

                    // from uctrl fifo

                    uctrlRecive(dout);
                }

                // memory to fifo abarbeiten

                memorySend(ref dump);
             }
           

            // tc

         //   m_thread_ctrl.setCoreIdle(m_hcore_id, idle);

            //

            isIdle = idle;

            
            pc = tpc;
            dout.pc_out = tpc;
            
            
        }

        private void uctrlRecive(components.stages_hcore.hc_s4_out_s5_in dout)
        {
            if (m_fifoFromUctrl.Count != 0)
            {
                if (m_fifoUctrlDestHold.Count == 0)
                {
                    throw new Exception("hcore:execute Error: got incomming data from uctrl but no destination is available!");
                }

                ooFromUctrl from = m_fifoFromUctrl.Dequeue();       // data
                ooUctrlDestHold rh = m_fifoUctrlDestHold.Dequeue(); // rc

                dout.rc_data = from.data;
                dout.rc = rh.rc;
                dout.rc_we = true;
                dout.rc_unlock = true;
            }
        }

        private void memoryRecive(components.stages_hcore.hc_s4_out_s5_in dout)
        {
            if (m_fifoFromMemory.Count != 0) // incoming ??
            {
                ooFromMemory ba = m_fifoFromMemory.Dequeue();

                dout.rc_data = ba.data;
                dout.rc = ba.rc;
                dout.rc_we = true;
                dout.rc_unlock = true;
            }
        }

        private void memorySend(ref bool dump)
        {
            while (m_fifoToMemory.Count != 0)
            {
                ooToMemory oo = m_fifoToMemory.Dequeue();

                switch (oo.code)
                {
                    // store

                    case ooMemoryCodes.oomc_st:
                        {
                            m_bus.write(oo.address, oo.data);
                                                      
                            //print(string.Format("st {0:X8} <- {1:X8}\n", oo.address, oo.data));
                        }
                        break;

                    case ooMemoryCodes.oomc_stw:
                        {
                            UInt16 da = (UInt16)oo.data;
                            m_bus.write(oo.address, da);

                           // print(string.Format("stw {0:X8} <- {1:X4}\n",oo.address,da));
                        }
                        break;

                    // load

                    case ooMemoryCodes.oomc_ld:
                        {
                            UInt32 data;
                            m_bus.read(oo.address, out data);

                            ooFromMemory ba = new ooFromMemory();

                            ba.data = data;
                            ba.rc = oo.rc;

                            m_fifoFromMemory.Enqueue(ba);
                        }
                        break;

                    case ooMemoryCodes.oomc_ldc:
                        {
                            UInt32 data;
                            m_bus.read(oo.address, out data);

                            m_dcache.write(oo.address, data); // write to dcache after read

                            ooFromMemory ba = new ooFromMemory();

                            ba.data = data;
                            ba.rc = oo.rc;

                            m_fifoFromMemory.Enqueue(ba);
                        }
                        break;
                    case ooMemoryCodes.oomc_ldw:
                    case ooMemoryCodes.oomc_ldws:
                        {
                            UInt16 data;
                            m_bus.read(oo.address, out data);

                            ooFromMemory ba = new ooFromMemory();

                            if (oo.code == ooMemoryCodes.oomc_ldw)
                            {
                                ba.data = ((UInt32)data);
                            }
                            else if (oo.code == ooMemoryCodes.oomc_ldws)
                            {
                                Int32 sdata = (Int32)data;
                                sdata <<= 16;
                                sdata >>= 16;
                                ba.data = (UInt32)sdata;
                            }

                            ba.rc = oo.rc;

                            m_fifoFromMemory.Enqueue(ba);
                        }
                        break;
                }
            }
        }
            
        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }
}
