﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Text.RegularExpressions;

namespace xcore_tools.Common
{
    public class UInt128
    {
        public UInt128()
        {
            l = h = 0;
        }

        public UInt128(UInt64 v)
        {
            l = v;
            h = 0;
        }

        public static UInt128 operator |(UInt128 a, UInt128 b)
        {
            UInt128 r = new UInt128();

            r.l = a.l | b.l;
            r.h = a.h | b.h;

            return r;
        }

        public static bool operator ==(UInt128 a,UInt128 b)
        {
            return a.l == b.l && a.h == b.h;
        }

        public static bool operator !=(UInt128 a, UInt128 b)
        {
            return a.l != b.l || a.h != b.h;
        }

        public static implicit operator UInt128(UInt64 a)
        {
            UInt128 r = new UInt128();

            r.l = a;
            r.h = 0;

            return r;
        }

        public static implicit operator UInt32(UInt128 a)
        {
            return (UInt32)a.low;
        }

        public static UInt128 operator &(UInt128 a, UInt128 b)
        {
            UInt128 r = new UInt128();

            r.l = a.l & b.l;
            r.h = a.h & b.h;

            return r;
        }

        public static UInt128 operator <<(UInt128 a, int sft)
        {
            UInt128 r = new UInt128();

        

            if (sft < 64)
            {
                r.l = a.l << sft;
            }
            else
            {
                r.l = 0;
            }

            if (sft < 64)
            {
                r.h = a.h << sft;
            }
            else
            {
                r.h = 0;
            }

            if ((64 - sft) < 64)   // UInt64 logical bug ....
            {
                if ((64 - sft) < 0)
                {
                    r.h |= a.l << -(64 - sft);
                }
                else
                {
                    r.h |= a.l >> (64 - sft);
                }
            }           

            return r;
        }

        public static UInt128 operator >>(UInt128 a, int sft)
        {
            UInt128 r = new UInt128();

            if (sft < 64)
            {
                r.l = a.l >> sft;
            }
            else
            {
                r.l = 0;
            }

            if ((64 - sft) < 64)   // UInt64 logical bug ....
            {
                if ((64 - sft) < 0)
                {
                    r.l |= a.h >> -(64 - sft);
                }
                else
                {
                    r.l |= a.h << (64 - sft);
                }
            }

            if (sft < 64)
            {
                r.h = a.h >> sft;
            }
            else
            {
                r.h = 0;
            }

            return r;
        }
       
        public UInt64 low
        {
            get { return l; }
            set { l = value; }
        }

        public UInt64 high
        {
            get { return h; }
            set { h = value; }
        }

        UInt64 l;
        UInt64 h;
    }

    public class Tools
    {
        public static int extractInteger(string txt)
        {
            int dataBase = 10;

            if(txt.Length > 1)
            { 
                if (txt[0] == '$')   // hex check
                {
                    txt = txt.Remove(0, 1);
                    dataBase = 16;
                }
                else if (txt[0] == '%')   // bin check
                {
                    txt = txt.Remove(0, 1);
                    dataBase = 2;
                }
                else if (txt[0] == '\'')   // char
                {
                    if (txt.Length > 2)
                    {
                        throw new Exception("error: accept only one char after '\n");
                    }
                    return (int) txt[1];
                }
            }

            return Convert.ToInt32(txt, dataBase);
        }
    }

    public class hiresCounter
    {
        private long m_startTime, m_stopTime;
        private long m_freq;

        // Constructor

        public hiresCounter()
        {
            m_startTime = 0;
            m_stopTime = 0;
            m_freq = Stopwatch.Frequency;
        }

        public void start()
        {
            m_startTime = current();
        }

        public void stop()
        {
            m_stopTime = current();
        }


        public double elapsed()
        {
            return calculate(m_stopTime, m_startTime);
        }

        // special use

        public double calculate(long s1, long s2)
        {
            return (double)(s1 - s2) / (double)frequency();
        }

        public long current()
        {
            return Stopwatch.GetTimestamp(); ;
        }

        public long frequency()
        {
            return m_freq;
        }
    }

    public class parsePattern<PATTERN_TYPE>
    {

        public parsePattern(Regex iregex, PATTERN_TYPE ipatternType)
        {
            regex = iregex;
            patternType = ipatternType;

        }

        public Regex regex;
        public PATTERN_TYPE patternType;
    }

    public class bitExtraction
    {
        static protected int exts(UInt16 din, int w)
        {
            int temp = 31 - w;
            int result = din;

            result <<= temp;
            result >>= temp;

            return result;
        }

        static protected int exts(UInt32 din, int w)
        {
            int temp = 31 - w;
            int result = (int)din;

            result <<= temp;
            result >>= temp;

            return result;
        }

        static protected Int64 exts(UInt64 din, int w)
        {
            int temp = 63 - w;
            Int64 result = (Int64)din;

            result <<= temp;
            result >>= temp;

            return result;
        }

        static protected UInt16 extr(UInt16 din, int e, int s)
        {
            // 0101010101, 5,2

            int temp = 15 - e;

            din = (UInt16)(din << temp);
            din = (UInt16)(din >> (temp + s));

            return din;
        }

        static protected UInt16 extr(UInt16 din, int e)
        {
            return extr(din, e, e);
        }

        static protected UInt32 extr(UInt32 din, int e, int s)
        {
            int temp = 31 - e;

            din = (UInt32)(din << temp);
            din = (UInt32)(din >> (temp + s));

            return din;
        }

        static protected UInt32 extr(UInt32 din, int e)
        {
            return extr(din, e, e);
        }

        static protected UInt64 extr(UInt64 din, int e, int s)
        {
            int temp = 63 - e;

            din = (UInt64)(din << temp);
            din = (UInt64)(din >> (temp + s));

            return din;
        }

        static protected UInt64 extr(UInt64 din, int e)
        {
            return extr(din, e, e) & 1;
        }

        static protected UInt128 extr(UInt128 din, int e, int s)
        {
            int temp = 127 - e;

            din = din << temp;
            din = din >> (temp + s);

            return din;
        }

        static protected UInt128 extr(UInt128 din, int e)
        {
            return extr(din, e, e) & 1;
        }
    }

    public class delay<T>
    {
        T[] buffer;
        uint counter = 0;
        uint length = 0;

        public delay(uint delayLength)
        {
            buffer = new T[delayLength];
            counter = 0;
            length = delayLength;
        }

        public T process(T data)
        {
            buffer[counter] = data;
            counter++;
            counter %= length;
            return buffer[counter]; 

        }
    }
}
