LIBRARY ieee;
USE ieee.std_logic_1164.all;
use IEEE.MATH_REAL.ALL;
LIBRARY altera_mf;
USE altera_mf.all;


ENTITY dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END dc_fifo;


ARCHITECTURE rtl OF dc_fifo IS

	COMPONENT dcfifo
	GENERIC (
		intended_device_family		: STRING;
		lpm_numwords		: NATURAL;
		lpm_showahead		: STRING;
		lpm_type		: STRING;
		lpm_width		: NATURAL;
		lpm_widthu		: NATURAL;
		overflow_checking		: STRING;
		rdsync_delaypipe		: NATURAL;
		read_aclr_synch		: STRING;
		underflow_checking		: STRING;
		use_eab		: STRING;
		write_aclr_synch		: STRING;
		wrsync_delaypipe		: NATURAL
	);
	PORT (
			rdclk	: IN STD_LOGIC ;
			q	: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
			rdempty	: OUT STD_LOGIC ;
			wrclk	: IN STD_LOGIC ;
			wrreq	: IN STD_LOGIC ;
			wrusedw	: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0);
			aclr	: IN STD_LOGIC ;
			data	: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
			rdreq	: IN STD_LOGIC 
	);
	END COMPONENT;

BEGIN

	dcfifo_component : dcfifo
	GENERIC MAP (
		intended_device_family => "Cyclone IV E",
		lpm_numwords => FIFO_DEPTH,
		lpm_showahead => FIFO_SHOWAHEAD,
		lpm_type => "dcfifo",
		lpm_width => FIFO_WIDTH,
		lpm_widthu => FIFO_USEDWWIDTH,
		overflow_checking => "OFF",
		rdsync_delaypipe => 2+FIFO_SYNCSTAGES,
		read_aclr_synch => "OFF",
		underflow_checking => "OFF",
		use_eab => "ON",
		write_aclr_synch => "ON",
		wrsync_delaypipe => 2+FIFO_SYNCSTAGES
	)
	PORT MAP (
		rdclk => rdclk,
		wrclk => wrclk,
		wrreq => wrreq,
		aclr => aclr,
		data => data,
		rdreq => rdreq,
		q => q,
		rdempty => rdempty,
		wrusedw => wrusedw
	);

END rtl;

