library ieee;
use work.bus_package.all;
use work.hcore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_dbuffer is
		generic
		(
			LINEWIDTH : integer := 6
		);
		port 
		(
				-- global

				clk					: in std_logic;
				reset					: in std_logic;
				clear					: in std_logic;					
				
				-- write a
				
				i_adr_a				: in std_logic_vector(30 downto 0);
				i_data_a				: in std_logic_vector(31 downto 0);
				i_data_a_we			: in std_logic;
				
				-- write b
				
				i_adr_b				: in std_logic_vector(30 downto 0);
				i_data_b				: in std_logic_vector(31 downto 0);
				i_data_b_we			: in std_logic;
				
				-- read
				
				i_read_adr			: in std_logic_vector(30 downto 0);
				o_read_data			: out std_logic_vector(31 downto 0);
				o_read_data_dok	: out std_logic
		
		);
end entity;

architecture rtl of hcore_dbuffer is

	type data_t is array(0 to (2**LINEWIDTH)-1) of std_logic_vector(31 downto 0);
	signal data : data_t;
	
	type valid_t is array(0 to (2**LINEWIDTH)-1) of std_logic;
	signal valid : valid_t;
	
	signal current_tag : std_logic_vector(30-LINEWIDTH downto 0) := (others => '0');
	
	-- extract functions
	
	function extractPosition(din : std_logic_vector(30 downto 0))
	return std_logic_vector is
	begin 
		return din(LINEWIDTH-1 downto 0);
	end;
	
	function extractTag(din : std_logic_vector(30 downto 0))
	return std_logic_vector is
	begin 
		return din(30 downto LINEWIDTH);
	end;

	--

begin

	process(clk)
	
		variable position : integer range 0 to (2**LINEWIDTH)-1;
		variable tag      : std_logic_vector(30-LINEWIDTH downto 0);
		
	
	begin
		if rising_edge(clk) then
			if reset = '1' or clear = '1' then
				current_tag <= (others => '0');
			else
			
				-- at write port a will win
			
				o_read_data_dok <= '0';
			
				if i_data_a_we = '1' then
				
					position := conv_integer(extractPosition(i_adr_a));
					tag		:= extractTag(i_adr_a);
				
					data(position)  <= i_data_a;
					
					if(current_tag /= tag) then						
						valid <= (others => '0');
					end if;
					
					valid(position) <= '1';
						
					current_tag <= tag;
					
					
				elsif i_data_b_we = '1' then
				
					position := conv_integer(extractPosition(i_adr_b));
					tag		:= extractTag(i_adr_b);
				
					data(position) <= i_data_b;
					
					if(current_tag /= tag) then						
						valid <= (others => '0');
					end if;
					
					valid(position) <= '1';
					
					current_tag <= tag;
				end if;
				
					
				position := conv_integer(extractPosition(i_read_adr));
				tag		:= extractTag(i_read_adr);
					
					
				if current_tag = tag and valid(position) = '1' then
					o_read_data_dok <= '1';
				end if;

				o_read_data <= data(position);
								
			end if;
		end if;
	end process;
	
end rtl;
