library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

package hcore_package is

type hopcode_oo_t is	-- out of order
(		
	  hope_oo_st,
	  hope_oo_stw,
	  hope_oo_uexec,
	  hope_oo_ld,
	  hope_oo_ldw,
	  hope_oo_ldws,
	  hope_oo_uexecs,
	  hope_oo_ldc,
     hope_oo_stc,
	  hope_oo_nop,
	  hope_oo_illegal
);

type hopcode_t is -- in order
( 
	  hope_move,
	  hope_moveh,
	  hope_add,
	  hope_sub,
	  hope_rsub,
	  hope_cmplo,
	  hope_cmplos,
	  hope_cmple,
	  hope_cmples,
	  hope_cmpeq,
	  hope_jmp,
	  hope_br,
	  hope_gpc,
	  hope_and,
	  hope_or,
	  hope_xor,
	  hope_bic,
	  hope_mul,
	  hope_mulh,
	  hope_lsr,
	  hope_lsl,
	  hope_asr,
	  hope_idle,
	  hope_dump,
	  hope_nop,
	  hope_gssr,
	  hope_sssr,
	  hope_roe,
 
	  -- internal
	  
	  hope_illegal
);

type hcondition_t is
(
	hcc_ever,
	hcc_ts,
	hcc_tc,
	hcc_never
);

type hmemory_codes_t is
(
	hmc_ld,
	hmc_ldc,
	hmc_ldw,
	hmc_ldws
);

-- memory path

type hcore_memory_loop_struct_t is record

	code		: hmemory_codes_t;					-- 2						
	rc			: std_logic_vector(6 downto 0);	-- 7
	address  : std_logic_vector(31 downto 0);	-- 32	
															-- 41
end record;

-- dbuffer queue

type hcore_memory_dbuffer_queue_struct_t is record
				
	rc			: std_logic_vector(6 downto 0);	-- 7
	data	   : std_logic_vector(31 downto 0);	-- 32	
															-- 39
end record;

-- uctrl

type hcore_uctrl_loop_struct_t is record
					
	rc			: std_logic_vector(6 downto 0);	-- 7
															-- 7
end record;

type hcore_to_uctrl_struct_t is record
			
	do_return	: std_logic;						   -- 1		
	hcore_id 	: std_logic_vector(7 downto 0);	-- 8		
	data		 	: std_logic_vector(31 downto 0);	-- 32	
	address  	: std_logic_vector(7 downto 0);	-- 8	
	code		 	: std_logic_vector(7 downto 0);	-- 8	
																-- 57
end record;

type uctrl_to_hcore_struct_t is record
				
	hcore_id 	: std_logic_vector(7 downto 0);	-- 8		
	data		 	: std_logic_vector(31 downto 0);	-- 32		
																-- 40
end record;

-- uctrl to hcores

type hcore_ctrl_struct_t is record
		
		 hcore_sel 	: std_logic_vector(7 downto 0);	-- 8
		 address 	: std_logic_vector(15 downto 0);	-- 16
		 data			: std_logic_vector(15 downto 0);	-- 16
																   -- = 40
end record;

type	hcore_ctrls_t			is array(0 to hcore_count-1) of hcore_ctrl_struct_t;
type	hcore_wes_t				is array(0 to hcore_count-1) of std_logic;

-- to uctrl (s)

type hcore_to_uctrl_structs_t	is array(0 to hcore_count-1) of hcore_to_uctrl_struct_t;
type hctrl_emptys					is array(0 to hcore_count-1) of std_logic;
type hctrl_rdreqs					is array(0 to hcore_count-1) of std_logic;

type uctrl_to_hcore_structs_t	is array(0 to hcore_count-1) of uctrl_to_hcore_struct_t;
type uctrl_fulls					is array(0 to hcore_count-1) of std_logic;
type uctrl_wrreqs					is array(0 to hcore_count-1) of std_logic;

-- inst arb

type hcore_send_data_t		is array(0 to hcore_count-1) of std_logic_vector(31 downto 0);
type hcore_send_address_t	is array(0 to hcore_count-1) of std_logic_vector(31 downto 0);
type hcore_send_we_t			is array(0 to hcore_count-1) of std_logic;
type hcore_send_dmode_t		is array(0 to hcore_count-1) of std_logic;
type hcore_send_busy_t		is array(0 to hcore_count-1) of std_logic;
type hcore_send_wrreq_t		is array(0 to hcore_count-1) of std_logic;

type hcore_recive_data_t	is array(0 to hcore_count-1) of std_logic_vector(31 downto 0);
type hcore_recive_busy_t	is array(0 to hcore_count-1) of std_logic;
type hcore_recive_ack_t		is array(0 to hcore_count-1) of std_logic;		
		
-- id generate

type use_id_t 	is array(0 to hcore_count-1) of integer range 0 to hcore_count-1;		
		
-- hcore id

type hcore_ids_t	is array(0 to hcore_count-1) of std_logic_vector(3 downto 0);	
type hcore_stalls_t	is array(0 to hcore_count-1) of std_logic;	
		
-- thread ctrl


--type thread_request_struct is record
--		
--		 request						: std_logic;
--		 max_cores 					: integer range 0 to hcore_count-1;
--		 handover_address 		: std_logic_vector(31 downto 0);	-- 32 
--		 jump_address				: std_logic_vector(31 downto 0);	-- 32 
--		 
--end record;
--
--type thread_accepted_struct is record
--		
--		 ready						: std_logic;
--		 max_cores 					: integer range 0 to hcore_count-1;
--		 current_core_number 	: integer range 0 to hcore_count-1;
--		 handover_address 		: std_logic_vector(31 downto 0);	-- 32 
--		 jump_address				: std_logic_vector(31 downto 0);	-- 32 
--		 
--end record;
--
--type thread_requests_t  is array(0 to hcore_count-1) of thread_request_struct;
--type thread_accepteds_t is array(0 to hcore_count-1) of thread_accepted_struct;
--type idles_t is array(0 to hcore_count-1) of std_logic;

	-- function declaration
	
	function hcore_ctrl_to_raw(din : hcore_ctrl_struct_t) return std_logic_vector;
	function raw_to_hcore_ctrl(din : std_logic_vector(39 downto 0)) return hcore_ctrl_struct_t;

	function memory_loop_to_raw(din : hcore_memory_loop_struct_t) return std_logic_vector;
	function raw_to_memory_loop(din : std_logic_vector(47 downto 0)) return hcore_memory_loop_struct_t;
	
	function dbuffer_queue_to_raw(din : hcore_memory_dbuffer_queue_struct_t) return std_logic_vector;
	function raw_to_dbuffer_queue(din : std_logic_vector(39 downto 0)) return hcore_memory_dbuffer_queue_struct_t;
	
	function uctrl_loop_to_raw(din : hcore_uctrl_loop_struct_t) return std_logic_vector;
	function raw_to_uctrl_loop(din : std_logic_vector(8 downto 0)) return hcore_uctrl_loop_struct_t;
	
	function to_uctrl_to_raw(din : hcore_to_uctrl_struct_t) return std_logic_vector; 
	function raw_to_to_uctrl(din : std_logic_vector(63 downto 0)) return hcore_to_uctrl_struct_t;
	
	function from_uctrl_to_raw(din : uctrl_to_hcore_struct_t) return std_logic_vector;
	function raw_to_from_uctrl(din : std_logic_vector(39 downto 0)) return uctrl_to_hcore_struct_t;
	
end hcore_package;

package body hcore_package is

	-------------------------------------------------------------------------------------------------
	-- ctrl
	-------------------------------------------------------------------------------------------------	

	function hcore_ctrl_to_raw(din : hcore_ctrl_struct_t)
		return std_logic_vector is
	variable result : std_logic_vector(39 downto 0);
	begin 
		result := 	din.hcore_sel		&
						din.address		&
						din.data;
						
		return  result;		
	end;
	
	function raw_to_hcore_ctrl(din : std_logic_vector(39 downto 0))
		return hcore_ctrl_struct_t is
	variable result : hcore_ctrl_struct_t;
	begin 
	
		result.hcore_sel 	:= din(39 downto 32);
		result.address 	:= din(31 downto 16);
		result.data			:= din(15 downto 0);

		return  result;		
	end;
	
	-------------------------------------------------------------------------------------------------
	-- memory loop
	-------------------------------------------------------------------------------------------------	
	
	function memory_loop_to_raw(din : hcore_memory_loop_struct_t)
		return std_logic_vector is
	variable result : std_logic_vector(47 downto 0);
	begin 
	
		case din.code is
			when hmc_ld		 =>	result(1 downto 0) := "00";
			when hmc_ldw	 =>	result(1 downto 0) := "01";
			when hmc_ldws	=>	result(1 downto 0) := "10";
			when hmc_ldc	 => result(1 downto 0) := "11";
			when others   => result(1 downto 0) := "11";
		end case;
	
		result(8 downto 2) :=  din.rc;		
		result(40 downto 9):=  din.address;	
		
		return  result;		
	end;
	
	function raw_to_memory_loop(din : std_logic_vector(47 downto 0))
		return hcore_memory_loop_struct_t is
	variable result : hcore_memory_loop_struct_t;
	begin 
	
		result.address := din(40 downto 9);
		result.rc 		:= din(8 downto 2);
		
		case din(1 downto 0) is
			when "00"    => result.code := hmc_ld;
			when "01"    => result.code := hmc_ldw;
			when "10"    => result.code := hmc_ldws;
			when "11"    => result.code := hmc_ldc;
			when others  => result.code := hmc_ldc;
		end case;

		return  result;		
	end;
	
	-------------------------------------------------------------------------------------------------
	-- dbuffer queue
	-------------------------------------------------------------------------------------------------	
	
	function dbuffer_queue_to_raw(din : hcore_memory_dbuffer_queue_struct_t)
		return std_logic_vector is
	variable result : std_logic_vector(39 downto 0);
	begin 
	
		result(6 downto 0) :=  din.rc;		
		result(38 downto 7):=  din.data;	
		
		return  result;		
	end;
	
	function raw_to_dbuffer_queue(din : std_logic_vector(39 downto 0))
		return hcore_memory_dbuffer_queue_struct_t is
	variable result : hcore_memory_dbuffer_queue_struct_t;
	begin 
	
		result.data := din(38 downto 7);
		result.rc 		:= din(6 downto 0);
			
		return  result;		
	end;
	
	-------------------------------------------------------------------------------------------------
	-- uctrl loop
	-------------------------------------------------------------------------------------------------	
	
	function uctrl_loop_to_raw(din : hcore_uctrl_loop_struct_t)
		return std_logic_vector is
	variable result : std_logic_vector(8 downto 0);
	begin 
		result(8 downto 7) := "00"; 	-- not used yet
		result(6 downto 0) :=  din.rc;					
		return  result;		
	end;
	
	function raw_to_uctrl_loop(din : std_logic_vector(8 downto 0))
		return hcore_uctrl_loop_struct_t is
	variable result : hcore_uctrl_loop_struct_t;
	begin 	
		result.rc 	:= din(6 downto 0);
		return  result;		
	end;
	
	-------------------------------------------------------------------------------------------------
	-- to uctrl 
	-------------------------------------------------------------------------------------------------	
	
	function to_uctrl_to_raw(din : hcore_to_uctrl_struct_t)
		return std_logic_vector is
	variable result : std_logic_vector(63 downto 0);
	begin 
	   result(63 downto 57) := (others => '0');
		result(56)				:= din.do_return;
		result(55 downto 48) := din.hcore_id; 	
		result(47 downto 16) := din.data; 	
		result(15 downto 8) 	:= din.address; 	
		result(7 downto 0) 	:= din.code;					
		return  result;		
	end;

	function raw_to_to_uctrl(din : std_logic_vector(63 downto 0))
		return hcore_to_uctrl_struct_t is
	variable result : hcore_to_uctrl_struct_t;
	begin 	
		result.do_return  := din(56);
		result.hcore_id	:= din(55 downto 48);
		result.data			:= din(47 downto 16);
		result.address		:= din(15 downto 8); 	 
		result.code			:= din(7 downto 0); 	 				
		return  result;		
	end;
	
	-------------------------------------------------------------------------------------------------
	-- from uctrl 
	-------------------------------------------------------------------------------------------------	
	
	function from_uctrl_to_raw(din : uctrl_to_hcore_struct_t)
		return std_logic_vector is
	variable result : std_logic_vector(39 downto 0);
	begin 
		result(39 downto 32) := din.hcore_id; 	
		result(31 downto 0) := din.data; 						
		return  result;		
	end;

	function raw_to_from_uctrl(din : std_logic_vector(39 downto 0))
		return uctrl_to_hcore_struct_t is
	variable result : uctrl_to_hcore_struct_t;
	begin 	
		result.hcore_id	:= din(39 downto 32);
		result.data			:= din(31 downto 0); 	 				
		return  result;		
	end;

end package body;	


