library ieee;
use work.bus_package.all;
use work.hcore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity performance_cnt is
		port 
		(
				-- global

				clk											: in std_logic;
				reset											: in std_logic;
				
				-- config
				
				i_update_cnt_max							: in std_logic_vector(31 downto 0);
				i_update_cnt_max_we						: in std_logic;
				
				i_clear_cnt									: in std_logic; -- clear counters (not update max)
				i_enable										: in std_logic;
				
				--	in
				
				i_icache_hit								: in std_logic;
				i_fetch_reload								: in std_logic;
				i_score_stall								: in std_logic;
				i_execute_stall							: in std_logic;
				i_dispatch_merge_stall					: in std_logic;
				i_dbuffer_hit								: in std_logic;
				i_dbuffer_miss								: in std_logic;
				i_soft_trigger								: in std_logic;
				
				-- out
				
				o_stored_icache_hit_cnt 				: out std_logic_vector(31 downto 0);
				o_stored_fetch_reload_cnt 			   : out std_logic_vector(31 downto 0);
				o_stored_score_stall_cnt 			   : out std_logic_vector(31 downto 0);
				o_stored_execute_stall_cnt 			: out std_logic_vector(31 downto 0);
				o_stored_dispatch_merge_stall_cnt   : out std_logic_vector(31 downto 0);
				o_stored_dbuffer_hit_cnt 				: out std_logic_vector(31 downto 0);
				o_stored_dbuffer_miss_cnt   			: out std_logic_vector(31 downto 0);
				o_stored_soft_cnt   						: out std_logic_vector(31 downto 0)
				
		);
end entity;

architecture rtl of performance_cnt is

	signal icache_hit_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal fetch_reload_cnt 			: std_logic_vector(31 downto 0) := (others => '0');
	signal score_stall_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal execute_stall_cnt 			: std_logic_vector(31 downto 0) := (others => '0');
	signal dispatch_merge_stall_cnt 	: std_logic_vector(31 downto 0) := (others => '0');
	signal dbuffer_hit_cnt  			: std_logic_vector(31 downto 0) := (others => '0');
	signal dbuffer_miss_cnt 			: std_logic_vector(31 downto 0) := (others => '0');
	signal soft_cnt 						: std_logic_vector(31 downto 0) := (others => '0');

	signal stored_icache_hit_cnt 					: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_fetch_reload_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_score_stall_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_execute_stall_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_dispatch_merge_stall_cnt 	: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_dbuffer_hit_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_dbuffer_miss_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
	signal stored_soft_cnt 							: std_logic_vector(31 downto 0) := (others => '0');
	
	signal update_cnt		 : std_logic_vector(31 downto 0) := (others => '0');
	signal update_cnt_max : std_logic_vector(31 downto 0) := x"02000000";
	
	signal clear : std_logic := '0';
	
begin

	o_stored_icache_hit_cnt 				<= stored_icache_hit_cnt; 				
	o_stored_fetch_reload_cnt 				<= stored_fetch_reload_cnt; 			
	o_stored_score_stall_cnt 				<= stored_score_stall_cnt; 			
	o_stored_execute_stall_cnt 			<= stored_execute_stall_cnt; 			
	o_stored_dispatch_merge_stall_cnt	<= stored_dispatch_merge_stall_cnt;
	o_stored_dbuffer_hit_cnt				<= stored_dbuffer_hit_cnt;
	o_stored_dbuffer_miss_cnt				<= stored_dbuffer_miss_cnt;
	o_stored_soft_cnt							<= stored_soft_cnt;
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				update_cnt_max <= x"02000000";
			else
				if i_update_cnt_max_we = '1' then
					update_cnt_max <= i_update_cnt_max;
				end if;
			end if;
		end if;
	end process;
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' or clear = '1' or i_clear_cnt = '1' then
				icache_hit_cnt	 				<= (others => '0');
			   fetch_reload_cnt 				<= (others => '0');
			   score_stall_cnt 				<= (others => '0');
			   execute_stall_cnt 			<= (others => '0');
			   dispatch_merge_stall_cnt 	<= (others => '0');
				dbuffer_hit_cnt 				<= (others => '0');
				dbuffer_miss_cnt				<= (others => '0');
				soft_cnt							<= (others => '0');
			else
			
				if i_enable = '1' then -- only if count enable
			
					if i_icache_hit = '1' then
						icache_hit_cnt <= icache_hit_cnt + 1;
					end if;
					
					if i_fetch_reload = '1' then
						fetch_reload_cnt <= fetch_reload_cnt + 1;
					end if;
					
					if i_score_stall = '1' then
						score_stall_cnt <= score_stall_cnt + 1;
					end if;
					
					if i_execute_stall = '1' then
						execute_stall_cnt <= execute_stall_cnt + 1;
					end if;
					
					if i_dispatch_merge_stall = '1' then
						dispatch_merge_stall_cnt <= dispatch_merge_stall_cnt + 1;
					end if;
					
					if i_dbuffer_hit = '1' then
						dbuffer_hit_cnt <= dbuffer_hit_cnt + 1;
					end if;
					
					if i_dbuffer_miss = '1' then
						dbuffer_miss_cnt <= dbuffer_miss_cnt + 1;
					end if;
					
					if i_soft_trigger = '1' then
						soft_cnt <= soft_cnt + 1;
					end if;
					
				end if;
				
			end if;
		end if;
	end process;
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' or i_clear_cnt = '1' then
				update_cnt 	<= (others => '0');
				clear			<= '0';
			else
			
				clear			<= '0'; -- default
			
				if i_enable = '1' then
					if update_cnt >= update_cnt_max then
					
						update_cnt 	<= (others => '0');
						clear			<= '1';
						
						stored_icache_hit_cnt 					<= icache_hit_cnt;					
						stored_fetch_reload_cnt 			   <= fetch_reload_cnt; 			
						stored_score_stall_cnt 			      <= score_stall_cnt; 			
						stored_execute_stall_cnt 			   <= execute_stall_cnt; 		
						stored_dispatch_merge_stall_cnt     <= dispatch_merge_stall_cnt;
						stored_dbuffer_hit_cnt 					<= dbuffer_hit_cnt;
						stored_dbuffer_miss_cnt					<= dbuffer_miss_cnt;
						stored_soft_cnt							<= soft_cnt;
						
					else						
						update_cnt <= update_cnt + 1;						
					end if;
				end if;
			end if;
		end if;
	end process;
	

end rtl;
