library ieee;
use work.hcore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_regio is
		port 
		(
				-- global

				clk						: in std_logic;
				reset						: in std_logic;
				
				-- in
				
				i_stall					: in std_logic;
						
				--	s3 / s5
						
				i_ra						: in std_logic_vector(6 downto 0);		-- read
				i_rb						: in std_logic_vector(6 downto 0);		-- read
					
				i_rc						: in std_logic_vector(6 downto 0);		-- write
				i_rc_we					: in std_logic;								--	enable			
				i_rc_data				: in std_logic_vector(31 downto 0);

				o_ra_data				: out std_logic_vector(31 downto 0);
			   o_rb_data				: out std_logic_vector(31 downto 0);
		
				-- hold / through	
	
				i_hold_rc				: in std_logic_vector(6 downto 0);		-- write
				i_hold_rc_we  			: in std_logic;								--			enable
				o_hold_rc				: out std_logic_vector(6 downto 0);		-- write
				o_hold_rc_we 			: out std_logic;								--			enable
								
				i_hold_imm				: in std_logic_vector(31 downto 0);	
				o_hold_imm				: out std_logic_vector(31 downto 0);	
				
				i_hold_a_as_reg		: in std_logic;
				i_hold_b_as_reg		: in std_logic;
				i_hold_c_as_reg		: in std_logic;
				
				o_hold_a_as_reg		: out std_logic;
				o_hold_b_as_reg		: out std_logic;
				o_hold_c_as_reg		: out std_logic;

				i_hold_condition		: in hcondition_t;
				i_hold_opcode			: in hopcode_t;
				i_hold_oo_opcode		: in hopcode_oo_t;
				o_hold_condition		: out hcondition_t;
				o_hold_opcode			: out hopcode_t;	
				o_hold_oo_opcode		: out hopcode_oo_t									
		);
end entity;



architecture rtl of hcore_regio is
		
	
		
begin

	------------------------------------------------------------
	-- reg
	------------------------------------------------------------
	
	process(all)
		type regs_t is array(0 to 127) of std_logic_vector(31 downto 0);	
		variable regs 	: regs_t 	:= (others => (others => '0'));	
		--attribute ramstyle : string;
		--attribute ramstyle of regs : variable is "no_rw_check, M512";
	begin
		if rising_edge(clk) then
			if reset = '1' then			
			else
			
				------------------------------------------------------------
				-- S5 (prio) write/unlock (not stalled because after exec)
				------------------------------------------------------------
					
				if i_rc_we = '1' then
						regs(conv_integer(i_rc)) := i_rc_data;
				end if;
				
				------------------------------------------------------------
				-- S3 read / lock
				------------------------------------------------------------
			
				if i_stall = '0' then 
					o_ra_data	<= regs(conv_integer(i_ra));
					o_rb_data	<= regs(conv_integer(i_rb));							
				end if;

			end if;
		end if;		
	end process;
	
	------------------------------------------------------------
	-- hold / through
	------------------------------------------------------------
	
	process(all)		
	begin
		if rising_edge(clk) then
			if reset = '1' then	

					o_hold_rc  				<=	(others => '0');
					o_hold_rc_we			<= '0';
					
					o_hold_imm				<= (others => '0');
					
					o_hold_a_as_reg		<= '0';
					o_hold_b_as_reg      <= '0';
					o_hold_c_as_reg      <= '0';
					
					o_hold_condition		<= hcc_never;	
					o_hold_opcode			<= hope_nop;	
					o_hold_oo_opcode		<= hope_oo_nop;		
			
			else
				
				------------------------------------------------------------
				-- S3 
				------------------------------------------------------------
			
				if i_stall = '0' then 
			
					-- through
					
					o_hold_rc  				<=	i_hold_rc;
					o_hold_rc_we			<= i_hold_rc_we;
					
					o_hold_imm				<= i_hold_imm;
					
					o_hold_a_as_reg		<= i_hold_a_as_reg;
					o_hold_b_as_reg      <= i_hold_b_as_reg;
					o_hold_c_as_reg      <= i_hold_c_as_reg;
					
					o_hold_condition		<= i_hold_condition;	
					o_hold_opcode			<= i_hold_opcode;	
					o_hold_oo_opcode		<= i_hold_oo_opcode;						
					
				end if;

			end if;
		end if;		
	end process;
	
end rtl;
