library ieee;
use work.hcore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_score is
		port 
		(
				-- global

				clk						: in std_logic;
				reset						: in std_logic;
				
				-- 
				
				i_stall					: in std_logic;
				i_clear_locks			: in std_logic;
				o_reg_lock				: out std_logic;
						
				--	s3 / s5
						
				i_ra						: in std_logic_vector(6 downto 0);		
				i_rb						: in std_logic_vector(6 downto 0);		
				o_ra						: out std_logic_vector(6 downto 0);
				o_rb						: out std_logic_vector(6 downto 0);
								
						
				i_rc_lock				: in std_logic;
				i_rc_unlock				: in std_logic;
				i_rc_num_unlock		: in std_logic_vector(6 downto 0);
			
				o_lock_exception		: out std_logic;
				o_unlock_exception	: out std_logic;
		
				i_rc				: in std_logic_vector(6 downto 0);		-- write
				i_rc_we  		: in std_logic;								--			enable
				o_rc				: out std_logic_vector(6 downto 0);		-- write
				o_rc_we 			: out std_logic;								--			enable
						
				i_imm				: in std_logic_vector(31 downto 0);	
				o_imm				: out std_logic_vector(31 downto 0);	
				
				i_a_as_reg		: in std_logic;
				i_b_as_reg		: in std_logic;
				i_c_as_reg		: in std_logic;
				
				o_a_as_reg		: out std_logic;
				o_b_as_reg		: out std_logic;
				o_c_as_reg		: out std_logic;

				i_condition		: in hcondition_t;
				i_opcode			: in hopcode_t;
				i_oo_opcode		: in hopcode_oo_t;
				o_condition		: out hcondition_t;
				o_opcode			: out hopcode_t;
				o_oo_opcode		: out hopcode_oo_t								
													
		);
end entity;



architecture rtl of hcore_score is
		
signal t_stall		: std_logic := '0';
signal stored_ra	: std_logic_vector(6 downto 0) := (others => '0');		
signal stored_rb	: std_logic_vector(6 downto 0) := (others => '0');		
signal tst_ra	   : std_logic := '0';
signal tst_rb	   : std_logic := '0';

begin
	
	o_reg_lock <= t_stall;
	
	------------------------------------------------------------
	-- locks
	------------------------------------------------------------
	
	process(all)	
		type locks_t is array(0 to 127) of std_logic;		
		variable locks : locks_t 	:= (others => '0');
	begin
		if rising_edge(clk) then
			if reset = '1' or i_clear_locks = '1' then	
			
				for i in 0 to 127 loop
					locks(i) := '0';
				end loop;
				
				o_lock_exception		<= '0';
				o_unlock_exception	<= '0';	
				t_stall 	<= '0';
				
			else
			
				-- default 
				
				o_lock_exception		<= '0';
				o_unlock_exception	<= '0';			
				
							

				------------------------------------------------------------
				-- S2 tst
				------------------------------------------------------------
				
				--if i_stall = '0' then 
				
					t_stall 	<= '0';	-- default
					
					if t_stall = '0' then
						if (locks(conv_integer(i_ra)) = '1' and i_a_as_reg = '1') or
							(locks(conv_integer(i_rb)) = '1' and i_b_as_reg = '1') then
								-- there is a lock
								t_stall 	<= '1';	-- so stall out
								stored_ra	<= i_ra;
								stored_rb	<= i_rb;
								tst_ra		<= i_a_as_reg;
								tst_rb		<= i_b_as_reg;								
						end if;
					else
						if (locks(conv_integer(stored_ra)) = '1' and tst_ra = '1') or
							(locks(conv_integer(stored_rb)) = '1' and tst_rb = '1') then
							t_stall 	<= '1';	-- no change so stall out	
						end if;							
					end if;
				--end if;
				
				------------------------------------------------------------
				-- S2 lock
				------------------------------------------------------------
			
				if i_stall = '0' then 
									
					-- set lock
					
					if i_rc_lock = '1' then						
						if locks(conv_integer(i_rc)) = '1' then --  already locked -> lock exception
							o_lock_exception <= '1';
						end if;
						
						locks(conv_integer(i_rc)) := '1';						
					end if;						
				end if;
				
				------------------------------------------------------------
				-- S5 unlock
				------------------------------------------------------------
				
				if i_rc_unlock = '1' then										
					if locks(conv_integer(i_rc_num_unlock)) = '0' then --  already unlocked -> unlock exception
						o_unlock_exception <= '1';
					end if;
						
					locks(conv_integer(i_rc_num_unlock)) := '0';					
				end if;			
				
			end if;
		end if;		
	end process;
	
	------------------------------------------------------------
	-- hold / through
	------------------------------------------------------------
	
	process(all)		
	begin
		if rising_edge(clk) then
			if reset = '1' then	
			
					o_ra			<=	(others => '0');
					o_rb			<= (others => '0');
													
					o_rc  		<=	(others => '0');
					o_rc_we		<= '0';
					
					o_imm			<= (others => '0');
					
					o_a_as_reg	<= '0';
					o_b_as_reg  <= '0';
					o_c_as_reg  <= '0';
					
					o_condition	<= hcc_never;	
					o_opcode		<= hope_nop;		
					o_oo_opcode	<= hope_oo_nop;
					
			else
				
				------------------------------------------------------------
				-- S2 
				------------------------------------------------------------
			
				if i_stall = '0' and t_stall = '0' then 
			
					-- through
					
					o_ra			<=	i_ra;
					o_rb			<= i_rb;
													
					o_rc  		<=	i_rc;
					o_rc_we		<= i_rc_we;
					
					o_imm			<= i_imm;
					
					o_a_as_reg	<= i_a_as_reg;
					o_b_as_reg  <= i_b_as_reg;
					o_c_as_reg  <= i_c_as_reg;
					
					o_condition	<= i_condition;	
					o_opcode		<= i_opcode;		
					o_oo_opcode	<= i_oo_opcode;					
					
				end if;

			end if;
		end if;		
	end process;
	
end rtl;
