library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity mem_flash is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	
				
				-- Flash 
				FL_ADDR			: out std_logic_vector(22 downto 0);
				FL_CE_N			: out std_logic;
				FL_DQ				: inout std_logic_vector(7 downto 0);
				FL_OE_N			: out std_logic;
				FL_RST_N			: out std_logic;
				FL_RY				: in std_logic;
				FL_WE_N			: out std_logic;
				FL_WP_N			: out std_logic;
	
				-- recive ports
				
				to_flash_data		: in flash_n_data_t;		
				to_flash_busy		: in flash_n_busy_t;
				to_flash_ack		: out flash_n_ack_t;	
			
				-- send ports
				
				from_flash_rclk	: in flash_rclk_t;
				from_flash_data	: out flash_n_data_t;
				from_flash_busy	: out flash_n_busy_t;
				from_flash_ack		: in flash_n_ack_t	
				
		);
end entity;

architecture rtl of mem_flash is

-- dcfifo (for send)

--component dcfifo_min IS
--	PORT
--	(
--		aclr		: IN STD_LOGIC  := '0';
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q			: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdempty	: OUT STD_LOGIC ;
--		wrusedw	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--END component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;

--

type send_xn_data_t is array(0 to mem_flash_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type send_xn_wrreq_t is array(0 to mem_flash_reciver_count-1) of STD_LOGIC;
type send_xn_wrfull_t is array(0 to mem_flash_reciver_count-1) of STD_LOGIC; 
type send_xn_wrusedw_t is array(0 to mem_flash_reciver_count-1) of STD_LOGIC_VECTOR (7 DOWNTO 0);

--

signal send_xn_data		: send_xn_data_t := (others => (others => '0'));
signal send_xn_wrreq		: send_xn_wrreq_t := (others => '0');
signal send_xn_wrfull	: send_xn_wrfull_t := (others => '0'); 
signal send_xn_wrusedw	: send_xn_wrusedw_t := (others => (others => '0'));

-- stats

type transferState_t is
	(
		WAIT_TRANSFER_START,
		S1,S2,S3,S4,S5,S6,S7,S8,S9,S10,S11,S12,S13,S14,S15,S16,S17,S18,S19,S20,S21,S22,S23	
	);
	
	
signal transferState : transferState_t := WAIT_TRANSFER_START;

-- stage 1

signal s1_push_out	: std_logic := '0';
signal s1_data			: port_struct := fill_port('0');
signal s1_id			: integer range 0 to mem_flash_reciver_count-1 := 0;

signal flash_ready	: std_logic := '0';
	
begin
	
	-- Flash 
	
	FL_CE_N				<= '0';	-- ever enable 
	FL_OE_N				<= '0';	-- only read implemented
	FL_RST_N				<= '1';	-- no reset
	FL_WE_N				<= '1';	-- dont write
	FL_WP_N				<= '1';
	
	-- check incomming
	
	process(clk)
	begin
		if rising_edge(clk) then
		
			s1_push_out <= '0';
			s1_data  	<= raw_to_port(to_flash_data(0));
		
			if reset = '0' then
				for i in 0 to mem_flash_reciver_count-1 loop
					if to_flash_busy(i) = '0' and flash_ready = '1' and send_xn_wrfull(i) = '0' then
						s1_data  	<= raw_to_port(to_flash_data(i));
						s1_id			<= i;
						s1_push_out	<= '1';
						exit; -- stop loop
					end if;	
				end loop;
			end if;
		end if;
	end process;
	
	-- ack
	
	process(all)
	begin
	
		for i in 0 to mem_flash_reciver_count-1 loop
			to_flash_ack(i) <= '0';
		end loop;
	
		if reset = '0' then
			for i in 0 to mem_flash_reciver_count-1 loop
				if to_flash_busy(i) = '0' and flash_ready = '1' and send_xn_wrfull(i) = '0' then
					to_flash_ack(i) <= '1';
				exit; -- stop loop
				end if;	
			end loop;
		end if;
	end process;
	
	--
	
	
	flash_ready <= '1' when transferState = WAIT_TRANSFER_START and s1_push_out = '0' and reset = '0'
							 else '0';
	
	-- stats ctrl
		
	process(clk)
		variable data_in : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
			if reset = '1' then
				transferState <= WAIT_TRANSFER_START;
			else
				
				--
				
				case transferState is
					when WAIT_TRANSFER_START => if s1_push_out = '1' then
																data_in := s1_data;
																transferState <= S1;
														 end if;
														 
					when S1  => transferState <= S2;  -- 10
					when S2  => transferState <= S3;  -- 20
					when S3  => transferState <= S4;  -- 30
					when S4  => transferState <= S5;  -- 40
					when S5  => transferState <= S6;  -- 50
					when S6  => transferState <= S7;  -- 60
					when S7  => transferState <= S8;  -- 70
					when S8  => transferState <= S9;  -- 80
					when S9  => transferState <= S10; -- 90
					when S10 => transferState <= S11; -- 100
					
					when S11 => transferState <= S12;
									
					when S12 => transferState <= S13;
					when S13 => transferState <= S14;
					when S14 => transferState <= S15;
					
					when S15 => 
									if data_in.dmode = '0' then -- 32 bit continue
										transferState <= S16;
									else -- 16 bit stop							
										transferState <= WAIT_TRANSFER_START;
									end if;
									
					when S16 => transferState <= S17;
					when S17 => transferState <= S18;
					when S18 => transferState <= S19;
					
					when S19 => transferState <= S20;
									
					when S20 => transferState <= S21;
					when S21 => transferState <= S22;
					when S22 => transferState <= S23;
					
					when S23 => transferState <= WAIT_TRANSFER_START;
				end case;
			end if;

		end if;
	end process;
	
	-- stats work
	
	process(clk)
		variable address_stored : std_logic_vector(22 downto 0) := (others => '0');
		variable data_in : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
		
			-- default
				
			for i in 0 to mem_flash_reciver_count-1 loop
				send_xn_wrreq(i) <= '0';
			end loop;
		
			if reset = '1' then
				-- nothing to do
			else

				--
			
				case transferState is
					when WAIT_TRANSFER_START => if s1_push_out = '1' then
																
																data_in := s1_data;
																address_stored  := data_in.address(21 downto 0) & '0';
																
																FL_ADDR <= address_stored;
														 end if;
					
					when S11 => data_in.data(31 downto 24) := FL_DQ;																	
									FL_ADDR <= address_stored + 1;							
					
					when S15 => data_in.data(23 downto 16) := FL_DQ;	
									
									FL_ADDR <= address_stored + 2;
									
									if data_in.dmode = '1' then -- 16 bit mode stop now
										
										data_in.data := x"0000" & data_in.data(31 downto 16); 										
										send_xn_wrreq(s1_id)	<= '1'; -- push

									end if;
						
					when S19 => data_in.data(15 downto 8) := FL_DQ;
																		
									FL_ADDR <= address_stored + 3;
																					
					when S23 => data_in.data(7 downto 0) := FL_DQ; -- 32 bit mode end
					
									send_xn_wrreq(s1_id)	<= '1'; -- push
									
					when others =>
				end case;
			end if;
			
			-- set to all fifo inputs (send_xn_wrreq arbit write)
			
			for i in 0 to mem_flash_reciver_count-1 loop
				send_xn_data(i) <= port_to_raw(data_in);				
			end loop;

		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	---- SEND DATA 
	-----------------------------------------------------------------------------------------------------------
	
	-- full generate

	process(clk)
	begin
		for i in 0 to mem_flash_reciver_count-1 loop
			if rising_edge(clk) then
				if send_xn_wrusedw(i) >=  x"fc" then --x"1fb" at 9 bit?
					send_xn_wrfull(i) <= '1';
				else
					send_xn_wrfull(i) <= '0';
				end if;
			end if;		
		end loop;		
	end process;
	
	-- data out
	
	dcfifo_loop : for i in 0 to mem_flash_reciver_count-1 generate
         dcfifo : dc_fifo 
					generic map
					(
						FIFO_DEPTH 			=> 256,
						FIFO_WIDTH 			=> 72,
						FIFO_SHOWAHEAD 	=> "ON",
						FIFO_USEDWWIDTH	=> 8,
						FIFO_SYNCSTAGES	=> 3
					)  
					port map
					(
						'0',
						send_xn_data(i),
						from_flash_rclk(i),		-- read with component clk
						from_flash_ack(i),	
						clk,					-- write with flash clk
						send_xn_wrreq(i),	
						from_flash_data(i),
						from_flash_busy(i),
						send_xn_wrusedw(i)
					);
   end generate;
	
end rtl;
