library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity mem_sdram is
		port 
		(
				-- global
				
				clk					: in std_logic;
				reset					: in std_logic;	
					
				-- SDRAM 
				
				DRAM_ADDR			: out std_logic_vector(12 downto 0);
				DRAM_BA				: out std_logic_vector(1 downto 0);
				DRAM_CAS_N			: out std_logic;				
				DRAM_CS_N			: out std_logic;
				DRAM_DQ				: inout std_logic_vector(31 downto 0);
				DRAM_DQM				: out std_logic_vector(3 downto 0);
				DRAM_RAS_N			: out std_logic;
				DRAM_WE_N			: out std_logic;
				
				-- recive ports
				
				to_sdram_data		: in sdram_n_data_t;		
				to_sdram_busy		: in sdram_n_busy_t;
				to_sdram_ack		: out sdram_n_ack_t;	
			
				-- send ports
				
				from_sdram_rclk	: in sdram_rclk_t;
				from_sdram_data	: out sdram_n_data_t;
				from_sdram_busy	: out sdram_n_busy_t;
				from_sdram_ack		: in sdram_n_ack_t	
				
		);
end entity;

architecture rtl of mem_sdram is

-- fifo for recive 

--component fifo_max
--	PORT
--	(
--		clock		: IN STD_LOGIC ;
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdreq		: IN STD_LOGIC ;
--		sclr		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		empty		: OUT STD_LOGIC ;
--		full		: OUT STD_LOGIC ;
--		q		: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		usedw		: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)		
--	);
--end component;

component sc_fifo_usedw
	GENERIC 
	(
		FIFO_DEPTH 			: natural;
		FIFO_WIDTH 			: natural;
		FIFO_SHOWAHEAD 	: string;
		FIFO_USERAM			: string;
		FIFO_USEDWWIDTH	: natural
		
	);
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		usedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
end component;

signal buf_data	: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');
signal buf_rdreq	: STD_LOGIC := '0';
signal buf_wrreq	: STD_LOGIC := '0';
signal buf_empty	: STD_LOGIC := '0';
signal buf_full	: STD_LOGIC := '0';
signal buf_dummy	: STD_LOGIC := '0';
signal buf_q		: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');
signal buf_usedw	: STD_LOGIC_VECTOR (7 DOWNTO 0) := (others => '0');

-- dcfifo (for send)

--component dcfifo_min IS
--	PORT
--	(
--		aclr		: IN STD_LOGIC  := '0';
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q			: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdempty	: OUT STD_LOGIC ;
--		wrusedw	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--END component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;

--

type send_xn_data_t is array(0 to mem_sdram_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type send_xn_wrreq_t is array(0 to mem_sdram_reciver_count-1) of STD_LOGIC;
type send_xn_wrfull_t is array(0 to mem_sdram_reciver_count-1) of STD_LOGIC; 
type send_xn_wrusedw_t is array(0 to mem_sdram_reciver_count-1) of STD_LOGIC_VECTOR (7 DOWNTO 0);

--

signal send_xn_data		: send_xn_data_t := (others => (others => '0'));
signal send_xn_wrreq		: send_xn_wrreq_t := (others => '0');
signal send_xn_wrfull	: send_xn_wrfull_t := (others => '0'); 
signal send_xn_wrusedw	: send_xn_wrusedw_t := (others => (others => '0'));

--

constant sd_cas_delay : integer range 2 to 3 := 3;

type SD_CTRL_STATE_t is
(
	CTRL_RESET,

	CTRL_WAIT,
	CTRL_ACTIVATE_CURRENT_BANK_ROW,
	
	--
	
	CTRL_NOP_N_RETURN,
	
	-- init
	
	CTRL_INIT_BRINGUP,
	CTRL_INIT_PRECHARGE_ALL,
	CTRL_INIT_AUTOREFRESH_1,
	CTRL_INIT_AUTOREFRESH_2,
	CTRL_INIT_AUTOREFRESH_3,
	CTRL_INIT_AUTOREFRESH_4,
	CTRL_INIT_AUTOREFRESH_5,
	CTRL_INIT_AUTOREFRESH_6,
	CTRL_INIT_AUTOREFRESH_7,
	CTRL_INIT_AUTOREFRESH_8,
	CTRL_INIT_SETMODE,
	
	-- refresh
	
	CTRL_REFRESH_PRECHARGE_ALL,
	CTRL_REFRESH_AUTOREFRESH
);

--attribute enum_encoding : string;
--attribute enum_encoding of SD_CTRL_STATE_t : type is "onehot"; -- one hot for better performance

constant tbu  : integer := 20000
-- synthesis translate_off
 -19990    -- bringup nop count 
-- synthesis translate_on
;

constant trp  : integer := 3;    -- precharge to activate
constant trc  : integer := 10;   -- refresh to refresh/activate
constant tmrd : integer := 2;    -- mode set to command delay
constant trcd : integer := 3;    -- activated to read/write
constant tref : integer := 977;  
-- refresh 64ms/8192 => every 781.2 ticks (100MHz) -10 ticks (pre-act, bad placmenent)
-- (64ms/8192)*clk -> 115 MHz (898.4)   - 10 = 887
-- (64ms/8192)*clk -> 125 MHz (976.6)   - 10 = 977
-- (64ms/8192)*clk -> 130 MHz (1015.6)  - 10 = 1004
-- (64ms/8192)*clk -> 133.33333 MHz (1041.66) - 10 = 1031
-- (64ms/8192)*clk -> 142 MHz (1109.37) - 10 = 1099

signal refresh_counter : integer range 0 to 2047 := 0;
signal do_refresh : std_logic := '0';
signal ack_refresh : std_logic := '0';

signal nop_counter 		: integer range 0 to 32767 := 0;
signal nop_counter_max 	: integer range 0 to 32767 := 0;

signal read_count 	: integer range 0 to 7 := 0;
signal write_count 	: integer range 0 to 7 := 0;	

constant sd_mode_data	: std_logic_vector(12 downto 0) :=	"0000000110000"; -- cas2 = 	"0000000100000";									

signal sd_ctrl_state 		     : SD_CTRL_STATE_t := CTRL_RESET;
signal sd_ctrl_state_next 		: SD_CTRL_STATE_t := CTRL_RESET;

type SD_CMD_t is 
(
	SD_DSEL, 
	SD_NOP,
	SD_BURST_STOP,
	SD_READ,
	SD_READA,
	SD_WRITE,
	SD_WRITEA,
	SD_ACTIVATE,
	SD_PRECHARGE,
	SD_PRECHARGE_ALL,
	SD_REFRESH,
	SD_MODREGSET
);


signal sd_cmd 		   	: SD_CMD_t := SD_NOP;
signal sd_cmd_ba	   	: std_logic_vector(1 downto 0) := (others => '0');
signal sd_cmd_row	  		: std_logic_vector(12 downto 0) := (others => '0');
signal sd_cmd_col	  		: std_logic_vector(9 downto 0) := (others => '0');
signal sd_word_mode		: std_logic := '0';
signal sd_address_lsb	: std_logic := '0';
signal sd_data				: std_logic_vector(31 downto 0) := (others => '0');
signal sd_indata			: port_struct := fill_port('0');

signal sd_active_row			: std_logic_vector(12 downto 0) := (others => '0');
signal sd_active_ba			: std_logic_vector(1 downto 0) := (others => '0');
signal sd_active_ba_row		: std_logic := '0';

--

signal rs0_push : std_logic := '0';
signal rs0_data : port_struct := fill_port('0');

signal rs1_push : std_logic := '0';
signal rs1_data : port_struct := fill_port('0');

signal rs2_push : std_logic := '0';
signal rs2_data : port_struct := fill_port('0');

signal rs3_push : std_logic := '0';
signal rs3_data : port_struct := fill_port('0');

--

begin

	-- recive fifo
	
	recive_buffer : sc_fifo_usedw
	generic map
	(
		FIFO_DEPTH 			=> 256,
		FIFO_WIDTH 			=> 72,
		FIFO_SHOWAHEAD 	=> "ON",
		FIFO_USERAM			=> "ON",
		FIFO_USEDWWIDTH	=> 8				
	) 
	port map
	(
		clk,
		buf_data,	
	   buf_rdreq,	
		reset,
	   buf_wrreq,	
	   buf_empty,	
	   buf_dummy,	
	   buf_q,		
		buf_usedw
	);


	process(clk)
		variable wor : std_logic := '0';
	begin
		if rising_edge(clk) then
		
			wor := '0';
		
			for i in 0 to mem_sdram_reciver_count-1 loop
				wor := wor or send_xn_wrfull(i);
			end loop;
			
		
			if buf_usedw >=  x"f" then --
					buf_full <= '1';
				else
					buf_full <= '0';-- or wor;
				end if;
			end if;		
	end process;

	-- recive ack
	
	process(all)
	begin
	
		for i in 0 to mem_sdram_reciver_count-1 loop
			to_sdram_ack(i) <= '0';
		end loop;
	
		if reset = '0' then
			for i in 0 to mem_sdram_reciver_count-1 loop
				if to_sdram_busy(i) = '0' and buf_full = '0' then						
					to_sdram_ack(i) <= '1';	
					exit; -- stop loop		
				end if;	
			end loop;
		end if;
	end process;
	
	process(clk)
		variable t_data : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
		
			-- default
		
			buf_wrreq 	<= '0';		
			--buf_data		<=  to_sdram_data(0); 
			
			--
			
			if reset = '1' then
				--
			else
				for i in 0 to mem_sdram_reciver_count-1 loop
				if to_sdram_busy(i) = '0' and buf_full = '0' then						
					
				
					t_data 		:= raw_to_port(to_sdram_data(i)); 	
					t_data.id	:= conv_std_logic_vector(i,6);	-- store id
			
					buf_data		<=  port_to_raw(t_data);
					buf_wrreq 	<= '1';
					
					exit; -- stop loop
				end if;	
			end loop;
			end if;
		end if;
	end process;

	-- buffer fifo read request
	
	process(all) 
		variable buf_in : port_struct := fill_port('0');
	begin
		
		-- default
		
		buf_in := raw_to_port(buf_q);
		buf_rdreq	<= '0';	
		
		-- accept data 
		
		if reset = '0' and												-- no reset
			sd_ctrl_state = CTRL_WAIT and 							-- in wait
			do_refresh = '0' and											-- no refresh
			buf_empty = '0' and											-- buf fifo is not empty
			sd_active_ba_row = '1' and									-- bank / row is activated
			sd_active_ba = buf_in.address(25 downto 24) and 
			sd_active_row = buf_in.address(23 downto 11) 
			--
			then 	-- same bank and row !
			
			if (buf_in.we = '1' and read_count = 0) or
				(buf_in.we = '0' and write_count = 0 and send_xn_wrfull(conv_integer(buf_in.id)) = '0') then			
				buf_rdreq	<= '1';	-- ack	
			end if;
		end if;	
	end process;	
  
	-- main states
	
	process(clk)
		 variable buf_in : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
		
			-- default
			
			buf_in := raw_to_port(buf_q);
			
			--
			
			sd_cmd_ba			<= buf_in.address(25 downto 24);
			sd_cmd_row			<= buf_in.address(23 downto 11);																		
			sd_cmd_col  		<= buf_in.address(10 downto 1);
			sd_word_mode		<= buf_in.dmode;
			sd_address_lsb		<= buf_in.address(0);
			sd_data				<= buf_in.data;
			
			ack_refresh 	<= '0';
			sd_cmd 			<= SD_NOP;	-- default is no operation
			
			if write_count > 0 then
				write_count 	<= write_count - 1;
			end if;
			
			if read_count > 0 then
				read_count 		<= read_count - 1;
			end if;
			
			-- state
		
			case sd_ctrl_state is
                              
				-- WAIT  
				  
				when CTRL_WAIT			=>			
				
												if do_refresh = '1' then -- refresh has to do now
													ack_refresh 		<= '1'; 									-- yes we do
													sd_ctrl_state 		<= CTRL_REFRESH_PRECHARGE_ALL;	-- start refresh
												else
													if reset = '0' then
														if buf_empty = '0' then	-- cmd is here
															if sd_active_ba_row = '0' then	-- there is no bank / row  activated
																if read_count = 0 and write_count = 0 then -- there is no read / write pending
																
																	
																	sd_cmd				   <= SD_ACTIVATE;						-- activate bank / row
																	
																	sd_ctrl_state        <= CTRL_NOP_N_RETURN;
																	sd_ctrl_state_next   <= CTRL_WAIT;
																	nop_counter_max      <= trcd-1;
																	nop_counter          <= 1;
																	
																	sd_active_ba_row 		<= '1';		-- now this bank and row are activated
																	sd_active_ba			<= buf_in.address(25 downto 24);
																	sd_active_row			<= buf_in.address(23 downto 11);		
																	
																end if;
															else -- there is a bank / row activated
																if sd_active_ba = buf_in.address(25 downto 24) and 
																	sd_active_row = buf_in.address(23 downto 11) then 	-- same bank and row !
																	
																	-- 
																	
																	if buf_in.we = '1' then	-- write
																		if read_count = 0 then -- no read is pending, we can write!
																			
																			sd_cmd					<= SD_WRITE;
																			sd_ctrl_state        <= CTRL_WAIT;
																																					
																			-- data accepted !!!
																			
																			write_count 			<= 3;
																		end if;
																	else
																		-- read
																		if write_count = 0 and send_xn_wrfull(conv_integer(buf_in.id)) = '0' then -- no write is pending, we can read!
																		
																			sd_cmd					<= SD_READ;
																			sd_ctrl_state        <= CTRL_WAIT;
																			
																			read_count 				<= 3;
																			
																			sd_indata				<= buf_in;
																			
																			
																			-- data accepted !!!
																		end if;
																	end if;
																	
																else
																	-- not same bank and row so close and activated
																	if read_count = 0 and write_count = 0 then -- there is no read / write pending
																	
																		sd_cmd				   <= SD_PRECHARGE_ALL;						-- close all bank / row (can optimized)
																		sd_ctrl_state        <= CTRL_NOP_N_RETURN;
																		sd_ctrl_state_next   <= CTRL_ACTIVATE_CURRENT_BANK_ROW;
																		nop_counter_max      <= trp-1;
																		nop_counter          <= 1;
																																																																			
																	end if;
																end if;
															end if;
														end if;
													end if;
												end if;
				
				
				when CTRL_ACTIVATE_CURRENT_BANK_ROW =>
				
												
												sd_cmd				   <= SD_ACTIVATE;						-- activate bank / row
												
												sd_ctrl_state        <= CTRL_NOP_N_RETURN;
												sd_ctrl_state_next   <= CTRL_WAIT;
												nop_counter_max      <= trcd-1;
												nop_counter          <= 1;
												
												sd_active_ba_row 		<= '1';		-- now this bank and row are activated (should set before)
												sd_active_ba			<= buf_in.address(25 downto 24);
												sd_active_row			<= buf_in.address(23 downto 11);											
									
			  -- RESET
			
				when CTRL_INIT_BRINGUP | CTRL_RESET=>
			  
			     						   sd_cmd				   <= SD_NOP;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_PRECHARGE_ALL;
				                     nop_counter_max      <= tbu-1;
				                     nop_counter          <= 1;
				                     
							
				when CTRL_INIT_PRECHARGE_ALL =>
				    
				                     sd_cmd				   <= SD_PRECHARGE_ALL;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_1;
				                     nop_counter_max      <= trp-1;
				                     nop_counter          <= 1;
				                     
				when CTRL_INIT_AUTOREFRESH_1 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_2;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_2 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_3;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_3 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_4;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_4 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_5;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_5 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_6;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_6 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_7;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_7 =>
	         
											sd_cmd				  	<= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_AUTOREFRESH_8;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
											
				when CTRL_INIT_AUTOREFRESH_8 =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_INIT_SETMODE;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;				                     				                     				                     				                     				                     				                     				                     				                     

				when CTRL_INIT_SETMODE =>		
				
				                     sd_cmd_row      	   <= sd_mode_data;					
				                     sd_cmd				   <= SD_MODREGSET;  
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_WAIT;
				                     nop_counter_max      <= tmrd-1;
				                     nop_counter          <= 1;
											ack_refresh 			<= '1'; 									-- reset refresh if occurred
				
				-- REFRESH
				
				when CTRL_REFRESH_PRECHARGE_ALL =>
				  
				                     sd_cmd				   <= SD_PRECHARGE_ALL;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_REFRESH_AUTOREFRESH;
				                     nop_counter_max      <= trp-1;
				                     nop_counter          <= 1;
											sd_active_ba_row		<= '0';	-- all banks rows are closed after this
				      
				when CTRL_REFRESH_AUTOREFRESH =>
	         
											sd_cmd				   <= SD_REFRESH;
				                     sd_ctrl_state        <= CTRL_NOP_N_RETURN;
				                     sd_ctrl_state_next   <= CTRL_WAIT;
				                     nop_counter_max      <= trc-1;
				                     nop_counter          <= 1;
										
				-- MULTI NOP
            
			   when CTRL_NOP_N_RETURN =>
								                    
								         if nop_counter >= nop_counter_max then
												sd_ctrl_state <= sd_ctrl_state_next;
											else
												sd_ctrl_state <= CTRL_NOP_N_RETURN; 
											end if;  
								                    
				                     nop_counter   			<= nop_counter + 1;
				                     sd_cmd				   <= SD_NOP;    
				--
				
				when others			=>			
												
			end case;
		end if;
	end process;
	
	--	process CMDs

	
		
	process(clk)
	begin
		
		if rising_edge(clk) then		
		
			rs0_push			<= '0';	-- nothing to push
			DRAM_DQ			<= (others => 'Z');
			DRAM_DQM			<= "0000";
		
			case sd_cmd is 
			  
			  when SD_DSEL =>
			    
										DRAM_CS_N 	<= '1';
										DRAM_RAS_N	<= '0';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '0';
										DRAM_BA		<= (others => '0');	-- dont care here
										DRAM_ADDR	<= (others => '0');	-- dont care here
			  
				when SD_NOP		=>	
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '1';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= (others => '0');	-- dont care here
										DRAM_ADDR	<= (others => '0');	-- dont care here
										
				when SD_BURST_STOP =>
				
				    					DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '1';
										DRAM_WE_N	<= '0';
										DRAM_BA		<= (others => '0');	-- dont care here
										DRAM_ADDR	<= (others => '0');	-- dont care here
										
				when SD_READ	=>	
				  
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= sd_cmd_ba;
										DRAM_ADDR	<= "000" & sd_cmd_col(9 downto 0);	
										DRAM_DQM		<= (others => '0');
										
										rs0_push				<= '1';	-- push read delay stages
										rs0_data				<= sd_indata;
										
				when SD_READA	=>	
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= sd_cmd_ba;
										DRAM_ADDR	<= "001" & sd_cmd_col(9 downto 0);	
										DRAM_DQM		<= (others => '0');
										
										rs0_push				<= '1';	-- push read delay stages
										rs0_data				<= sd_indata;
										
				when SD_WRITE	=>	
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '0';
										DRAM_BA		<= sd_cmd_ba;
										DRAM_ADDR	<= "000" & sd_cmd_col(9 downto 0);	
										
										if sd_word_mode = '1' then -- 16 bit
											if sd_address_lsb = '1' then
												DRAM_DQM		<= "1100";
											else
												DRAM_DQM		<= "0011";
											end if;
											DRAM_DQ		<= sd_data(15 downto 0) & sd_data(15 downto 0);
										else -- 32 bit
											DRAM_DQM		<= "0000";
											DRAM_DQ		<= sd_data;
										end if;
					
				when SD_WRITEA	=>	
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= sd_cmd_ba;
										DRAM_ADDR	<= "001" & sd_cmd_col(9 downto 0);	
										
										if sd_word_mode = '1' then -- 16 bit
											if sd_address_lsb = '1' then
												DRAM_DQM		<= "1100";
											else
												DRAM_DQM		<= "0011";
											end if;
											DRAM_DQ		<= sd_data(15 downto 0) & sd_data(15 downto 0);
										else -- 32 bit
											DRAM_DQM		<= "0000";
											DRAM_DQ		<= sd_data;
										end if;
										
				when SD_ACTIVATE	=>
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '0';
										DRAM_CAS_N	<= '1';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= sd_cmd_ba;	-- bank sel
										DRAM_ADDR	<= sd_cmd_row;	-- col select
										
				when SD_PRECHARGE	=> 
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '0';
										DRAM_CAS_N	<= '1';
										DRAM_WE_N	<= '0';
										DRAM_BA		<= sd_cmd_ba;	-- bank sel
										DRAM_ADDR	<= (others => '0'); --A10 = low others dont care
				
				when SD_PRECHARGE_ALL =>
				  
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '0';
										DRAM_CAS_N	<= '1';
										DRAM_WE_N	<= '0';
										DRAM_BA		<= (others => '0');
										DRAM_ADDR	<= (others => '1'); --A10 = high (precharge all)
						
				when SD_REFRESH	=> 
				
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '0';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= (others => '0');	-- dont care here
										DRAM_ADDR	<= (others => '0');	-- dont care here
										
				when SD_MODREGSET =>
				  
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '0';
										DRAM_CAS_N	<= '0';
										DRAM_WE_N	<= '0';
										DRAM_BA		<= "00";	
										DRAM_ADDR	<= sd_cmd_row;	-- A10 have to be 0
										
				when others		=> -- equal to nop
										DRAM_CS_N 	<= '0';
										DRAM_RAS_N	<= '1';
										DRAM_CAS_N	<= '1';
										DRAM_WE_N	<= '1';
										DRAM_BA		<= (others => '0');	-- dont care here
										DRAM_ADDR	<= (others => '0');	-- dont care here			
			end case;
		end if;
	end process;
	
	-- read delay stage 1..2
	
	process(clk)
	begin
		if rising_edge(clk) then
				rs1_push <= rs0_push;
				rs1_data	<= rs0_data;
				rs2_push <= rs1_push;
				rs2_data	<= rs1_data;
				rs3_push <= rs2_push;
				rs3_data	<= rs2_data;
		end if;
	end process;
	
	-- read capture
	
	process(clk)
		variable tmp_data : std_logic_vector(31 downto 0);
		variable tmp_fifo : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
							
			-- default
			
			for i in 0 to mem_sdram_reciver_count-1 loop
				send_xn_wrreq(i) <= '0';
			end loop;
			
			--
			
			if rs3_push = '1' and reset = '0' then

				tmp_fifo 		:= rs3_data;	

				if tmp_fifo.dmode = '1' then -- 16 bit
					if tmp_fifo.address(0) = '1' then
							tmp_fifo.data  := DRAM_DQ;							
						else
							tmp_fifo.data  := x"0000" & DRAM_DQ(31 downto 16);
					end if;
				else -- 32 bit
					tmp_fifo.data  := DRAM_DQ;
				end if;
				
				send_xn_wrreq(conv_integer(tmp_fifo.id))	<= '1'; -- push

			end if;		
		end if;
		
		-- set to all fifo inputs (send_xn_wrreq arbit write)
			
		for i in 0 to mem_sdram_reciver_count-1 loop
			send_xn_data(i) <= port_to_raw(tmp_fifo);				
		end loop;
			
	end process;
	
	-- full generate
		
	process(clk)
	begin
		for i in 0 to mem_sdram_reciver_count-1 loop
			if rising_edge(clk) then
					if send_xn_wrusedw(i) >=  x"f0" then -- fb
						send_xn_wrfull(i) <= '1';
					else
						send_xn_wrfull(i) <= '0';
					end if;
				end if;		
			end loop;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	---- SEND DATA 
	-----------------------------------------------------------------------------------------------------------
	
	 dcfifo_loop : for i in 0 to mem_sdram_reciver_count-1 generate
         dcfifo : dc_fifo 
					generic map
					(
						FIFO_DEPTH 			=> 256,
						FIFO_WIDTH 			=> 72,
						FIFO_SHOWAHEAD 	=> "ON",
						FIFO_USEDWWIDTH	=> 8,
						FIFO_SYNCSTAGES	=> 2
					)  
					port map
               (
						'0',
						send_xn_data(i),
						from_sdram_rclk(i),		-- read with component clk
						from_sdram_ack(i),	
						clk,					-- write with clk
						send_xn_wrreq(i),	
						from_sdram_data(i),
						from_sdram_busy(i),
						send_xn_wrusedw(i)
					);
   end generate;
	
	-- refresh counter
	
	process(clk)
	begin
		if rising_edge(clk) then
			
			refresh_counter <= refresh_counter + 1;
			
			if ack_refresh = '1' then
				do_refresh		 <= '0';
			end if;
			
			if refresh_counter >= tref then
				refresh_counter <= 0;
				do_refresh		 <= '1';
			end if;			
		end if;
	end process;
	
	
end rtl;		