library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity sdData_io_channel is
		port 
		(
				-- global

				clk						: in std_logic;
				reset						: in std_logic;	

				-- SDCARD 
				
				sd_data					: in std_logic;
				
				-- clk in
				
				clk_read_edge			: in std_logic;
				
				-- recive
				
				recive_reset				: in std_logic;
				recive_fifo_usage			: out std_logic_vector (12 DOWNTO 0);				
				recive_data_in_cnt		: in integer range 0 to 8191; -- cnt + 1			
				recive_crc_pass			: out std_logic;
				recive_finish				: out std_logic;
				recive_error				: out std_logic;
				recive_blocks_readed		: out integer range 0 to 65535;
								
				recive_data					: out std_logic;
				recive_data_available	: out std_logic;
				recive_data_ack			: in std_logic				
				
		);
end entity;

architecture rtl of sdData_io_channel is

-- crc16 in

component sdCtrl_crc16t is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	

				--  
				
				dclr				: in std_logic;
				din				: in std_logic;
				din_en			: in std_logic;
				dout 				: out std_logic_vector(15 downto 0)

		);
end component;


signal crc16_in_dclr		: std_logic := '0';
signal crc16_in_din		: std_logic := '0';
signal crc16_in_din_en	: std_logic := '0';
signal crc16_in_dout		: std_logic_vector(15 downto 0) := (others => '0');

-- fifo in

--component sdData_io_in_fifo
--	PORT
--	(
--		clock		: IN STD_LOGIC ;
--		data		: IN STD_LOGIC_VECTOR (0 DOWNTO 0);
--		rdreq		: IN STD_LOGIC ;
--		sclr		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		almost_empty		: OUT STD_LOGIC ;
--		almost_full		: OUT STD_LOGIC ;
--		empty		: OUT STD_LOGIC ;
--		full		: OUT STD_LOGIC ;
--		q		: OUT STD_LOGIC_VECTOR (0 DOWNTO 0);
--		usedw		: OUT STD_LOGIC_VECTOR (12 DOWNTO 0)
--	);
--end component;

component sc_fifo_usedw
	GENERIC 
	(
		FIFO_DEPTH 			: natural;
		FIFO_WIDTH 			: natural;
		FIFO_SHOWAHEAD 	: string;
		FIFO_USERAM			: string;
		FIFO_USEDWWIDTH	: natural
		
	);
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		usedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
end component;

signal fifo_in_data				: STD_LOGIC_VECTOR (0 DOWNTO 0) := (others => '0');
signal fifo_in_rdreq				: STD_LOGIC := '0';
signal fifo_in_reset				: std_logic := '0';
signal fifo_in_wrreq				: STD_LOGIC := '0';
signal fifo_in_empty				: STD_LOGIC := '0';
signal fifo_in_full				: STD_LOGIC := '0';
signal fifo_in_q					: STD_LOGIC_VECTOR (0 DOWNTO 0) := (others => '0');
signal fifo_in_usedw				: STD_LOGIC_VECTOR (12 DOWNTO 0) := (others => '0');

--

signal recive_crc16_sampled	: std_logic_vector(15 downto 0) := (others => '0');
signal recive_state_cnt 		: integer range 0 to 20 := 0;
signal recive_sub_state_cnt  	: integer range 0 to 8191 := 0;
signal recive_block_cnt			: integer range 0 to 65535 := 0;

--

signal sd_read					: std_logic := '1';

begin
	
	recive_blocks_readed <= recive_block_cnt;
	
	-----------------------------------------------------------------------------------------------------------
	-- crc16
	-----------------------------------------------------------------------------------------------------------
	
	crc16_in : sdCtrl_crc16t port map 
	(
		clk,
		reset,
		crc16_in_dclr,	
		crc16_in_din,		
		crc16_in_din_en,	
		crc16_in_dout	
	);
	
	-----------------------------------------------------------------------------------------------------------
	-- in fifo
	-----------------------------------------------------------------------------------------------------------
	
	fifo_in_reset <= reset;
	
	fifo_in : sc_fifo_usedw
	generic map
	(
		FIFO_DEPTH 			=> 8192,
		FIFO_WIDTH 			=> 1,
		FIFO_SHOWAHEAD 	=> "ON",
		FIFO_USERAM			=> "ON",
		FIFO_USEDWWIDTH	=> 13				
	)
	port map
	(
		clk,
		fifo_in_data,
		fifo_in_rdreq,
		fifo_in_reset,
		fifo_in_wrreq,
		fifo_in_empty,			
		fifo_in_full,			
		fifo_in_q,				
		fifo_in_usedw			
	);

	
	recive_fifo_usage	<= fifo_in_usedw;
	recive_data			<= fifo_in_q(0);
	recive_data_available	<= not fifo_in_empty;
	
	
	fifo_in_rdreq		<= recive_data_ack;
	
	-----------------------------------------------------------------------------------------------------------
	-- data in
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' or recive_reset = '1' then
				sd_read <= '1';
			else
				sd_read <= sd_data;
			end if;
		end if;
	end process;
	
	process(clk)
		variable t_crc_ok : std_logic := '0';
	begin
		if rising_edge(clk) then
			if reset = '1' or recive_reset = '1' then
				recive_finish 		<= '0';
				recive_error 		<= '0';
				recive_crc_pass 	<= '0';
				recive_state_cnt	<= 0;
				recive_block_cnt	<= 0;
				crc16_in_dclr		<= '1';
			else
			
				-- default
				
				crc16_in_din_en 	<= '0';
				crc16_in_dclr		<= '0';	
				
				fifo_in_wrreq		<= '0';
				
				--
			
				if clk_read_edge = '1' then
					if recive_state_cnt = 0 then	-- wait for data
						if sd_read = '0' then
							recive_state_cnt 		<= 1; -- next recive_state_cnt
							recive_sub_state_cnt	<= 0;
							
							crc16_in_din_en 	<= '1';
							crc16_in_din		<= '0';	
							
						end if;
					elsif recive_state_cnt = 1 then
					
						-- fifo (no check current)
						
						
						fifo_in_wrreq		<= '1';
						fifo_in_data(0) 	<= sd_read;
					
						-- crc
						
						crc16_in_din_en 	<= '1';
						crc16_in_din		<= sd_read;	
					
						if recive_sub_state_cnt = recive_data_in_cnt then
							recive_state_cnt 		<= 2; -- finished
						else
							recive_sub_state_cnt <= recive_sub_state_cnt + 1;
						end if;
					
					elsif recive_state_cnt >=2 and recive_state_cnt<=17 then -- crc sample
						
						recive_crc16_sampled(15-(recive_state_cnt-2)) <= sd_read;
						
						recive_state_cnt  <= recive_state_cnt + 1;
						
					elsif recive_state_cnt = 18 then -- wait for stop
					
						if recive_crc16_sampled = crc16_in_dout then
							t_crc_ok := '1';
						else
							t_crc_ok := '0';
						end if;
						
						recive_crc_pass 	<= t_crc_ok;
						recive_finish  	<= '1';
						
						if sd_read = '1'  then --stop 															
							recive_state_cnt 	<= 0; -- restart
							recive_block_cnt 	<= recive_block_cnt + 1; -- if pass cnt block up
						else														
							recive_state_cnt 	<= 19; -- stop and error
						end if;
						
						
						crc16_in_dclr		<= '1';	-- clear crc 16 after this
						
					else -- 19
						recive_error 		<= '1';
					end if;
								
				end if;
			
			end if;
		end if;
	end process;


end rtl;
