library ieee;
use work.bus_package.all;
use work.hcore_package.all;
use work.video_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl is
		port 
		(
				-- 
			
				clk				: in std_logic;
				reset				: in std_logic;
				clk_audio		: in std_logic;
				reset_audio		: in std_logic;
				reset_fall_audio	: in std_logic;
				
				-- utility core port
				
				ctrl_port_a_adress  	: in std_logic_vector(6 downto 0);
				ctrl_port_a_q		   : out std_logic_vector(15 downto 0);
				ctrl_port_a_data 		: in std_logic_vector(15 downto 0);
				ctrl_port_a_we			: in std_logic;
		
				--
		
				irq_out					: out std_logic;
				
				-- hcore config
				
				hcore_clk				: in std_logic;
				hcore_data				: out hcore_ctrl_struct_t;
				hcore_busy				: out std_logic;
				hcore_ack				: in std_logic;
				
				-- from hcores
				
				i_from_hcores_data		: in hcore_to_uctrl_struct_t;
				i_from_hcores_empty		: in std_logic;
				o_from_hcores_rdreq		: out std_logic;
		
				-- to hcores
				
				o_to_hcores_data			: out uctrl_to_hcore_struct_t;
				o_to_hcores_empty			: out std_logic;
				i_to_hcores_rdreq			: in std_logic;
		
				-- SW 
				
				SW					: in std_logic_vector(15 downto 0);

				-- SEG7 
				HEX0				: out std_logic_vector(6 downto 0);
				HEX1				: out std_logic_vector(6 downto 0);
				HEX2				: out std_logic_vector(6 downto 0);
				HEX3				: out std_logic_vector(6 downto 0);
				HEX4				: out std_logic_vector(6 downto 0);
				HEX5				: out std_logic_vector(6 downto 0);
				HEX6				: out std_logic_vector(6 downto 0);
				HEX7				: out std_logic_vector(6 downto 0);		

				-- LED 
				
				LEDG				: out std_logic_vector(8 downto 0);
				LEDR				: out std_logic_vector(8 downto 0);
			
				-- LCD 
				LCD_BLON			: out std_logic;
				LCD_DATA			: inout std_logic_vector(7 downto 0);
				LCD_EN			: out std_logic;
				LCD_ON			: out std_logic;
				LCD_RS			: out std_logic;
				LCD_RW			: out std_logic;
				
				-- VGA 
				VGA_B				: out std_logic_vector(7 downto 0);
				VGA_BLANK_N		: out std_logic;
				VGA_CLK			: out std_logic;
				VGA_G				: out std_logic_vector(7 downto 0);
				VGA_HS			: out std_logic;
				VGA_R				: out std_logic_vector(7 downto 0);
				VGA_SYNC_N		: out std_logic;
				VGA_VS			: out std_logic;
				
				--PS2
				
				PS2_CLK			: inout std_logic;
				PS2_CLK2			: inout std_logic;
				PS2_DAT			: inout std_logic;
				PS2_DAT2			: inout std_logic;
				
				-- I2C for EEPROM 
				EEP_I2C_SCLK	: out std_logic;
				EEP_I2C_SDAT	: inout std_logic;

				-- I2C for Audio Tv-Decoder HSMC  
				I2C_SCLK			: out std_logic;
				I2C_SDAT			: inout std_logic;
				
				-- AUDIO
				
				AUD_ADCDAT		: in std_logic;
				AUD_ADCLRCK		: inout std_logic;
				AUD_DACDAT		: out std_logic;
				AUD_DACLRCK		: inout std_logic;
				
				-- external
				
				-- send
				
				ext_send_address	: out std_logic_vector(31 downto 0);
				ext_send_busy		: in std_logic;
				ext_send_wrreq		: out std_logic;
				
				-- recive
				
				ext_recive_data	: in std_logic_vector(31 downto 0);
				ext_recive_busy	: in std_logic;
				ext_recive_ack		: out std_logic;		
			
				-- errors
				
				error_audio_dac_no_data			: out std_logic;
				error_audio_dac_full				: out std_logic;
				error_audio_adc_no_data			: out std_logic;
				error_audio_adc_full				: out std_logic
		
		);
end entity;



architecture rtl of utility_ctrl is

component utility_ctrl_audio is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- audio 
				
				clk_audio		: in std_logic;
				reset_audio		: in std_logic;
				reset_fall_audio	: in std_logic;
				
				-- dac fifo
				
				dac_fifo_data			: in std_logic_vector(63 downto 0); 
				dac_fifo_wreq			: in std_logic;
				dac_used_low_irq		: in std_logic_vector(7 downto 0);
				dac_irq					: out std_logic;
				
				-- adc fifo
				
				adc_fifo_data		: out std_logic_vector(47 downto 0);
				adc_fifo_rdreq		: in std_logic;
				adc_used_high_irq	: in std_logic_vector(7 downto 0);
				adc_irq				: out std_logic;
				
				-- error
				
				error_dac_no_data			: out std_logic;
				error_dac_full				: out std_logic;
				
				error_adc_no_data			: out std_logic;
				error_adc_full				: out std_logic;
				
				-- wm8731
				
				AUD_ADCDAT		: in std_logic;
				AUD_ADCLRCK		: inout std_logic;
				AUD_DACDAT		: out std_logic;
				AUD_DACLRCK		: inout std_logic															
		);
end component;

component utility_ctrl_lcd is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				--
				
				lcd_data_write  : in std_logic := '0';
				lcd_data_read	 : in std_logic := '0';
				lcd_data_rs     : in std_logic;
				lcd_data_data   : in std_logic_vector(7 downto 0);
				lcd_data_q	    : out std_logic_vector(7 downto 0);
				lcd_wait			 : out std_logic;
				
				-- LCD 
				
				LCD_BLON			: out std_logic;
				LCD_DATA			: inout std_logic_vector(7 downto 0);
				LCD_EN			: out std_logic;
				LCD_ON			: out std_logic;
				LCD_RS			: out std_logic;
				LCD_RW			: out std_logic

		);
end component;

	-- lcd

	signal lcd_data_rs    : std_logic := '0';
	signal lcd_data_data  : std_logic_vector(7 downto 0) := (others => '0');
	signal lcd_data_q	    : std_logic_vector(7 downto 0) := (others => '0');
	signal lcd_data_write : std_logic := '0';
	signal lcd_data_read	 : std_logic := '0';
	signal lcd_wait		 : std_logic := '0';


	-- timer
	
component utility_ctrl_timer is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				--
				
				timer_load_value 		: in std_logic_vector(31 downto 0);
				timer_start 			: in std_logic := '0';
				timer_stop  			: in std_logic := '0';
				timer_auto_reload  	: in std_logic := '0';
				timer_hit  				: out std_logic := '0';
				
				timer_current_value	: out std_logic_vector(31 downto 0)
				
				
		
		);
end component;
	
	signal timer1_tmp_load_value : std_logic_vector(15 downto 0) 	:= (others => '0');
	signal timer1_load_value : std_logic_vector(31 downto 0) 		:= (others => '0');
	signal timer1_start : std_logic := '0';
	signal timer1_stop  : std_logic := '0';
	signal timer1_auto_reload  : std_logic := '0';
	signal timer1_hit  : std_logic := '0';
	signal timer1_current_value : std_logic_vector(31 downto 0) 		:= (others => '0');
	signal timer1_captured_value : std_logic_vector(31 downto 0) 		:= (others => '0');
	
	signal timer2_tmp_load_value : std_logic_vector(15 downto 0) 	:= (others => '0');
	signal timer2_load_value : std_logic_vector(31 downto 0) 		:= (others => '0');
	signal timer2_start : std_logic := '0';
	signal timer2_stop  : std_logic := '0';
	signal timer2_auto_reload  : std_logic := '0';
	signal timer2_hit  : std_logic := '0';
	signal timer2_current_value : std_logic_vector(31 downto 0) 		:= (others => '0');
	signal timer2_captured_value : std_logic_vector(31 downto 0) 		:= (others => '0');

	-- vga
	
component utility_ctrl_vga is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset		: in std_logic;
						
				-- VGA 
				
				VGA_B				: out std_logic_vector(7 downto 0);
				VGA_BLANK_N		: out std_logic;
				VGA_CLK			: out std_logic;
				VGA_G				: out std_logic_vector(7 downto 0);
				VGA_HS			: out std_logic;
				VGA_R				: out std_logic_vector(7 downto 0);
				VGA_SYNC_N		: out std_logic;
				VGA_VS			: out std_logic;
				
				--
				
				HSYNC				: in integer range 0 to 4095;
				HSTART			: in integer range 0 to 4095;
				HMEMSTART		: in integer range 0 to 4095;
				HSTOP				: in integer range 0 to 4095;
				HTOTAL			: in integer range 0 to 4095;

				HLOADSTART		: in integer range 0 to 4095;
				HLOADSTOP		: in integer range 0 to 4095;

				VSYNC				: in integer range 0 to 2047;
				VSTART			: in integer range 0 to 2047;
				VSTOP				: in integer range 0 to 2047;
				VTOTAL			: in integer range 0 to 2047;

				VLOADSTART		: in integer range 0 to 4095;
				VLOADSTOP		: in integer range 0 to 4095;
				VLOADNEXT		: in integer range 0 to 4095;

				MEM_STARTADR	: in std_logic_vector(31 downto 0);
				MEM_LINEOFFSET	: in std_logic_vector(31 downto 0);	
				MEM_MODE			: in std_logic_vector(15 downto 0);

				video_on			: in std_logic;
				lc_add			: in std_logic_vector(31 DOWNTO 0);
				lc_start			: in std_logic_vector(10 DOWNTO 0);
				
				video_hcount	: out integer range 0 to 4095;
				video_vcount	: out integer range 0 to 2047;
				
				PVCOUNT			: in integer range 0 to 2047;
				PVCOUNT_WE		: in std_logic;
				PVCOUNT_HIT		: out std_logic;
				
				-- external
				
				-- send
				
				ext_send_address	: out std_logic_vector(31 downto 0);
				ext_send_busy		: in std_logic;
				ext_send_wrreq		: out std_logic;
				
				-- recive
				
				ext_recive_data	: in std_logic_vector(31 downto 0);
				ext_recive_busy	: in std_logic;
				ext_recive_ack		: out std_logic	
		
		);
end component;

	signal HSYNC		: integer range 0 to 4095	:= HSYNC_default;	
	signal HSTART		: integer range 0 to 4095	:= HSTART_default;
	signal HMEMSTART	: integer range 0 to 4095	:= HMEMSTART_default;
	signal HSTOP		: integer range 0 to 4095	:= HSTOP_default;
	signal HTOTAL		: integer range 0 to 4095	:= HTOTAL_default;
	
	signal HLOADSTART	: integer range 0 to 4095  := HLOADSTART_default;		-- external memory start
	signal HLOADSTOP	: integer range 0 to 4095  := HLOADSTOP_default;
	
	signal VSYNC		: integer range 0 to 2047	:= VSYNC_default;	
	signal VSTART		: integer range 0 to 2047	:= VSTART_default;
	signal VSTOP		: integer range 0 to 2047	:= VSTOP_default;
	signal VTOTAL		: integer range 0 to 2047	:= VTOTAL_default;	
	
	signal VLOADSTART	: integer range 0 to 4095  := VLOADSTART_default;		-- external memory start
	signal VLOADSTOP	: integer range 0 to 4095  := VLOADSTOP_default; 
	signal VLOADNEXT	: integer range 0 to 4095  := VLOADNEXT_default;	
	
	signal MEM_STARTADR		: std_logic_vector(31 downto 0) := MEM_STARTADR_default;		-- external memory start
	signal MEM_LINEOFFSET	: std_logic_vector(31 downto 0) := MEM_LINEOFFSET_default;		
	signal MEM_MODE			: std_logic_vector(15 downto 0) := MEM_MODE_default;
	
	signal video_on	: std_logic := VIDEO_ON_default;
	signal lc_add		: std_logic_vector(31 DOWNTO 0) := lc_add_default;
	signal lc_start	: std_logic_vector(10 DOWNTO 0) := lc_start_default;
		
	signal video_hcount	: integer range 0 to 4095 := 0;
	signal video_vcount	: integer range 0 to 2047 := 0;
	
	signal PVCOUNT			: integer range 0 to 2047 := 0;
	signal PVCOUNT_WE		: std_logic := '0';
	signal PVCOUNT_HIT	: std_logic := '0';
	
-- ps2

component utility_ctrl_ps2 is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				-- data out
				
				data_out			: out std_logic_vector(7 downto 0);
				data_out_ready : out std_logic;
				data_out_ack	: in std_logic;
				
				-- data in
				
				data_in					: in std_logic_vector(7 downto 0);
				data_in_we				: in std_logic;
				data_in_ready			: out std_logic;
				data_in_wrs_finish	: out std_logic;
				
				--
				
				ps2_clk_inout			: inout std_logic;
				ps2_dat_inout			: inout std_logic
									
		);
end component;	

signal ps2_a_data_out			: std_logic_vector(7 downto 0) := (others => '0');
signal ps2_a_data_out_ready	: std_logic := '0';
signal ps2_a_data_out_ack		: std_logic := '0';

signal ps2_a_data_in			: std_logic_vector(7 downto 0) := (others => '0');
signal ps2_a_data_in_we		: std_logic := '0';
signal ps2_a_data_in_ready		: std_logic := '0';		
signal ps2_a_data_in_wrs_finish		: std_logic := '0';		


signal ps2_b_data_out			: std_logic_vector(7 downto 0) := (others => '0');
signal ps2_b_data_out_ready	: std_logic := '0';
signal ps2_b_data_out_ack		: std_logic := '0';

signal ps2_b_data_in			: std_logic_vector(7 downto 0) := (others => '0');
signal ps2_b_data_in_we		: std_logic := '0';
signal ps2_b_data_in_ready		: std_logic := '0';		
signal ps2_b_data_in_wrs_finish		: std_logic := '0';		

component i2c_port is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- ext
				
				I2C_SCLK			: out std_logic;
				I2C_SDAT			: inout std_logic;
				
				-- internal
				
				clk_in			: in std_logic;
				direction_in	: in std_logic;
				data_write		: in std_logic;
				data_read		: out std_logic															
		);
end component;

signal i2c_a_clk_in			: std_logic := '1';
signal i2c_a_direction_in	: std_logic := '0';
signal i2c_a_data_write		: std_logic := '0';
signal i2c_a_data_read		: std_logic	:= '0';
	
signal i2c_b_clk_in			: std_logic := '1';
signal i2c_b_direction_in	: std_logic := '0';
signal i2c_b_data_write		: std_logic := '0';
signal i2c_b_data_read		: std_logic	:= '0';	
	
-- irq
	
signal irq_msk 				: std_logic_vector(15 downto 0) := (others => '0');
signal irq_memory 			: std_logic_vector(15 downto 0) := (others => '0');
signal clear_irq_stored 	: std_logic := '0';
signal clear_irq_value		: std_logic_vector(15 downto 0) := (others => '0');
	
-- audio

signal dac_fifo_data			: std_logic_vector(63 downto 0) := (others => '0'); 
signal dac_fifo_wreq			: std_logic := '0';
signal dac_used_low_irq		: std_logic_vector(7 downto 0) := x"80"; -- 128
signal dac_irq					: std_logic := '0';	

signal adc_fifo_data			: std_logic_vector(47 downto 0) := (others => '0'); 
signal adc_fifo_rdreq		: std_logic := '0';
signal adc_used_high_irq	: std_logic_vector(7 downto 0)  := x"80"; -- 128
signal adc_irq					: std_logic := '0';	

-- hcore ctrl

--component dcfifo_hcore
--	PORT
--	(
--		data		: IN STD_LOGIC_VECTOR (39 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q		: OUT STD_LOGIC_VECTOR (39 DOWNTO 0);
--		rdempty		: OUT STD_LOGIC ;
--		wrusedw		: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--end component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;


signal hcore_ctrl_fifo_data		: std_logic_vector (39 DOWNTO 0) := (others => '0'); 
signal hcore_ctrl_fifo_rdclk		: std_logic := '0';	
signal hcore_ctrl_fifo_rdreq		: std_logic := '0';	
signal hcore_ctrl_fifo_wrclk		: std_logic := '0';	
signal hcore_ctrl_fifo_wrreq		: std_logic := '0';	
signal hcore_ctrl_fifo_q			: std_logic_vector (39 DOWNTO 0) := (others => '0'); 
signal hcore_ctrl_fifo_rdempty	: std_logic := '0';	
signal hcore_ctrl_fifo_wrusedw	: std_logic_vector (7 DOWNTO 0) := (others => '0'); 

signal hcore_ctrl_fifofull		: std_logic := '1';	
signal hcore_ctrl_data		: hcore_ctrl_struct_t := (others => (others => '0'));

-- hcore data

signal hcore_data_fifo_data		: std_logic_vector (39 DOWNTO 0) := (others => '0'); 
signal hcore_data_fifo_rdclk		: std_logic := '0';	
signal hcore_data_fifo_rdreq		: std_logic := '0';	
signal hcore_data_fifo_wrclk		: std_logic := '0';	
signal hcore_data_fifo_wrreq		: std_logic := '0';	
signal hcore_data_fifo_q			: std_logic_vector (39 DOWNTO 0) := (others => '0'); 
signal hcore_data_fifo_rdempty	: std_logic := '0';	
signal hcore_data_fifo_wrusedw	: std_logic_vector (7 DOWNTO 0) := (others => '0'); 

signal hcore_data_fifofull			:  std_logic := '1';	
--	

constant MODELSIM : boolean := false
-- synthesis translate_off
     	or true
-- synthesis translate_on
	;

--

signal to_ctrl_port : std_logic_vector (15 DOWNTO 0) := (others => '0');

--
	
begin

	hcore_ctrl_fifo_data <= hcore_ctrl_to_raw(hcore_ctrl_data);

	-- to hcore (ctrl (reset/start pc etc.))
		
	hcore_ctrl_fifo : dc_fifo 
	generic map
	(
		FIFO_DEPTH 			=> 256,
		FIFO_WIDTH 			=> 40,
		FIFO_SHOWAHEAD 	=> "ON",
		FIFO_USEDWWIDTH	=> 8,
		FIFO_SYNCSTAGES	=> 3
	)   
	port map
	(
		'0',
		hcore_ctrl_fifo_data,	
		hcore_ctrl_fifo_rdclk,	
		hcore_ctrl_fifo_rdreq,	
		hcore_ctrl_fifo_wrclk,	
		hcore_ctrl_fifo_wrreq,	
		hcore_ctrl_fifo_q,		
		hcore_ctrl_fifo_rdempty,
		hcore_ctrl_fifo_wrusedw
	);
	
	hcore_ctrl_fifo_wrclk	<= clk;
	hcore_data 					<= raw_to_hcore_ctrl(hcore_ctrl_fifo_q);
	hcore_ctrl_fifo_rdclk	<= hcore_clk;
	hcore_ctrl_fifo_rdreq	<= hcore_ack;
	hcore_busy					<= hcore_ctrl_fifo_rdempty;

	process(clk)
	begin 
		if rising_edge(clk) then
			if reset = '0' and hcore_ctrl_fifo_wrusedw < x"fc" then
				hcore_ctrl_fifofull <= '0';
			else
				hcore_ctrl_fifofull <= '1';
			end if;
		end if;
	end process;
			
	-- to hcore data (response)
		
	hcore_data_fifo : dc_fifo 
	generic map
	(
		FIFO_DEPTH 			=> 256,
		FIFO_WIDTH 			=> 40,
		FIFO_SHOWAHEAD 	=> "ON",
		FIFO_USEDWWIDTH	=> 8,
		FIFO_SYNCSTAGES	=> 3
	)   
	port map
	(
		'0',
		hcore_data_fifo_data,	
		hcore_data_fifo_rdclk,	
		hcore_data_fifo_rdreq,	
		hcore_data_fifo_wrclk,	
		hcore_data_fifo_wrreq,	
		hcore_data_fifo_q,		
		hcore_data_fifo_rdempty,
		hcore_data_fifo_wrusedw
	);
		
	hcore_data_fifo_wrclk	<= clk;
	o_to_hcores_data 			<= raw_to_from_uctrl(hcore_data_fifo_q);
	hcore_data_fifo_rdclk	<= hcore_clk;
	hcore_data_fifo_rdreq	<= i_to_hcores_rdreq;
	o_to_hcores_empty			<= hcore_data_fifo_rdempty;

	process(clk)
	begin 
		if rising_edge(clk) then
			if reset = '0' and hcore_data_fifo_wrusedw < x"fc" then
				hcore_data_fifofull <= '0';
			else
				hcore_data_fifofull <= '1';
			end if;
		end if;
	end process;
	
	-- audio
	
	audio : utility_ctrl_audio port map
	(
		clk,
		reset,
		clk_audio,		
		reset_audio,
		reset_fall_audio,
		dac_fifo_data,
		dac_fifo_wreq,
		dac_used_low_irq,
		dac_irq,
		adc_fifo_data,			
		adc_fifo_rdreq,	
		adc_used_high_irq,	
		adc_irq,					
		error_audio_dac_no_data,
		error_audio_dac_full,
		error_audio_adc_no_data,		
		error_audio_adc_full,			
		AUD_ADCDAT,		
		AUD_ADCLRCK,		
		AUD_DACDAT,		
		AUD_DACLRCK																
	);

	-- i2c ports
	
	i2c_port_a : i2c_port port map
	(
		clk,
		reset,
		EEP_I2C_SCLK,
		EEP_I2C_SDAT,
		i2c_a_clk_in,			
		i2c_a_direction_in,	
		i2c_a_data_write,		
		i2c_a_data_read		
	);
	
	i2c_port_b : i2c_port port map
	(
		clk,
		reset,
		I2C_SCLK,
		I2C_SDAT,
		i2c_b_clk_in,			
		i2c_b_direction_in,	
		i2c_b_data_write,		
		i2c_b_data_read		
	);

	-- ps2
	-- port a
	
	ps2_port_a : utility_ctrl_ps2 port map
	(
		clk,
		reset,
		ps2_a_data_out,		
		ps2_a_data_out_ready,
		ps2_a_data_out_ack,
		ps2_a_data_in,	
		ps2_a_data_in_we,
		ps2_a_data_in_ready,
		ps2_a_data_in_wrs_finish,
		PS2_CLK,
		PS2_DAT	
	);
	
	-- port b
	ps2_port_b : utility_ctrl_ps2 port map
	(
		clk,
		reset,
		ps2_b_data_out,		
		ps2_b_data_out_ready,
		ps2_b_data_out_ack,
		ps2_b_data_in,	
		ps2_b_data_in_we,
		ps2_b_data_in_ready,
		ps2_b_data_in_wrs_finish,
		PS2_CLK2,
		PS2_DAT2	
	);

	-- timers
	
	timer1 : utility_ctrl_timer port map 
	( 
			clk,
			reset,
			timer1_load_value,
			timer1_start,
			timer1_stop,
			timer1_auto_reload,
			timer1_hit,
			timer1_current_value
	);
	
	timer2 : utility_ctrl_timer port map 
	( 
			clk,
			reset,
			timer2_load_value,
			timer2_start,
			timer2_stop,
			timer2_auto_reload,
			timer2_hit,
			timer2_current_value
	);
	
	-- lcd
	
	lcd : utility_ctrl_lcd  port map 
	(			
			clk,
			reset,
			lcd_data_write,
			lcd_data_read,
			lcd_data_rs,
			lcd_data_data,
			lcd_data_q,
			lcd_wait,
			LCD_BLON,
			LCD_DATA,
			LCD_EN,
			LCD_ON,
			LCD_RS,
			LCD_RW
	);
	
	-- vga
	
	vga : utility_ctrl_vga port map 
	(			
			clk,
			reset,			
			VGA_B,
			VGA_BLANK_N,
			VGA_CLK,
			VGA_G,
			VGA_HS,
			VGA_R,
			VGA_SYNC_N,
			VGA_VS,	
			HSYNC,
			HSTART,
			HMEMSTART,
			HSTOP,
			HTOTAL,
			HLOADSTART,
			HLOADSTOP,
			VSYNC,
			VSTART,
			VSTOP,
			VTOTAL,	
			VLOADSTART,
			VLOADSTOP,
			VLOADNEXT,
			MEM_STARTADR,
			MEM_LINEOFFSET,	
			MEM_MODE,
			video_on,
			lc_add,
			lc_start,	
			video_hcount,
			video_vcount,
			PVCOUNT,
			PVCOUNT_WE,
			PVCOUNT_HIT,
			ext_send_address,
			ext_send_busy,
			ext_send_wrreq,
			ext_recive_data,
			ext_recive_busy,
			ext_recive_ack
			
	);
	
	-----------------------------------------------------------------------------------------------------------						
	-- irq generate
	-----------------------------------------------------------------------------------------------------------	

	process ( clk )
		variable irq_current : std_logic_vector(15 downto 0)  := (others => '0');
		variable irq_stored  : std_logic_vector(15 downto 0) := (others => '0');
	begin
		if rising_edge(clk) then	
			if reset = '1' then
				irq_stored := (others => '0');
				irq_out <= '0'; 				
			else
				-- default 

				if clear_irq_stored = '1' then
					irq_stored := irq_stored and clear_irq_value;
					irq_out <= '0'; 
				end if;
		
				--
				
				irq_current := "00000000000" & (not i_from_hcores_empty) & adc_irq & dac_irq & timer2_hit & timer1_hit;	-- irqs that can occure
				
				if(irq_current and irq_msk) /= x"0000" then -- irq occure (masked)
					irq_stored 	:= irq_stored or irq_current;
					irq_out 		<= '1' ;
				end if;
				
			end if;
			
			irq_memory <= irq_stored;
		end if;
	end process;	
	
	-----------------------------------------------------------------------------------------------------------	
	-- ucore port
	-----------------------------------------------------------------------------------------------------------	
	
	process(clk)
	begin
		if rising_edge(clk) then
			ctrl_port_a_q <= to_ctrl_port;	-- one delay like iram
		end if;
	end process;
	
	process ( clk )
		variable temp_from_uctrl_data	: uctrl_to_hcore_struct_t := raw_to_from_uctrl((others => '0'));
		variable lc_add_store			: std_logic_vector(31 DOWNTO 0) := lc_add_default;
		variable MEM_STARTADR_store	: std_logic_vector(31 DOWNTO 0) := MEM_STARTADR_default;		
		variable MEM_LINEOFFSET_store	: std_logic_vector(31 DOWNTO 0) := MEM_LINEOFFSET_default;	
		variable MEM_MODE_store			: std_logic_vector(15 DOWNTO 0) := MEM_MODE_default;			
	begin
		if rising_edge(clk) then	
			if reset = '1' then
				HEX0 <= (others => '1');
				HEX1 <= (others => '1');
				HEX2 <= (others => '1');
				HEX3 <= (others => '1');
				HEX4 <= (others => '1');
				HEX5 <= (others => '1');
				HEX6 <= (others => '1');
				HEX7 <= (others => '1');	
				LEDG <= (others => '0');
				LEDR <= (others => '0');							
			
				to_ctrl_port <= (others => '0');
				
				lcd_data_write <= '0';
				lcd_data_read	<= '0';
				
				timer1_start <= '0';
				timer1_stop  <= '0';
				timer2_start <= '0';
				timer2_stop  <= '0';
				
				irq_msk <= (others => '0');
				
				-- vga
								
				HSYNC <= HSYNC_default;			
				HSTART <= HSTART_default;
				HMEMSTART <= HMEMSTART_default;		
				HSTOP <= HSTOP_default;			
				HTOTAL <= HTOTAL_default;			

				VSYNC <= VSYNC_default;			
				VSTART <= VSTART_default;			
				VSTOP <= VSTOP_default;			
				VTOTAL <= VTOTAL_default;			
				
				video_on <= VIDEO_ON_default;

				-- external load

				HLOADSTART <= HLOADSTART_default;	
				HLOADSTOP <= HLOADSTOP_default;		

				VLOADSTART <= VLOADSTART_default;	
				VLOADSTOP <= VLOADSTOP_default;		
				VLOADNEXT <= VLOADNEXT_default;		

				-- memory processing

				MEM_STARTADR <= MEM_STARTADR_default;			
				MEM_LINEOFFSET <= MEM_LINEOFFSET_default;		
				MEM_MODE <= MEM_MODE_default;			

				-- line cache

				lc_add <= lc_add_default;		
				lc_start <= lc_start_default;	
				
				-- i2c
				
				i2c_a_clk_in			<= '1';
				i2c_a_direction_in	<= '0';
				i2c_a_data_write		<= '0';
				
				i2c_b_clk_in			<= '1';
				i2c_b_direction_in	<= '0';
				i2c_b_data_write		<= '0';
				
				-- audio default
				
				dac_fifo_wreq 	<= '0';	
				adc_fifo_rdreq <= '0';	
				
				-- hcore
				
				hcore_ctrl_fifo_wrreq	<= '0';
				
				-- hcore uctrl
				
				o_from_hcores_rdreq 		<= '0';
				hcore_data_fifo_wrreq	<= '0';
		
			else
			
				-- default
			
				lcd_data_write <= '0';
				lcd_data_read	<= '0';
				
				
				-- default timer
				
				timer1_start <= '0';
				timer1_stop  <= '0';
				
				timer2_start <= '0';
				timer2_stop  <= '0';
				
				-- ps 2
				
				ps2_a_data_out_ack 	<= '0';
				ps2_a_data_in_we		<= '0';
				
				ps2_b_data_out_ack 	<= '0';
				ps2_b_data_in_we		<= '0';
				
				-- default irq
				
				clear_irq_stored <= '0';
				
				-- audio default
				
				dac_fifo_wreq 	<= '0';	
				adc_fifo_rdreq <= '0';	
				
				-- hcore
				
				hcore_ctrl_fifo_wrreq	<= '0';
				
				-- hcore uctrl
				
				o_from_hcores_rdreq 		<= '0';
				hcore_data_fifo_wrreq	<= '0';
			
				-- video
				
				PVCOUNT_WE <= '0';
								
				--
			
				if ctrl_port_a_we = '0' then -- read
					case ctrl_port_a_adress is
					
						-----------------------------------------------------------------------------------------------------------	
						-- switches
						-----------------------------------------------------------------------------------------------------------	
					
						when "000" & x"0" => 	to_ctrl_port <= SW(15 downto 0);
						when "000" & x"1" => 	to_ctrl_port <= (others => '0');
													if MODELSIM = true then
														to_ctrl_port(1) <= '1';
													end if;
						-----------------------------------------------------------------------------------------------------------	
						-- de lcd
						-----------------------------------------------------------------------------------------------------------	
						
						when "000" & x"6" => 	to_ctrl_port(15 downto 10) <= (others => '0');
													to_ctrl_port(9)	<= lcd_wait;											
													to_ctrl_port(7 downto 0) <= lcd_data_q;
											
						-----------------------------------------------------------------------------------------------------------	
						-- irq related
						-----------------------------------------------------------------------------------------------------------						
											
						when "000" & x"8" => 	to_ctrl_port <= irq_msk;
						
						when "000" & x"9" => 	to_ctrl_port <= irq_memory;
						
						-----------------------------------------------------------------------------------------------------------	
						-- timer 1/2
						-----------------------------------------------------------------------------------------------------------											
					
						when "000" & x"A" => 	to_ctrl_port <= timer1_captured_value(15 downto 0);
						when "000" & x"B" => 	to_ctrl_port <= timer1_captured_value(31 downto 16);
						
						when "000" & x"D" => 	to_ctrl_port <= timer2_captured_value(15 downto 0);
						when "000" & x"E" => 	to_ctrl_port <= timer2_captured_value(31 downto 16);
						
						-----------------------------------------------------------------------------------------------------------	
						-- video
						-----------------------------------------------------------------------------------------------------------							
								
						when "001" & x"0" =>	to_ctrl_port <= conv_std_logic_vector(video_hcount,16);
						when "001" & x"1" =>	to_ctrl_port <= conv_std_logic_vector(video_vcount,16);
						when "001" & x"2" =>	to_ctrl_port <= x"000" & "000" & PVCOUNT_HIT;
						
						-----------------------------------------------------------------------------------------------------------	
						-- ps2
						-- port a/b
						-----------------------------------------------------------------------------------------------------------	
						
						when "011" & x"0" =>	to_ctrl_port(7 downto 0) 	 	<= ps2_a_data_out;
													to_ctrl_port(8) 				 	<= ps2_a_data_out_ready;
													to_ctrl_port(9) 				 	<= ps2_a_data_in_ready;
													to_ctrl_port(10) 			 	<= ps2_a_data_in_wrs_finish;
													to_ctrl_port(15 downto 11) 	<= (others => '0');
													
						when "011" & x"1" =>	to_ctrl_port(7 downto 0) 		<= ps2_b_data_out;
													to_ctrl_port(8) 					<= ps2_b_data_out_ready;
													to_ctrl_port(9) 					<= ps2_b_data_in_ready;
													to_ctrl_port(10) 				<= ps2_b_data_in_wrs_finish;	
													to_ctrl_port(15 downto 11) 	<= (others => '0');
						
						-----------------------------------------------------------------------------------------------------------	
						-- i2c	
						-- port a/b
						-----------------------------------------------------------------------------------------------------------	
						
						when "011" & x"2"	=> to_ctrl_port(2 downto 0) 		<= (others => '0');
													to_ctrl_port(3)					<= i2c_a_data_read;	
													to_ctrl_port(15 downto 4) 		<= (others => '0');		
						
						when "011" & x"3"	=> to_ctrl_port(2 downto 0) 		<= (others => '0');
													to_ctrl_port(3)					<= i2c_b_data_read;	
													to_ctrl_port(15 downto 4) 	<= (others => '0');
							
						-----------------------------------------------------------------------------------------------------------				
						-- audio 	
						-----------------------------------------------------------------------------------------------------------			
									
						when "011" & x"4"	=>	to_ctrl_port <= adc_fifo_data(47 downto 32);
						when "011" & x"5"	=>	to_ctrl_port <= x"00" & adc_fifo_data(31 downto 24);
						when "011" & x"6"	=>	to_ctrl_port <= adc_fifo_data(23 downto 8);
						when "011" & x"7"	=> to_ctrl_port <= x"00" & adc_fifo_data(7 downto 0);						
										
						
						
						-----------------------------------------------------------------------------------------------------------
						-- hcores
						-----------------------------------------------------------------------------------------------------------
						
						when "011" & x"c"	=>	to_ctrl_port	<= conv_std_logic_vector(hcore_count,16);				
						when "011" & x"f"	=>	to_ctrl_port 	<= x"000" & "000" & hcore_ctrl_fifofull;	
						
						-----------------------------------------------------------------------------------------------------------
						-- hcores UCTRL
						-----------------------------------------------------------------------------------------------------------
					
						when "100" & x"0"	=> to_ctrl_port	<= i_from_hcores_data.data(15 downto 0);
						when "100" & x"1"	=> to_ctrl_port	<= i_from_hcores_data.data(31 downto 16);
						when "100" & x"2"	=> to_ctrl_port	<= i_from_hcores_data.address & i_from_hcores_data.code;
						when "100" & x"3"	=> to_ctrl_port	<= i_from_hcores_data.do_return & i_from_hcores_data.hcore_id & "0000000";
						
						when "100" & x"4"	=> to_ctrl_port <= (others => '0');
						
													if i_from_hcores_empty = '0' then
														to_ctrl_port(0) <= '1';
													end if;
													
													if hcore_data_fifofull = '1' then
														to_ctrl_port(1) <= '1';
													end if;
						
						when others => 		to_ctrl_port <= (others => '0');
					end case;
				else								  -- write
					case ctrl_port_a_adress is
					
						-----------------------------------------------------------------------------------------------------------	
						-- LEDs
						-----------------------------------------------------------------------------------------------------------	
					
						when "000" & x"0" => 	LEDG <= ctrl_port_a_data(8 downto 0);																				
						when "000" & x"1" => 	LEDR <= ctrl_port_a_data(8 downto 0);
						
						-----------------------------------------------------------------------------------------------------------	
						-- HEX (7Segs)
						-----------------------------------------------------------------------------------------------------------	
						
						when "000" & x"2" => 	HEX0 <= not ctrl_port_a_data(6 downto 0); HEX1 <= not ctrl_port_a_data(14 downto 8);
						when "000" & x"3" =>	HEX2 <= not ctrl_port_a_data(6 downto 0); HEX3 <= not ctrl_port_a_data(14 downto 8);
						when "000" & x"4" =>	HEX4 <= not ctrl_port_a_data(6 downto 0); HEX5 <= not ctrl_port_a_data(14 downto 8);
						when "000" & x"5" =>	HEX6 <= not ctrl_port_a_data(6 downto 0); HEX7 <= not ctrl_port_a_data(14 downto 8);
						
						-----------------------------------------------------------------------------------------------------------	
						-- de lcd
						-----------------------------------------------------------------------------------------------------------	
						
						
						when "000" & x"6" =>	lcd_data_rs			<= ctrl_port_a_data(8);
													lcd_data_data  	<= ctrl_port_a_data(7 downto 0);
											
													if ctrl_port_a_data(9) = '0' then
														lcd_data_write 	<= '1';
													else
														lcd_data_read		<= '1';
													end if;
											
						-----------------------------------------------------------------------------------------------------------	
						-- irq related
						-----------------------------------------------------------------------------------------------------------																
											
						when "000" & x"8" =>	irq_msk <= ctrl_port_a_data;	
						
						when "000" & x"9" => clear_irq_value	<= ctrl_port_a_data;	
													clear_irq_stored 	<= '1';
						
						-----------------------------------------------------------------------------------------------------------	
						-- timer 1/2
						-----------------------------------------------------------------------------------------------------------											
						
						when "000" & x"a" =>	timer1_tmp_load_value(15 downto 0) <= ctrl_port_a_data;
						when "000" & x"b" =>	timer1_load_value <= ctrl_port_a_data & timer1_tmp_load_value;
						when "000" & x"c" =>	timer1_start <= ctrl_port_a_data(0);
													timer1_stop  <= ctrl_port_a_data(1);
													timer1_auto_reload <= ctrl_port_a_data(2);
													
													if ctrl_port_a_data(3) = '1' then
														timer1_captured_value <= timer1_current_value;
													end if;
											
						when "000" & x"d" =>	timer2_tmp_load_value(15 downto 0) <= ctrl_port_a_data;
						when "000" & x"e" =>	timer2_load_value <= ctrl_port_a_data & timer2_tmp_load_value;
						when "000" & x"f" =>	timer2_start <= ctrl_port_a_data(0);
													timer2_stop  <= ctrl_port_a_data(1);
													timer2_auto_reload <= ctrl_port_a_data(2);
								
													if ctrl_port_a_data(3) = '1' then
														timer2_captured_value <= timer2_current_value;
													end if;
						
						-----------------------------------------------------------------------------------------------------------	
						-- video
						-----------------------------------------------------------------------------------------------------------						
												
						when "001" & x"0" =>	video_on 				<= ctrl_port_a_data(0);
						when "001" & x"1" =>	HSYNC 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"2" =>	HSTART 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"3" =>	HMEMSTART 				<= conv_integer(ctrl_port_a_data);
						when "001" & x"4" =>	HSTOP 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"5" =>	HTOTAL 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"6" =>	VSYNC 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"7" =>	VSTART 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"8" =>	VSTOP 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"9" =>	VTOTAL 					<= conv_integer(ctrl_port_a_data);
						when "001" & x"a" =>	lc_add_store(15 downto 0) 	:= ctrl_port_a_data;
						when "001" & x"b" =>	lc_add_store(31 downto 16) := ctrl_port_a_data;
						when "001" & x"c" =>	lc_add 					<= lc_add_store;
						when "001" & x"d" =>	lc_start					<= ctrl_port_a_data(10 downto 0);
						when "001" & x"e" =>	PVCOUNT					<= conv_integer(ctrl_port_a_data(10 downto 0));
													PVCOUNT_WE				<= '1';
						
						when "010" & x"0" =>	HLOADSTART 				<= conv_integer(ctrl_port_a_data);
						when "010" & x"1" =>	HLOADSTOP				<= conv_integer(ctrl_port_a_data);
						when "010" & x"2" =>	VLOADSTART 				<= conv_integer(ctrl_port_a_data);
						when "010" & x"3" =>	VLOADSTOP 				<= conv_integer(ctrl_port_a_data);
						when "010" & x"4" =>	VLOADNEXT				<= conv_integer(ctrl_port_a_data);
						
						when "010" & x"5" =>	MEM_STARTADR_store(15 downto 0) 		:= ctrl_port_a_data;
						when "010" & x"6" =>	MEM_STARTADR_store(31 downto 16) 	:= ctrl_port_a_data;
						when "010" & x"7" =>	MEM_LINEOFFSET_store(15 downto 0) 	:= ctrl_port_a_data;
						when "010" & x"8" =>	MEM_LINEOFFSET_store(31 downto 16) 	:= ctrl_port_a_data;
						when "010" & x"9" =>	MEM_MODE_store 							:= ctrl_port_a_data;
						
						when "010" & x"a" =>	MEM_STARTADR 	<= MEM_STARTADR_store; 
													MEM_LINEOFFSET <= MEM_LINEOFFSET_store;
													MEM_MODE 		<= MEM_MODE_store;
								
						-----------------------------------------------------------------------------------------------------------		
						-- ps2
						-- port a
						-----------------------------------------------------------------------------------------------------------		
									
						when "011" & x"0" =>	if ctrl_port_a_data(8) = '1' then
														ps2_a_data_out_ack <= '1'; -- send request to fifo
													end if;
													
													ps2_a_data_in <= ctrl_port_a_data(7 downto 0);	
													
													if ctrl_port_a_data(9) = '1' and ps2_a_data_in_ready ='1' then
														ps2_a_data_in_we <= '1';
													end if;
						-----------------------------------------------------------------------------------------------------------
						-- ps2	
						-- port b
						-----------------------------------------------------------------------------------------------------------		
						
						when "011" & x"1" =>	if ctrl_port_a_data(8) = '1' then
														ps2_b_data_out_ack <= '1'; -- send request to fifo
													end if;
													
													ps2_b_data_in <= ctrl_port_a_data(7 downto 0);	
													
													if ctrl_port_a_data(9) = '1' and ps2_b_data_in_ready ='1' then
														ps2_b_data_in_we <= '1';
													end if;
							
						-----------------------------------------------------------------------------------------------------------			
						-- i2c
						-----------------------------------------------------------------------------------------------------------
						
						when "011" & x"2"	=> -- port a
						
													if ctrl_port_a_data(8) = '1' then				-- clk ena 
														 i2c_a_data_write <= ctrl_port_a_data(0);
													end if;
													
													if ctrl_port_a_data(9) = '1' then				-- data out ena 
														i2c_a_clk_in <= ctrl_port_a_data(1);
													end if;
													
													if ctrl_port_a_data(10) = '1' then				-- direction
														i2c_a_direction_in <= ctrl_port_a_data(2);
													end if;
													
						when "011" & x"3"	=>  -- port b
						
													if ctrl_port_a_data(8) = '1' then				-- clk ena 
														i2c_b_data_write <= ctrl_port_a_data(0);
													end if;
													
													if ctrl_port_a_data(9) = '1' then				-- data out ena 
														 i2c_b_clk_in <= ctrl_port_a_data(1);
													end if;
													
													if ctrl_port_a_data(10) = '1' then			-- direction
														i2c_b_direction_in <= ctrl_port_a_data(2);
													end if;
						
						-----------------------------------------------------------------------------------------------------------					
						-- audio 
						-----------------------------------------------------------------------------------------------------------
									
						when "011" & x"4"	=>	dac_fifo_data(63 downto 48) <= ctrl_port_a_data;
						when "011" & x"5"	=>	dac_fifo_data(47 downto 40) <= ctrl_port_a_data(7 downto 0);
						when "011" & x"6"	=>	dac_fifo_data(39 downto 24) <= ctrl_port_a_data;
						when "011" & x"7"	=>	dac_fifo_data(23 downto 16) <= ctrl_port_a_data(7 downto 0);
						when "011" & x"8"	=>	dac_fifo_data(15 downto 0)  <= ctrl_port_a_data;
													dac_fifo_wreq 					 <= '1';		
								
						when "011" & x"9"	=>	dac_used_low_irq				<= ctrl_port_a_data(7 downto 0);
						when "011" & x"a"	=>	adc_used_high_irq				<= ctrl_port_a_data(7 downto 0);
						when "011" & x"b"	=>	adc_fifo_rdreq					<= '1';
						
						-----------------------------------------------------------------------------------------------------------
						-- hcores
						-----------------------------------------------------------------------------------------------------------
						
						when "011" & x"c"	=>	hcore_ctrl_data.hcore_sel	<= ctrl_port_a_data(7 downto 0);
						when "011" & x"d"	=>	hcore_ctrl_data.address		<= ctrl_port_a_data;
						when "011" & x"e"	=>	hcore_ctrl_data.data			<= ctrl_port_a_data;
						when "011" & x"f"	=>	if hcore_ctrl_fifofull = '0' then
														hcore_ctrl_fifo_wrreq	<= '1';
													end if;	
						-----------------------------------------------------------------------------------------------------------
						-- hcores UCTRL
						-----------------------------------------------------------------------------------------------------------
					
						when "100" & x"0" => temp_from_uctrl_data.data(15 downto 0) 	:= ctrl_port_a_data;
						when "100" & x"1" => temp_from_uctrl_data.data(31 downto 16) 	:= ctrl_port_a_data;
						when "100" & x"2" =>
						when "100" & x"3" => temp_from_uctrl_data.hcore_id := ctrl_port_a_data(14 downto 7);
						
						when "100" & x"4"	=> if ctrl_port_a_data(0) = '1' and hcore_data_fifofull = '0' then
														hcore_data_fifo_wrreq <= '1';
													end if;
													
													if ctrl_port_a_data(1) = '1' and i_from_hcores_empty = '0' then
														o_from_hcores_rdreq <= '1';
													end if;
													
						
									
						when others => 
					end case;
				end if;
			end if;
			
			hcore_data_fifo_data <= from_uctrl_to_raw(temp_from_uctrl_data);
		end if;
	end process;	
	
		
end rtl;
