library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl_counter is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				--
					
				counter_restart 	: in std_logic := '0';
				counter_meas  		: in std_logic := '0';
				counter_overflow	: out std_logic;
				counter_diff		: out std_logic_vector(31 downto 0)
		);
end entity;

architecture rtl of utility_ctrl_counter is

	-- counter
	
	signal icounter_value : std_logic_vector(31 downto 0) := (others => '0');
	signal idiff_value 	 : std_logic_vector(31 downto 0) := (others => '0');
	signal icounter_overflow : std_logic := '0';
	
begin
	
	counter_diff  		<= idiff_value;
	counter_overflow	<= icounter_overflow;
	
	-- counter
	
	process ( clk )
	begin
		if rising_edge(clk) then	
			if reset = '1' then
				icounter_overflow <= '0';
			else
			
				icounter_value <= icounter_value + 1;
				
				if counter_restart = '1' then
					icounter_value 	<= (others => '0'); 
					icounter_overflow <= '0';
				else
					if icounter_value = x"ffffffff" then
						icounter_overflow <= '1';
					end if;
				end if;
				
				if counter_meas = '1' then
					idiff_value <= icounter_value;
				end if;
				
			end if;
		end if;
	end process;	

	

end rtl;
